Package: BEKKs
Title: Multivariate Conditional Volatility Modelling and Forecasting
Version: 1.4.7
Author: Markus J. Fülle [aut, cre],
  Alexander Lange [aut],
  Christian M. Hafner [aut],
  Helmut Herwartz [aut]
Authors@R: 
    c(person("Markus J.",
            "Fülle",
           role = c("aut", "cre"),
           email = "markus.fuelle@gmail.com"),
           person("Alexander", "Lange", 
          role = c("aut"),
          email = "alexander.lange@uni-goettingen.de"),
      person("Christian M.", "Hafner", 
           role = c("aut"), 
           email = "christian.hafner@uclouvain.be"),
      person("Helmut", "Herwartz", 
           role = c("aut"), 
           email = "hherwartz@uni-goettingen.de"))
Maintainer: Markus J. Fülle <markus.fuelle@gmail.com>
Description: Methods and tools for estimating, simulating and forecasting of so-called BEKK-models (named after Baba, Engle, Kraft and Kroner) based on the fast Berndt–Hall–Hall–Hausman (BHHH) algorithm described in Hafner and Herwartz (2008) <doi:10.1007/s00184-007-0130-y>. For an overview, we refer the reader to Fülle et al. (2024) <doi:10.18637/jss.v111.i04>.  
Depends: R (>= 3.5.0)
Imports: Rcpp, reshape2, ggplot2, mathjaxr, gridExtra, grid, ggfortify,
        parallel, xts, stats, future, future.apply, ks, lubridate,
        utils, pbapply, numDeriv, moments
LinkingTo: Rcpp, RcppArmadillo
NeedsCompilation: yes
SystemRequirements: C++17
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: true
Suggests: testthat (>= 2.1.0)
RdMacros: mathjaxr
RoxygenNote: 7.3.3
Packaged: 2026-05-03 13:26:14 UTC; Markus
Repository: CRAN
Date/Publication: 2026-05-03 14:00:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-03 14:32:12 UTC; unix
Archs: BEKKs.so.dSYM

Package: BEKKs
Title: Multivariate Conditional Volatility Modelling and Forecasting
Version: 1.4.7
Author: Markus J. Fülle [aut, cre],
  Alexander Lange [aut],
  Christian M. Hafner [aut],
  Helmut Herwartz [aut]
Authors@R: 
    c(person("Markus J.",
            "Fülle",
           role = c("aut", "cre"),
           email = "markus.fuelle@gmail.com"),
           person("Alexander", "Lange", 
          role = c("aut"),
          email = "alexander.lange@uni-goettingen.de"),
      person("Christian M.", "Hafner", 
           role = c("aut"), 
           email = "christian.hafner@uclouvain.be"),
      person("Helmut", "Herwartz", 
           role = c("aut"), 
           email = "hherwartz@uni-goettingen.de"))
Maintainer: Markus J. Fülle <markus.fuelle@gmail.com>
Description: Methods and tools for estimating, simulating and forecasting of so-called BEKK-models (named after Baba, Engle, Kraft and Kroner) based on the fast Berndt–Hall–Hall–Hausman (BHHH) algorithm described in Hafner and Herwartz (2008) <doi:10.1007/s00184-007-0130-y>. For an overview, we refer the reader to Fülle et al. (2024) <doi:10.18637/jss.v111.i04>.  
Depends: R (>= 3.5.0)
Imports: Rcpp, reshape2, ggplot2, mathjaxr, gridExtra, grid, ggfortify,
        parallel, xts, stats, future, future.apply, ks, lubridate,
        utils, pbapply, numDeriv, moments
LinkingTo: Rcpp, RcppArmadillo
NeedsCompilation: yes
SystemRequirements: C++17
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: true
Suggests: testthat (>= 2.1.0)
RdMacros: mathjaxr
RoxygenNote: 7.3.3
Packaged: 2026-05-03 13:26:14 UTC; Markus
Repository: CRAN
Date/Publication: 2026-05-03 14:00:02 UTC
Built: R 4.4.3; x86_64-apple-darwin20; 2026-05-03 16:08:17 UTC; unix
Archs: BEKKs.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(VaR,bekkFit)
S3method(VaR,bekkForecast)
S3method(backtest,bekkFit)
S3method(bekk_fit,bekk)
S3method(bekk_fit,bekka)
S3method(bekk_fit,dbekk)
S3method(bekk_fit,dbekka)
S3method(bekk_fit,sbekk)
S3method(bekk_fit,sbekka)
S3method(logLik,bekkFit)
S3method(plot,backtest)
S3method(plot,bekkFit)
S3method(plot,bekkMC)
S3method(plot,var)
S3method(plot,virf)
S3method(portmanteau.test,bekkFit)
S3method(predict,bekk)
S3method(predict,bekka)
S3method(predict,dbekk)
S3method(predict,dbekka)
S3method(predict,sbekk)
S3method(predict,sbekka)
S3method(print,backtest)
S3method(print,bekkFit)
S3method(print,var)
S3method(process_object,bekkFit)
S3method(process_object,bekkSpec)
S3method(residuals,bekkFit)
S3method(simulate,bekk)
S3method(simulate,bekka)
S3method(simulate,dbekk)
S3method(simulate,dbekka)
S3method(simulate,sbekk)
S3method(simulate,sbekka)
S3method(summary,backtest)
S3method(summary,bekk)
S3method(summary,bekka)
S3method(summary,dbekk)
S3method(summary,dbekka)
S3method(summary,sbekk)
S3method(summary,sbekka)
S3method(virf,bekk)
S3method(virf,dbekk)
S3method(virf,sbekk)
export(VaR)
export(backtest)
export(bekk_fit)
export(bekk_spec)
export(portmanteau.test)
export(virf)
import(future)
import(future.apply)
import(ggfortify)
import(ggplot2)
import(grid)
import(gridExtra)
import(ks)
import(lubridate)
import(mathjaxr)
import(moments)
import(numDeriv)
import(pbapply)
import(reshape2)
import(stats)
import(utils)
import(xts)
importFrom(Rcpp,sourceCpp)
useDynLib(BEKKs)

#' @import ggplot2
#' @import ggfortify
#' @import reshape2
#' @export

plot.var <- function(x, ...) {

  obs <- NULL
  V1 <- NULL
  lower <- NULL
  type <- NULL
  upper <- NULL

  if (any(class(x) == 'bekkFit')) {
    if(is.null(x$portfolio_weights)) {
      if (inherits(x$bekk$data, c("ts","xts","zoo"))) {
        autoplot(x$VaR) + theme_bw() + ylab('VaR')
      } else {
        x$VaR$obs <- 1:nrow(x$VaR)
        VaR <- melt(x$VaR, id = 'obs')
        ggplot(VaR) + geom_line(aes(x = obs, y = value)) + theme_bw() + xlab('') + ylab('VaR') + facet_wrap(~variable, scales = 'free_y', ncol = 1)
      }
    } else {
      if (inherits(x$bekk$data, c("ts","xts","zoo"))) {
        autoplot(x$VaR) + theme_bw() + ylab('VaR') + ggtitle('Portfolio VaR')
      } else {
        ggplot(x$VaR) + geom_line(aes(x = 1:nrow(x$VaR), y = V1)) + theme_bw() + xlab('') + ylab('VaR') + ggtitle('Portfolio VaR')
      }
    }
  } else if (any(class(x) == 'bekkForecast')) {
    if(is.null(x$portfolio_weights)) {
      sample <- x$VaR[1:(nrow(x$VaR)-x$n.ahead),]
      forc <- x$VaR[(nrow(x$VaR)-x$n.ahead+1):nrow(x$VaR),]
      cb_lower <- x$VaR_lower[(nrow(x$VaR)-x$n.ahead+1):nrow(x$VaR),]
      cb_upper <- x$VaR_upper[(nrow(x$VaR)-x$n.ahead+1):nrow(x$VaR),]

      sample$obs <- as.character(1:nrow(sample))
      forc$obs <- as.character((nrow(sample)+1):(nrow(sample)+x$n.ahead))
      cb_lower$obs <- as.character((nrow(sample)+1):(nrow(sample)+x$n.ahead))
      cb_upper$obs <- as.character((nrow(sample)+1):(nrow(sample)+x$n.ahead))

      sample <- sample[(nrow(sample)-4*x$n.ahead):nrow(sample),]
      sample$type <- as.factor('Sample')
      forc$type <- as.factor('Forecast')
      cb_lower$type <- as.factor('Forecast')
      cb_upper$type <- as.factor('Forecast')

      cb_l <- melt(cb_lower, id = c('obs', 'type'))
      cb_u <- melt(cb_upper, id = c('obs', 'type'))

      cb <- cbind(cb_l, cb_u$value)
      colnames(cb)[4:5] <- c('lower', 'upper')



      total <- rbind(sample, forc)

      VaR <- melt(total, id = c('obs', 'type'))

      cc <- merge(VaR, cb, all.x = TRUE, all.y = TRUE)

      if (x$n.ahead > 1) {
        ggplot(cc, aes(x = obs, y = value)) +
          geom_line(aes(y = lower, group = type, color = type, linetype = type), na.rm = TRUE, color = 'red') +
          geom_line(aes(y = upper, group = type, color = type, linetype = type), na.rm = TRUE, color = 'red') +
          geom_line(aes(group = type, color = type)) +
          geom_point(aes(shape = type)) +
          theme_bw() + xlab('') + ylab('VaR') +
          scale_color_manual(values = c('black', 'blue')) +
          facet_wrap(~variable, scales = 'free_y', ncol = 1) +
          theme(legend.position="bottom", legend.title = element_blank())
      } else {
        ggplot(cc, aes(x = obs, y = value)) +
          geom_line(data = cc[cc$type == 'Sample',], aes(x = obs, y = value, group = type)) +
          geom_errorbar( aes(ymin=lower, ymax=upper), width=.2, color = 'red') +
          geom_point(aes(x = obs, y = value, shape = type), size = 2.5) +
          theme_bw() + xlab('') + ylab('VaR') +
          scale_color_manual(values = c('black', 'blue')) +
          facet_wrap(~variable, scales = 'free_y', ncol = 1) +
          theme(legend.position="bottom", legend.title = element_blank())
      }
    } else {
      sample <- as.data.frame(x$VaR[1:(nrow(x$VaR)-x$n.ahead),])
      forc <- as.data.frame(x$VaR[(nrow(x$VaR)-x$n.ahead+1):nrow(x$VaR),])
      cb_lower <- as.data.frame(x$VaR_lower[(nrow(x$VaR)-x$n.ahead+1):nrow(x$VaR),])
      cb_upper <- as.data.frame(x$VaR_upper[(nrow(x$VaR)-x$n.ahead+1):nrow(x$VaR),])

      sample$obs <- as.character(1:nrow(sample))
      forc$obs <- as.character((nrow(sample)+1):(nrow(sample)+x$n.ahead))
      cb_lower$obs <- as.character((nrow(sample)+1):(nrow(sample)+x$n.ahead))
      cb_upper$obs <- as.character((nrow(sample)+1):(nrow(sample)+x$n.ahead))

      sample <- sample[(nrow(sample)-4*x$n.ahead):nrow(sample),]
      sample$type <- as.factor('Sample')
      forc$type <- as.factor('Forecast')

      colnames(sample)[1] <- colnames(forc)[1] <- colnames(cb_lower)[1] <- colnames(cb_upper)[1] <- 'V1'

      cb_lower$type <- as.factor('Forecast')
      cb_upper$type <- as.factor('Forecast')

      cb_l <- melt(cb_lower, id = c('obs', 'type'))
      cb_u <- melt(cb_upper, id = c('obs', 'type'))

      cb <- cbind(cb_l, cb_u$value)
      colnames(cb)[4:5] <- c('lower', 'upper')

      total <- rbind(sample, forc)

      VaR <- melt(total, id = c('obs', 'type'))

      cc <- merge(VaR, cb, all.x = TRUE, all.y = TRUE)

      if (x$n.ahead > 1) {
        ggplot(cc, aes(x = obs, y = value)) +
          geom_line(aes(y = lower, group = type, linetype = type), color = 'red', na.rm = TRUE) +
          geom_line(aes(y = upper, group = type, linetype = type), color = 'red', na.rm = TRUE) +
          geom_line(aes(group = type, color = type)) +
          geom_point(aes(shape = type)) +
          theme_bw() + xlab('') + ylab('VaR') +
          scale_color_manual(values = c('black', 'blue')) +
          theme(legend.position="bottom", legend.title = element_blank()) + ggtitle('Portfolio VaR')
      } else {
        ggplot(cc, aes(x = obs, y = value)) +
          geom_line(data = cc[cc$type == 'Sample',], aes(x = obs, y = value, group = type)) +
          geom_errorbar( aes(ymin=lower, ymax=upper), width=.2, color = 'red') +
          geom_point(aes(x = obs, y = value, shape = type), size = 2.5) +
          theme_bw() + xlab('') + ylab('VaR') +
          scale_color_manual(values = c('black', 'blue')) +
          theme(legend.position="bottom", legend.title = element_blank()) + ggtitle('Portfolio VaR')
      }


          }
  }


}

#' Calculating Value-at-Risk (VaR)
#'
#' @description Method for calculating VaR from estimated covariance processes (\link{bekk_fit}) or predicted covariances (\link{predict}).
#'
#' @param x An object of class "bekkFit" from the function \link{bekk_fit} or an object of class "bekkForecast" from the function \link{predict}.
#' @param p A numerical value that determines the confidence level. The default value is set at 0.99 in accordance with the Basel Regulation.
#' @param portfolio_weights A vector determining the portfolio weights to calculate the portfolio VaR. If set to "NULL", the univariate VaR for each series are calculated.
#' @param distribution A character string determining the assumed distribution of the residuals. Implemented are "normal", "empirical" and "t". The default is using the empirical distribution of the residuals.
#' @return  Returns a S3 class "var" object containing the VaR forecast and respective confidence bands.
#' @examples
#' \donttest{
#'
#' data(StocksBonds)
#' obj_spec <- bekk_spec()
#' x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)
#'
#' # single VaRs of series
#' x2 <- VaR(x1, distribution="normal")
#' plot(x2)
#'
#' # VaR of equally-weighted portfolio
#' portfolio_weights <- c(0.5, 0.5)
#' x3 <- VaR(x1, portfolio_weights = portfolio_weights)
#' plot(x3)
#'
#' # VaR of traditional 30/70 weighted bond and stock portfolio
#' portfolio_weights <- c(0.3, 0.7)
#' x4 <- VaR(x1, portfolio_weights = portfolio_weights)
#' plot(x4)
#'
#' }
#'
#' @import xts
#' @import stats
#' @import moments
#' @export

VaR <- function(x, p = 0.99, portfolio_weights = NULL, distribution = "empirical") {

  UseMethod('VaR')
}

#' @export
VaR.bekkFit <-  function(x, p = 0.99, portfolio_weights = NULL, distribution = "empirical" )
{
  if(nrow(x$data) < 1000 && distribution == "empirical"){
    stop("Using the empirical distribution is not stable for time series with less than 1000 observations!")
  }

  alpha = p
  N = ncol(x$data)
  n = nrow(x$data)
  match.arg(distribution, c("empirical", "t", "normal"))

  if(length(portfolio_weights)!= N && !is.null(portfolio_weights)){
    stop("Portfolio weights do no match number of time series")
  }


  #specify quantiles here
  if(distribution == "t"){
    #fit skewed t
    skew_t <- function(i){
      kurtos = moments::kurtosis(x$e_t[,i])-3
      df = 6/kurtos+4
      if(df <= 4){
        df=4.001
      }
      return(qt(1-alpha, df = df)/sqrt(df/(df-2)))
    }

    qtls <- sapply(1:N, skew_t)
  }else if(distribution == "empirical"){
    empirical <- function(i){
      quantile(x$e_t[,i],1-alpha)
    }
    qtls <- sapply(1:N, empirical)

  } else if(distribution == "normal"){
    #fit skewed t

    qtls <- rep(qnorm(1-alpha),ncol(x$data))
  } else{
    qtls <- rep(qnorm(1-alpha),ncol(x$data))
  }
  #quantile(x$e_t, )

  if (is.null(portfolio_weights)) {
    columns = ncol(x$data)
    csd <- extract_csd(x)
    VaR <- matrix(NA, nrow = nrow(x$data), ncol = ncol(x$data))

    for(i in 1:n) {
       for(j in 1: ncol(x$data)){
      VaR[i, j] =  sqrt(matrix(x$H_t[i,],N,N)[j,j]) * qtls[j]
        }
    }
    VaR <- as.data.frame(VaR)
    for(column in 1:columns) {
      r = as.vector(na.omit(x$data[,column]))
      if (!is.numeric(r)) stop("The selected column is not numeric")
      m2 =  csd[, column]
      #VaR[, column] = - qnorm(alpha)*m2
      #VaR <- as.data.frame(VaR)

      for (i in 1:ncol(x$data)) {
        colnames(VaR)[i] <- paste('VaR of', colnames(x$data)[i])
      }
    }
  } else {
    if(distribution == "t"){
      #fit skewed t

        kurtos = moments::kurtosis(x$e_t%*%portfolio_weights)-3
        df = 6/kurtos+4
        if(df <= 4){
          df=4.001
        }
        qtls <- qt(1-alpha, df = df)/sqrt(df/(df-2))
      }
    else if(distribution == "empirical"){

      qtls <- quantile(x$e_t%*%portfolio_weights,1-alpha)

      } else if(distribution == "normal"){
      #fit skewed t

      qtls <-qnorm(1-alpha)
    } else{
      qtls <- qnorm(1-alpha)
    }


    VaR <- matrix(NA, nrow = nrow(x$data), ncol = 1)
    for(i in 1:nrow(x$H_t)) {
      VaR[i,] <- qtls*sqrt(portfolio_weights%*%matrix(x$H_t[i,], ncol = ncol(x$data))%*%portfolio_weights)
      #VaR[i,] <- portfolio_weights%*%eigen_value_decomposition(matrix(x$H_t[i,], ncol = ncol(x$data)))%*%qtls
      }
    VaR <- as.data.frame(VaR)
  }

  if (inherits(x$data, "ts")) {
    VaR <- ts(VaR, start = time(x$data)[1], frequency = frequency(x$data))
  }else if(inherits(x$data, "xts") || inherits(x$data, "zoo") ){
    VaR <- xts(VaR, order.by = time(x$data[1:nrow(x$data),]))
      }

  result <- list(VaR = VaR,
                 p = p,
                 portfolio_weights = portfolio_weights,
                 bekk = x)
  class(result) <- c('var', 'bekkFit')
  return(result)
}


#' @export
VaR.bekkForecast <-  function(x, p = 0.99, portfolio_weights = NULL, distribution = "empirical")
{

  if(nrow(x$bekkfit$data) < 1000 && distribution == "empirical"){
    stop("Using the empirical distribution is not stable for time series with less than 1000 observations!")
  }

  alpha = p


  obj <- x$bekkfit
  obj$H_t <- rbind(x$bekkfit$H_t, x$H_t_forecast)
  obj$sigma_t <- rbind(x$bekkfit$sigma_t, x$volatility_forecast)
  N = ncol(obj$data)
  n = nrow(obj$H_t)
  if(length(portfolio_weights)!= N && !is.null(portfolio_weights)){
    stop("Portfolio weights do no match number of time series")
  }

  if(distribution == "t"){
    #fit skewed t
    skew_t <- function(i){
      kurtos = moments::kurtosis(obj$e_t[,i])-3
      df = 6/kurtos+4
      if(df <= 4){
        df=4.001
      }
      return(qt(1-alpha, df = df)/sqrt(df/(df-2)))
    }

    qtls <- sapply(1:N, skew_t)
  }else if(distribution == "empirical"){
    empirical <- function(i){
      quantile(obj$e_t[,i],1-alpha)
    }
    qtls <- sapply(1:N, empirical)

  } else if(distribution == "normal"){
    #fit skewed t

    qtls <- rep(qnorm(1-alpha),ncol(obj$data))
  } else{
    qtls <- rep(qnorm(1-alpha),ncol(obj$data))
  }

  if (is.null(portfolio_weights)) {
    columns = ncol(x$bekkfit$data)
    #csd <- extract_csd(obj)
    VaR <- matrix(NA, nrow = nrow(x$bekkfit$data) + x$n.ahead, ncol = ncol(x$bekkfit$data))


    for(i in 1:nrow(obj$H_t)) {
      for(j in 1: ncol(x$bekkfit$data)){
        VaR[i, j] =  sqrt(matrix(obj$H_t[i,],N,N)[j,j]) * qtls[j]
      }
    }
    VaR <- as.data.frame(VaR)

      for (i in 1:ncol(x$bekkfit$data)) {
        colnames(VaR)[i] <- paste('VaR of', colnames(x$bekkfit$data)[i])
      }


    # Confidence intervals
    H_t_lower <- rbind(x$bekkfit$H_t[-nrow(x$bekkfit$H_t),], x$H_t_lower_conf_band)
    H_t_upper <- rbind(x$bekkfit$H_t[-nrow(x$bekkfit$H_t),], x$H_t_upper_conf_band)

    colnames(x$volatility_lower_conf_band) = colnames(x$volatility_forecast)
    obj$sigma_t <- rbind(x$bekkfit$sigma_t, x$volatility_lower_conf_band)

    csd_lower <- extract_csd(obj)
    VaR_lower <- matrix(NA, nrow = nrow(x$bekkfit$data) + x$n.ahead, ncol = ncol(x$bekkfit$data))

    for(i in 1:nrow(obj$H_t)) {
      for(j in 1: ncol(x$bekkfit$data)){
        VaR_lower[i, j] =  sqrt(matrix(H_t_lower[i,],N,N)[j,j]) * qtls[j]
      }
    }

      VaR_lower <- as.data.frame(VaR_lower)

      for (i in 1:ncol(x$bekkfit$data)) {
        colnames(VaR_lower)[i] <- paste('VaR of', colnames(x$bekkfit$data)[i])
      }

    colnames(x$volatility_upper_conf_band) = colnames(x$volatility_forecast)
    obj$sigma_t <- rbind(x$bekkfit$sigma_t, x$volatility_upper_conf_band)

    csd_upper <- extract_csd(obj)
    VaR_upper <- matrix(NA, nrow = nrow(x$bekkfit$data) + x$n.ahead, ncol = ncol(x$bekkfit$data))
    for(i in 1:nrow(obj$H_t)) {
      for(j in 1: ncol(x$bekkfit$data)){
        VaR_upper[i, j] =  sqrt(matrix(H_t_upper[i,],N,N)[j,j]) * qtls[j]
      }
    }

      VaR_upper <- as.data.frame(VaR_upper)

      for (i in 1:ncol(x$bekkfit$data)) {
        colnames(VaR_upper)[i] <- paste('VaR of', colnames(x$bekkfit$data)[i])
      }



  } else {

    if(distribution == "t"){
      #fit skewed t

      kurtos = moments::kurtosis(obj$e_t%*%portfolio_weights)-3
      df = 6/kurtos+4
      if(df <= 4){
        df=4.001
      }
      qtls <- qt(1-alpha, df = df)/sqrt(df/(df-2))
    }
    else if(distribution == "empirical"){

      qtls <- quantile(obj$e_t%*%portfolio_weights,1-alpha)

    } else if(distribution == "normal"){
      #fit skewed t

      qtls <-qnorm(1-alpha)
    } else{
      qtls <- qnorm(1-alpha)
    }

    VaR <- matrix(NA, nrow = nrow(x$bekkfit$data) + x$n.ahead, ncol = 1)

    for(i in 1:nrow(obj$H_t)) {
      #VaR[i,] <- -qnorm(alpha)*sqrt(portfolio_weights%*%matrix(x$H_t[i,], ncol = ncol(x$data))%*%portfolio_weights)
      VaR[i,] <- sqrt(portfolio_weights%*%matrix(obj$H_t[i,], ncol = ncol(x$bekkfit$data))%*%portfolio_weights)*qtls
    }
    # for(i in 1:nrow(obj$H_t)) {
    #   VaR[i,] <- -qnorm(alpha)*sqrt(portfolio_weights%*%matrix(obj$H_t[i,], ncol = ncol(x$bekkfit$data))%*%portfolio_weights)
    # }
    VaR <- as.data.frame(VaR)

    # Confidence intervals
    VaR_lower <- matrix(NA, nrow = nrow(x$bekkfit$data) + x$n.ahead, ncol = 1)
    VaR_upper <- matrix(NA, nrow = nrow(x$bekkfit$data) + x$n.ahead, ncol = 1)

    H_t_lower <- rbind(x$bekkfit$H_t[-nrow(x$bekkfit$H_t),], x$H_t_lower_conf_band)
    H_t_upper <- rbind(x$bekkfit$H_t[-nrow(x$bekkfit$H_t),], x$H_t_upper_conf_band)

    # for(i in 1:nrow(obj$H_t)) {
    #   VaR_lower[i,] <- -qnorm(alpha)*sqrt(portfolio_weights%*%matrix(H_t_lower[i,], ncol = ncol(x$bekkfit$data))%*%portfolio_weights)
    #   VaR_upper[i,] <- -qnorm(alpha)*sqrt(portfolio_weights%*%matrix(H_t_upper[i,], ncol = ncol(x$bekkfit$data))%*%portfolio_weights)
    # }

    for(i in 1:nrow(obj$H_t)) {
      VaR_lower[i,] <- sqrt(portfolio_weights%*%matrix(H_t_lower[i,], ncol = ncol(x$bekkfit$data))%*%portfolio_weights)*qtls
      VaR_upper[i,] <- sqrt(portfolio_weights%*%matrix(H_t_upper[i,], ncol = ncol(x$bekkfit$data))%*%portfolio_weights)*qtls
    }
    VaR_lower <- as.data.frame(VaR_lower)
    VaR_upper <- as.data.frame(VaR_upper)
  }

  if (inherits(x$data, "ts")) {
    VaR <- ts(VaR, start = time(x$bekkfit$data)[1], frequency = frequency(x$bekkfit$data))
  }  else if(inherits(x$data, "xts") || inherits(x$data, "zoo") ){
    VaR <- xts(VaR, order.by = time(x$data[1:nrow(x$data),]))
  }

  result <- list(VaR = VaR,
                 VaR_lower = VaR_lower,
                 VaR_upper = VaR_upper,
                 p = p,
                 portfolio_weights = portfolio_weights,
                 n.ahead = x$n.ahead,
                 bekk = x)
  class(result) <- c('var', 'bekkForecast')
  return(result)
}

#' @import ggplot2
#' @import grid
#' @import gridExtra
#' @export

plot.bekkFit <- function(x, diagnostic = FALSE, ...){

  V1 <- NULL
  l <- NULL

  if (diagnostic == FALSE) {
    trianglePlotGrid <- function(plots){
      #take a list of plots and returns a single plot where the elements in the list arranged in a triangular grid
      #plots should be a list of 1 or 3 or 6... plots to be arranged in a trianglular structure with 1 plot in the top row
      ncols <- (-1 + sqrt(1 + 8*length(plots)))/2

      grobs <- list()

      k <-1

      for (i in 1:ncols) {
        for (j in 1:ncols) {
          if (i <= j) {
            grobs[[length(grobs)+1]] <- plots[[k]]
            k <- k+1
          } else {
            grobs[[length(grobs)+1]] <- nullGrob()
          }
        }
      }
      do.call("grid.arrange", c(grobs, ncol=ncols))
    }




    if (inherits(x$sigma_t, c("ts","zoo","xts")) ){
      plist <- vector(mode = "list", length = ncol(x$sigma_t))
      xxc <- colnames(x$sigma_t)
      for (i in 1:ncol(x$sigma_t)) {
        if (grepl('correlation', xxc[i])) {
          plist[[i]] <- suppressMessages(autoplot(x$sigma_t[,i]) + theme_bw() + ylab("") + ylim(-1,1) + geom_hline(yintercept = 0, col = 'red'))
        } else {
          plist[[i]] <- autoplot(x$sigma_t[,i]) + theme_bw() + ylab("")
        }
      }
    } else {
      xxc <- colnames(x$sigma_t)

      plist <- vector(mode = "list", length = ncol(x$sigma_t))

      for (i in 1:ncol(x$sigma_t)) {
        xx1 <- data.frame(x$sigma_t[,i])
        colnames(xx1) <- 'V1'
        if (grepl('correlation', xxc[i])) {
          plist[[i]] <- ggplot(xx1, aes(x = 1:nrow(x$sigma_t), y = V1)) + geom_line() + theme_bw()+ xlab('') + ylab('') + ylim(-1,1) + geom_hline(yintercept = 0, col = 'red')
        } else {
          plist[[i]] <- ggplot(xx1, aes(x = 1:nrow(x$sigma_t), y = V1)) + geom_line() + theme_bw()+ xlab('') + ylab('')
        }
      }
    }


    for (i in 1:ncol(x$sigma_t)) {
      plist[[i]] <- plist[[i]] + ggtitle(xxc[i])
    }

    trianglePlotGrid(plist)
  } else {
    dat <- data.frame(l = x$likelihood_iter)
    ggplot(dat, aes(x = 1:nrow(dat), y = l)) + geom_line() + ggtitle('BHHH-algorithm convergence') +
      xlab('Iteration') + ylab('log-likelihood') + theme_bw()
  }

}

.onLoad <- function(libname, pkgname) {

  # CRAN OMP THREAD LIMIT
  Sys.setenv("OMP_THREAD_LIMIT" = 1)
}
.onUnload <- function(libpath) {
  library.dynam.unload("BEKKs", libpath)
  invisible(NULL)
}

#' @import future
#' @import future.apply


bekk_mc_eval <- function(object, spec, sample_sizes, iter, nc = 1) {
  xx <- process_object(object)
  theta <- xx$theta

  mse <- rep(NA, length(sample_sizes))
  index <- 1

  for(j in sample_sizes) {
    print(paste('Sample size: ', j))
    sim_dat <- vector(mode = "list", iter)

    sim_dat <- future_lapply(1:iter, function(x){simulate(object, nobs = j)}, future.seed=TRUE)

    dd <- future_lapply(sim_dat, function(x){bekk_fit(spec = spec, data = x, max_iter = 200)},future.seed=TRUE)

    mse[index] <- sum(unlist(lapply(dd, RMSE, theta_true = theta)))/iter
    print(mse[index])
    index <- index +1

  }

  result <- data.frame(Sample_size = sample_sizes, MSE = mse)
  colnames(result) <- c('Sample', 'MSE')
  result <- list(result)

  class(result) <- 'bekkMC'
  return(result)
}


RMSE <- function(x, theta_true) {
  theta_est <- x$theta
  return(mean(sqrt(((theta_true - theta_est) / theta_true)^2)))
}
#'@exportS3Method
plot.bekkMC <- function(x, ...) {
  Sample <- NULL
  MSE <- NULL

  msep <- x[[1]]
  ggplot(msep) + geom_line(aes(x = Sample, y = MSE)) + theme_bw()
}

#' Estimating multivariate BEKK-type volatility models
#'
#' @description Method for fitting a variety of N-dimensional BEKK models.
#'
#' @param spec An object of class "bekkSpec" from function \link{bekk_spec}.
#' @param data A multivariate data object. Can be a numeric matrix or ts/xts/zoo object.
#' @param QML_t_ratios Logical. If QML_t_ratios = 'TRUE', the t-ratios of the BEKK parameter matrices
#'                     are exactly calculated via second order derivatives.
#' @param max_iter Maximum number of BHHH algorithm iterations.
#' @param crit Determines the precision of the BHHH algorithm.
#' @return  Returns a S3 class "bekkFit" object containing the estimated parameters, t-values, standard errors and volatility process of the model defined by the BEKK_spec object.
#'
#' @details The BEKK optimization routine is based on the Berndt–Hall–Hall–Hausman (BHHH) algorithm and is inspired by the study of Hafner and Herwartz (2008).
#' The authors provide analytical formulas for the score and Hessian of several MGARCH models in a QML framework and show that analytical derivations significantly outperform numerical methods.
#'
#' @references Hafner and Herwartz (2008). Analytical quasi maximum likelihood inference in multivariate volatility models. Metrika, 67, 219-239.
#' @references Fülle, M. J., A. Lange, C. M. Hafner, and H. Herwartz (2024). BEKKs: An R package for estimation of
#' conditional volatility of multivariate time series. Journal of Statistical Software 111 (4), 1–34. <doi:10.18637/jss.v111.i04>.
#' @examples
#' \donttest{
#'
#' data(StocksBonds)
#'
#' # Fitting a symmetric BEKK model
#' obj_spec <- bekk_spec()
#' x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)
#'
#' summary(x1)
#'
#' plot(x1)
#'
#' # Fitting an asymmetric BEKK model
#' obj_spec <- bekk_spec(model = list(type = "bekk", asymmetric = TRUE))
#' x1 <- bekk_fit(obj_spec, StocksBonds)
#'
#' summary(x1)
#'
#' plot(x1)
#'
#' # Fitting a symmetric diagonal BEKK model
#' obj_spec <- bekk_spec(model = list(type = "dbekk", asymmetric = FALSE))
#' x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)
#'
#' summary(x1)
#'
#' plot(x1)
#'
#'
#' # Fitting a symmetric scalar BEKK model
#' obj_spec <- bekk_spec(model = list(type = "sbekk", asymmetric = FALSE))
#' x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)
#'
#' summary(x1)
#'
#' plot(x1)
#'
#' }
#' @import xts
#' @import stats
#' @import utils
#' @export

bekk_fit <- function(spec, data, QML_t_ratios = FALSE,
                     max_iter = 50, crit = 1e-9){

  if (!inherits(spec, 'bekkSpec')) {
    stop('Please provide an object of class "bekkSpec" for spec.')
  }

  if (any(is.na(data))) {
    stop("\nNAs in data.\n")
  }
  if (is.null(ncol(data))) {
    stop("The data matrix should contain at least two variables.")
  }
  if (ncol(data) < 2) {
    stop("The data matrix should contain at least two variables.")
  }
  if (is.null(colnames(data))) {
    colnames(data) <- paste("y", 1:ncol(data), sep = "")
  }

  UseMethod('bekk_fit')

}

#' @export
bekk_fit.bekk <- function(spec, data, QML_t_ratios = FALSE,
                          max_iter = 50, crit = 1e-9) {

  init_values <- spec$init_values
  N <- ncol(data)

  if(!is.numeric(init_values)) {
    if (is.null(init_values)) {
      theta <- gridSearch_BEKK(data)
      theta <- theta[[1]]
    } else if (init_values == 'random') {
      cat('Generating starting values \n')

      theta <- random_grid_search_BEKK(data)
      theta <- theta[[1]]
    } else if (init_values == 'simple') {
      uncond_var <- crossprod(data)/nrow(data)
      A <- matrix(0, ncol = N, nrow = N)
      G <- matrix(0, ncol = N, nrow = N)
      C <- matrix(0, ncol = N, nrow = N)
      #th0=numeric(2*n^2+n*(n+1)/2)

      diag(A) <- 0.3
      diag(G) <- 0.92
      diag(C) <- 0.05*diag(uncond_var)


      for (i in 1:N){
        for (j in seq(i,N)){

          cij <- uncond_var[i, j]/sqrt(uncond_var[i, i]*uncond_var[j, j])
          C[i,j] <- cij*sqrt(C[i, i]*C[j, j])
          C[j,i] <- C[i, j]

        }
      }

      C = t(chol(C))
      C0 = C[,1]

      if (N > 2) {
        for (i in 2:(N-1)){
          C0 = c(C0, C[i:N, i])
        }
      }

      C0 = c(C0, C[N, N])

      theta = c(C0, c(A), c(G))
     }
  } else {
    if(length(init_values) != 2 * N^2 + N * (N + 1)/2) {
      stop('Number of initial parameter does not match dimension of data.')
    }
    theta <- init_values
  }

  theta <- matrix(theta, ncol =1)

  params <- bhh_bekk(data, theta, max_iter, crit)

  if (QML_t_ratios == TRUE) {
    tratios <- QML_t_ratios(params$theta, data)
    tratios_mat <- coef_mat(tratios, N)
    sds <- QML_sd(params$theta, data)
    sd_mat <- coef_mat(sds, N)
  } else {
    tratios_mat <- coef_mat(params$t_val, N)
    sd_mat <- coef_mat(params$sd, N)
  }

  param_mat <- coef_mat(params$theta, N)


  var_process <- sigma_bekk(data, t(param_mat$c0), param_mat$a, param_mat$g)
  sigma_t <- as.data.frame(var_process$sigma_t)
  colnames(sigma_t) <- rep(1, N^2)

  k <- 1
  k2 <- 1
  for (i in 1:N) {
    for (j in 1:N) {
      if (i == j) {
        colnames(sigma_t)[k2] <- paste('Conditional standard deviation of \n', colnames(data)[k])
        k <- k + 1
        k2 <- k2 +1
      } else {
        colnames(sigma_t)[k2] <- paste('Conditional correlation of \n', colnames(data)[i], ' and ', colnames(data)[j])
        k2 <- k2 +1
      }
    }
  }

  for (i in 1:nrow(sigma_t)) {
    tm <- matrix(unlist(sigma_t[i,]), N, N, byrow = T)
    tm2 <- sqrt(solve(diag(diag(tm))))%*%tm%*%sqrt(solve(diag(diag(tm))))
    diag(tm2) <- sqrt(diag(tm))
    sigma_t[i,] <- c(tm2)
  }

  elim <- elimination_mat(N)
  sigma_t <- sigma_t[, which(colSums(elim) == 1)]


  if (inherits(data, "ts")) {
    sigma_t <- ts(sigma_t, start = time(data)[1], frequency = frequency(data))
    H_t <- ts(var_process$sigma_t, start = time(data)[1], frequency = frequency(data))
    e_t <-ts(var_process$e_t, start = time(data)[1], frequency = frequency(data))

   }
  else if(inherits(data, "xts") || inherits(data, "zoo") ){
    sigma_t <- xts(sigma_t, order.by = time(data))
    H_t <-  xts(var_process$sigma_t, order.by = time(data))
    e_t <-  xts(var_process$e_t, order.by = time(data))
  }else{
    H_t <-  var_process$sigma_t
    e_t <- var_process$e_t
  }

  # Final check if BEKK is valid
  BEKK_valid <- valid_bekk(param_mat$c0, param_mat$a, param_mat$g)


  params$likelihood_iter <- params$likelihood_iter[params$likelihood_iter != 0]

  result <- list(C0 =  param_mat$c0,
                 A = param_mat$a,
                 G = param_mat$g,
                 C0_t = tratios_mat$c0,
                 A_t = tratios_mat$a,
                 G_t = tratios_mat$g,
                 C0_sd = sd_mat$c0,
                 A_sd = sd_mat$a,
                 G_sd = sd_mat$g,
                 theta = params$theta,
                 log_likelihood = params$likelihood,
                 BEKK_valid = BEKK_valid,
                 sigma_t = sigma_t,
                 H_t = H_t,
                 e_t = e_t,
                 Second_moments_of_residuals = cov(var_process$e_t),
                 iter = params$iter,
                 likelihood_iter = params$likelihood_iter,
                 asymmetric = FALSE,
                 data = data,
                 spec = spec,
                 QML_t_ratios = QML_t_ratios)
  class(result) <- c('bekkFit', 'bekk')
  result$AIC <- logLik(result)$AIC
  result$BIC <- logLik(result)$BIC
  result$Portmanteau.test <- portmanteau.test(result, lags = 5)

  return(result)
}

#' @export
bekk_fit.bekka <- function(spec, data, QML_t_ratios = FALSE,
                   max_iter = 50, crit = 1e-9) {

  init_values <- spec$init_values
  N <- ncol(data)

  if(is.null(spec$model$signs)){
    spec$model$signs = matrix(rep(-1, N), ncol = 1)
  }
  if(length(spec$model$signs) != N){
    stop('Length of "signs" does not match dimension of data.')
  }

  if(!is.numeric(init_values)) {
    if (is.null(init_values)) {
      theta <- gridSearch_asymmetricBEKK(data, spec$model$signs)
      theta <- theta[[1]]
    } else if (init_values == 'random') {

      cat('Generating starting values \n')
      theta = random_grid_search_asymmetric_BEKK(data, spec$model$signs)[[1]]
    } else if (init_values == 'simple') {
      uncond_var <- crossprod(data)/nrow(data)
      A <- matrix(0, ncol = N, nrow = N)
      B <- matrix(0, ncol = N, nrow = N)
      G <- matrix(0, ncol = N, nrow = N)
      C <- matrix(0, ncol = N, nrow = N)
      #th0=numeric(2*n^2+n*(n+1)/2)

      diag(A) <- 0.25
      diag(B) <- 0.05
      diag(G) <- 0.92
      diag(C) <- 0.05*diag(uncond_var)


      for (i in 1:N){
        for (j in seq(i,N)){

          cij <- uncond_var[i, j]/sqrt(uncond_var[i, i]*uncond_var[j, j])
          C[i,j] <- cij*sqrt(C[i, i]*C[j, j])
          C[j,i] <- C[i, j]

        }
      }

      C = t(chol(C))
      C0 = C[,1]

      if (N > 2) {
        for (i in 2:(N-1)){
          C0 = c(C0, C[i:N, i])
        }
      }

      C0 = c(C0, C[N, N])

      theta = c(C0, c(A), c(G))

    }
  } else {
    if(length(init_values) != 3 * N^2 + N * (N + 1)/2) {
      stop('Number of initial parameter does not match dimension of data.')
    }
    theta <- init_values
  }

  theta <- matrix(theta, ncol =1)

  params <- bhh_asymm_bekk(data, theta, max_iter, crit, spec$model$signs)

  if (QML_t_ratios == TRUE) {
    tratios <- QML_t_ratios_asymm(params$theta, data, spec$model$signs)
    tratios_mat <- coef_mat_asymm(tratios, N)
    sds <- QML_sd_asymm(params$theta, data, spec$model$signs)
    sd_mat <- coef_mat_asymm(sds, N)
  } else {
    tratios_mat <- coef_mat_asymm(params$t_val, N)
    sd_mat <- coef_mat_asymm(params$sd, N)
  }

  param_mat <- coef_mat_asymm(params$theta, N)

  var_process <- sigma_bekk_asymm(data, t(param_mat$c0), param_mat$a, param_mat$b, param_mat$g, spec$model$signs)
  sigma_t <- as.data.frame(var_process$sigma_t)
  colnames(sigma_t) <- rep(1, N^2)

  k <- 1
  k2 <- 1
  for (i in 1:N) {
    for (j in 1:N) {
      if (i == j) {
        colnames(sigma_t)[k2] <- paste('Conditional SD of', colnames(data)[k])
        k <- k + 1
        k2 <- k2 +1
      } else {
        colnames(sigma_t)[k2] <- paste('Conditional correlation of', colnames(data)[i], ' and ', colnames(data)[j])
        k2 <- k2 +1
      }
    }
  }

  for (i in 1:nrow(sigma_t)) {
    tm <- matrix(unlist(sigma_t[i,]), N, N, byrow = T)
    tm2 <- sqrt(solve(diag(diag(tm))))%*%tm%*%sqrt(solve(diag(diag(tm))))
    diag(tm2) <- sqrt(diag(tm))
    sigma_t[i,] <- c(tm2)
  }

  elim <- elimination_mat(N)
  sigma_t <- sigma_t[, which(colSums(elim) == 1)]

  if (inherits(data, "ts")) {
    sigma_t <- ts(sigma_t, start = time(data)[1], frequency = frequency(data))
    H_t <- ts(var_process$sigma_t, start = time(data)[1], frequency = frequency(data))
    e_t <-ts(var_process$e_t, start = time(data)[1], frequency = frequency(data))

  }
  else if(inherits(data, "xts") || inherits(data, "zoo") ){
    sigma_t <- xts(sigma_t, order.by = time(data))
    H_t <-  xts(var_process$sigma_t, order.by = time(data))
    e_t <-  xts(var_process$e_t, order.by = time(data))
  }else{
    H_t <-  var_process$sigma_t
    e_t <- var_process$e_t
  }

  # Final check if BEKK is valid
  BEKK_valid <- valid_asymm_bekk(param_mat$c0, param_mat$a, param_mat$b, param_mat$g, data, spec$model$signs)

  expected_signs=expected_indicator_value(data,spec$model$signs)

  params$likelihood_iter <- params$likelihood_iter[params$likelihood_iter != 0]

  result <- list(C0 =  param_mat$c0,
                 A = param_mat$a,
                 B = param_mat$b,
                 G = param_mat$g,
                 C0_t = tratios_mat$c0,
                 A_t = tratios_mat$a,
                 B_t = tratios_mat$b,
                 G_t = tratios_mat$g,
                 C0_sd = sd_mat$c0,
                 A_sd = sd_mat$a,
                 B_sd = sd_mat$b,
                 G_sd = sd_mat$g,
                 theta = params$theta,
                 signs = spec$model$signs,
                 log_likelihood = params$likelihood,
                 BEKK_valid = BEKK_valid,
                 sigma_t = sigma_t,
                 H_t = H_t,
                 e_t = e_t,
                 Second_moments_of_residuals = cov(var_process$e_t),
                 iter = params$iter,
                 likelihood_iter = params$likelihood_iter,
                 asymmetric = TRUE,
                 expected_signs = expected_signs,
                 data = data,
                 spec = spec,
                 QML_t_ratios = QML_t_ratios)
  class(result) <- c('bekkFit', 'bekka')

  result$AIC <- AIC(result)$AIC
  result$BIC <- BIC(result)$BIC
  result$Portmanteau.test <- portmanteau.test(result, lags = 5)

  return(result)
}

#' @export
bekk_fit.dbekk <- function(spec, data, QML_t_ratios = FALSE,
                          max_iter = 50, crit = 1e-9) {

  init_values <- spec$init_values
  N <- ncol(data)

  if(!is.numeric(init_values)) {
    if (is.null(init_values)) {
      theta <- gridSearch_dBEKK(data)
      theta <- theta[[1]]
    } else if (init_values == 'random') {
      cat('Generating starting values \n')

      theta <- random_grid_search_dBEKK(data)
      theta <- theta[[1]]
    }
  } else {
    if(length(init_values) != 2 * N + N * (N + 1)/2) {
      stop('Number of initial parameter does not match dimension of data.')
    }
    theta <- init_values
  }

  theta <- matrix(theta, ncol =1)

  params <- bhh_dbekk(data, theta, max_iter, crit)

  if (QML_t_ratios == TRUE) {
    tratios <- QML_t_ratios_dbekk(params$theta, data)
    tratios_mat <- coef_mat_diagonal(tratios, N)
    sds <- QML_sd_dbekk(params$theta, data)
    sd_mat <- coef_mat_diagonal(sds, N)
  } else {
    tratios_mat <- coef_mat_diagonal(params$t_val, N)
    sd_mat <- coef_mat_diagonal(params$sd, N)
  }

  param_mat <- coef_mat_diagonal(params$theta, N)


  var_process <- sigma_bekk(data, t(param_mat$c0), param_mat$a, param_mat$g)
  sigma_t <- as.data.frame(var_process$sigma_t)
  colnames(sigma_t) <- rep(1, N^2)

  k <- 1
  k2 <- 1
  for (i in 1:N) {
    for (j in 1:N) {
      if (i == j) {
        colnames(sigma_t)[k2] <- paste('Conditional standard deviation of \n', colnames(data)[k])
        k <- k + 1
        k2 <- k2 +1
      } else {
        colnames(sigma_t)[k2] <- paste('Conditional correlation of \n', colnames(data)[i], ' and ', colnames(data)[j])
        k2 <- k2 +1
      }
    }
  }

  for (i in 1:nrow(sigma_t)) {
    tm <- matrix(unlist(sigma_t[i,]), N, N, byrow = T)
    tm2 <- sqrt(solve(diag(diag(tm))))%*%tm%*%sqrt(solve(diag(diag(tm))))
    diag(tm2) <- sqrt(diag(tm))
    sigma_t[i,] <- c(tm2)
  }

  elim <- elimination_mat(N)
  sigma_t <- sigma_t[, which(colSums(elim) == 1)]


  if (inherits(data, "ts")) {
    sigma_t <- ts(sigma_t, start = time(data)[1], frequency = frequency(data))
    H_t <- ts(var_process$sigma_t, start = time(data)[1], frequency = frequency(data))
    e_t <-ts(var_process$e_t, start = time(data)[1], frequency = frequency(data))

  }
  else if(inherits(data, "xts") || inherits(data, "zoo") ){
    sigma_t <- xts(sigma_t, order.by = time(data))
    H_t <-  xts(var_process$sigma_t, order.by = time(data))
    e_t <-  xts(var_process$e_t, order.by = time(data))
  }else{
    H_t <-  var_process$sigma_t
    e_t <- var_process$e_t
  }

  # Final check if BEKK is valid
  BEKK_valid <- valid_bekk(param_mat$c0, param_mat$a, param_mat$g)


  params$likelihood_iter <- params$likelihood_iter[params$likelihood_iter != 0]

  result <- list(C0 =  param_mat$c0,
                 A = param_mat$a,
                 G = param_mat$g,
                 C0_t = tratios_mat$c0,
                 A_t = tratios_mat$a,
                 G_t = tratios_mat$g,
                 C0_sd = sd_mat$c0,
                 A_sd = sd_mat$a,
                 G_sd = sd_mat$g,
                 theta = params$theta,
                 log_likelihood = params$likelihood,
                 BEKK_valid = BEKK_valid,
                 sigma_t = sigma_t,
                 H_t = H_t,
                 e_t = e_t,
                 Second_moments_of_residuals = cov(var_process$e_t),
                 iter = params$iter,
                 likelihood_iter = params$likelihood_iter,
                 asymmetric = FALSE,
                 data = data,
                 spec = spec,
                 QML_t_ratios = QML_t_ratios)
  class(result) <- c('bekkFit', 'dbekk')

  result$AIC <- AIC(result)$AIC
  result$BIC <- BIC(result)$BIC
  result$Portmanteau.test <- portmanteau.test(result, lags = 5)

  return(result)
}

#' @export
bekk_fit.dbekka <- function(spec, data, QML_t_ratios = FALSE,
                           max_iter = 50, crit = 1e-9) {

  init_values <- spec$init_values
  N <- ncol(data)

  if(is.null(spec$model$signs)){
    spec$model$signs = matrix(rep(-1, N), ncol = 1)
  }
  if(length(spec$model$signs) != N){
    stop('Length of "signs" does not match dimension of data.')
  }

  if(!is.numeric(init_values)) {
    if (is.null(init_values)) {
      theta <- gridSearch_asymmetricdBEKK(data, spec$model$signs)
      theta <- theta[[1]]
    } else if (init_values == 'random') {

      cat('Generating starting values \n')
      theta = random_grid_search_asymmetric_dBEKK(data, spec$model$signs)[[1]]
    }
  } else {
    if(length(init_values) != 3 * N + N * (N + 1)/2) {
      stop('Number of initial parameter does not match dimension of data.')
    }
    theta <- init_values
  }

  theta <- matrix(theta, ncol =1)

  params <- bhh_asymm_dbekk(data, theta, max_iter, crit, spec$model$signs)

  if (QML_t_ratios == TRUE) {
    tratios <- QML_t_ratios_dbekka(params$theta, data, spec$model$signs)
    tratios_mat <- coef_mat_asymm_diagonal(tratios, N)
    sds <- QML_sd_dbekka(params$theta, data, spec$model$signs)
    sd_mat <- coef_mat_asymm_diagonal(sds, N)
  } else {
    tratios_mat <- coef_mat_asymm_diagonal(params$t_val, N)
    sd_mat <- coef_mat_asymm_diagonal(params$sd, N)
  }

  param_mat <- coef_mat_asymm_diagonal(params$theta, N)

  var_process <- sigma_bekk_asymm(data, t(param_mat$c0), param_mat$a, param_mat$b, param_mat$g, spec$model$signs)
  sigma_t <- as.data.frame(var_process$sigma_t)
  colnames(sigma_t) <- rep(1, N^2)

  k <- 1
  k2 <- 1
  for (i in 1:N) {
    for (j in 1:N) {
      if (i == j) {
        colnames(sigma_t)[k2] <- paste('Conditional SD of', colnames(data)[k])
        k <- k + 1
        k2 <- k2 +1
      } else {
        colnames(sigma_t)[k2] <- paste('Conditional correlation of', colnames(data)[i], ' and ', colnames(data)[j])
        k2 <- k2 +1
      }
    }
  }

  for (i in 1:nrow(sigma_t)) {
    tm <- matrix(unlist(sigma_t[i,]), N, N, byrow = T)
    tm2 <- sqrt(solve(diag(diag(tm))))%*%tm%*%sqrt(solve(diag(diag(tm))))
    diag(tm2) <- sqrt(diag(tm))
    sigma_t[i,] <- c(tm2)
  }

  elim <- elimination_mat(N)
  sigma_t <- sigma_t[, which(colSums(elim) == 1)]

  if (inherits(data, "ts")) {
    sigma_t <- ts(sigma_t, start = time(data)[1], frequency = frequency(data))
    H_t <- ts(var_process$sigma_t, start = time(data)[1], frequency = frequency(data))
    e_t <-ts(var_process$e_t, start = time(data)[1], frequency = frequency(data))

  }
  else if(inherits(data, "xts") || inherits(data, "zoo") ){
    sigma_t <- xts(sigma_t, order.by = time(data))
    H_t <-  xts(var_process$sigma_t, order.by = time(data))
    e_t <-  xts(var_process$e_t, order.by = time(data))
  }else{
    H_t <-  var_process$sigma_t
    e_t <- var_process$e_t
  }

  # Final check if BEKK is valid
  BEKK_valid <- valid_asymm_bekk(param_mat$c0, param_mat$a, param_mat$b, param_mat$g, data, spec$model$signs)

  expected_signs=expected_indicator_value(data,spec$model$signs)

  params$likelihood_iter <- params$likelihood_iter[params$likelihood_iter != 0]

  result <- list(C0 =  param_mat$c0,
                 A = param_mat$a,
                 B = param_mat$b,
                 G = param_mat$g,
                 C0_t = tratios_mat$c0,
                 A_t = tratios_mat$a,
                 B_t = tratios_mat$b,
                 G_t = tratios_mat$g,
                 C0_sd = sd_mat$c0,
                 A_sd = sd_mat$a,
                 B_sd = sd_mat$b,
                 G_sd = sd_mat$g,
                 theta = params$theta,
                 signs = spec$model$signs,
                 log_likelihood = params$likelihood,
                 BEKK_valid = BEKK_valid,
                 sigma_t = sigma_t,
                 H_t = H_t,
                 e_t = e_t,
                 Second_moments_of_residuals = cov(var_process$e_t),
                 iter = params$iter,
                 likelihood_iter = params$likelihood_iter,
                 asymmetric = TRUE,
                 expected_signs = expected_signs,
                 data = data,
                 spec = spec,
                 QML_t_ratios = QML_t_ratios)
  class(result) <- c('bekkFit', 'dbekka')

  result$AIC <- AIC(result)$AIC
  result$BIC <- BIC(result)$BIC
  result$Portmanteau.test <- portmanteau.test(result, lags = 5)

  return(result)
}

#' @export
bekk_fit.sbekk <- function(spec, data, QML_t_ratios = FALSE,
                          max_iter = 50, crit = 1e-9) {

  init_values <- spec$init_values
  N <- ncol(data)

  if(!is.numeric(init_values)) {
    if (is.null(init_values)) {
      theta <- gridSearch_sBEKK(data)
      theta <- theta[[1]]
    } else if (init_values == 'random') {
      cat('Generating starting values \n')

      theta <- random_grid_search_sBEKK(data)
      theta <- theta[[1]]
    }
  } else {
    if(length(init_values) != 2  + N * (N + 1)/2) {
      stop('Number of initial parameter does not match dimension of data.')
    }
    theta <- init_values
  }

  theta <- matrix(theta, ncol =1)

  params <- bhh_sbekk(data, theta, max_iter, crit)

  if (QML_t_ratios == TRUE) {
    tratios <- QML_t_ratios_sbekk(params$theta, data)
    tratios_mat <- coef_mat_scalar(tratios, N)
    sds <- QML_sd_sbekk(params$theta, data)
    sd_mat <- coef_mat_scalar(sds, N)
  } else {
    tratios_mat <- coef_mat_scalar(params$t_val, N)
    sd_mat <- coef_mat_scalar(params$sd, N)
  }

  param_mat <- coef_mat_scalar(params$theta, N)


  var_process <- sigma_sbekk(data, t(param_mat$c0), param_mat$a, param_mat$g)
  sigma_t <- as.data.frame(var_process$sigma_t)
  colnames(sigma_t) <- rep(1, N^2)

  k <- 1
  k2 <- 1
  for (i in 1:N) {
    for (j in 1:N) {
      if (i == j) {
        colnames(sigma_t)[k2] <- paste('Conditional standard deviation of \n', colnames(data)[k])
        k <- k + 1
        k2 <- k2 +1
      } else {
        colnames(sigma_t)[k2] <- paste('Conditional correlation of \n', colnames(data)[i], ' and ', colnames(data)[j])
        k2 <- k2 +1
      }
    }
  }

  for (i in 1:nrow(sigma_t)) {
    tm <- matrix(unlist(sigma_t[i,]), N, N, byrow = T)
    tm2 <- sqrt(solve(diag(diag(tm))))%*%tm%*%sqrt(solve(diag(diag(tm))))
    diag(tm2) <- sqrt(diag(tm))
    sigma_t[i,] <- c(tm2)
  }

  elim <- elimination_mat(N)
  sigma_t <- sigma_t[, which(colSums(elim) == 1)]


  if (inherits(data, "ts")) {
    sigma_t <- ts(sigma_t, start = time(data)[1], frequency = frequency(data))
    H_t <- ts(var_process$sigma_t, start = time(data)[1], frequency = frequency(data))
    e_t <-ts(var_process$e_t, start = time(data)[1], frequency = frequency(data))

  }
  else if(inherits(data, "xts") || inherits(data, "zoo") ){
    sigma_t <- xts(sigma_t, order.by = time(data))
    H_t <-  xts(var_process$sigma_t, order.by = time(data))
    e_t <-  xts(var_process$e_t, order.by = time(data))
  }else{
    H_t <-  var_process$sigma_t
    e_t <- var_process$e_t
  }

  # Final check if BEKK is valid
  BEKK_valid <- valid_sbekk(param_mat$c0, param_mat$a, param_mat$g)


  params$likelihood_iter <- params$likelihood_iter[params$likelihood_iter != 0]

  result <- list(C0 =  param_mat$c0,
                 a = param_mat$a,
                 g = param_mat$g,
                 C0_t = tratios_mat$c0,
                 a_t = tratios_mat$a,
                 g_t = tratios_mat$g,
                 C0_sd = sd_mat$c0,
                 a_sd = sd_mat$a,
                 g_sd = sd_mat$g,
                 theta = params$theta,
                 log_likelihood = params$likelihood,
                 BEKK_valid = BEKK_valid,
                 sigma_t = sigma_t,
                 H_t = H_t,
                 e_t = e_t,
                 Second_moments_of_residuals = cov(var_process$e_t),
                 iter = params$iter,
                 likelihood_iter = params$likelihood_iter,
                 asymmetric = FALSE,
                 data = data,
                 spec = spec,
                 QML_t_ratios = QML_t_ratios)
  class(result) <- c('bekkFit', 'sbekk')

  result$AIC <- AIC(result)$AIC
  result$BIC <- BIC(result)$BIC
    result$Portmanteau.test <- portmanteau.test(result, lags = 5)

  return(result)
}

#' @export
bekk_fit.sbekka <- function(spec, data, QML_t_ratios = FALSE,
                           max_iter = 50, crit = 1e-9) {

  init_values <- spec$init_values
  N <- ncol(data)

  if(is.null(spec$model$signs)){
    spec$model$signs = matrix(rep(-1, N), ncol = 1)
  }
  if(length(spec$model$signs) != N){
    stop('Length of "signs" does not match dimension of data.')
  }

  if(!is.numeric(init_values)) {
    if (is.null(init_values)) {
      theta <- gridSearch_asymmetricsBEKK(data, spec$model$signs)
      theta <- theta[[1]]
    } else if (init_values == 'random') {

      cat('Generating starting values \n')
      theta = random_grid_search_asymmetric_sBEKK(data, spec$model$signs)[[1]]
    }
  } else {
    if(length(init_values) != 3 + N * (N + 1)/2) {
      stop('Number of initial parameter does not match dimension of data.')
    }
    theta <- init_values
  }

  theta <- matrix(theta, ncol =1)

  params <- bhh_asymm_sbekk(data, theta, max_iter, crit, spec$model$signs)

  if (QML_t_ratios == TRUE) {
    tratios <- QML_t_ratios_sbekk_asymm(params$theta, data, spec$model$signs)
    tratios_mat <- coef_mat_asymm_scalar(tratios, N)
    sds <- QML_sd_sbekk_asymm(params$theta, data, spec$model$signs)
    sd_mat <- coef_mat_asymm_scalar(sds, N)
  } else {
    tratios_mat <- coef_mat_asymm_scalar(params$t_val, N)
    sd_mat <- coef_mat_asymm_scalar(params$sd, N)
  }

  param_mat <- coef_mat_asymm_scalar(params$theta, N)

  var_process <- sigma_sbekk_asymm(data, t(param_mat$c0), param_mat$a, param_mat$b, param_mat$g, spec$model$signs)
  sigma_t <- as.data.frame(var_process$sigma_t)
  colnames(sigma_t) <- rep(1, N^2)

  k <- 1
  k2 <- 1
  for (i in 1:N) {
    for (j in 1:N) {
      if (i == j) {
        colnames(sigma_t)[k2] <- paste('Conditional SD of', colnames(data)[k])
        k <- k + 1
        k2 <- k2 +1
      } else {
        colnames(sigma_t)[k2] <- paste('Conditional correlation of', colnames(data)[i], ' and ', colnames(data)[j])
        k2 <- k2 +1
      }
    }
  }

  for (i in 1:nrow(sigma_t)) {
    tm <- matrix(unlist(sigma_t[i,]), N, N, byrow = T)
    tm2 <- sqrt(solve(diag(diag(tm))))%*%tm%*%sqrt(solve(diag(diag(tm))))
    diag(tm2) <- sqrt(diag(tm))
    sigma_t[i,] <- c(tm2)
  }

  elim <- elimination_mat(N)
  sigma_t <- sigma_t[, which(colSums(elim) == 1)]
  if (inherits(data, "ts")) {
    sigma_t <- ts(sigma_t, start = time(data)[1], frequency = frequency(data))
    H_t <- ts(var_process$sigma_t, start = time(data)[1], frequency = frequency(data))
    e_t <-ts(var_process$e_t, start = time(data)[1], frequency = frequency(data))

  }
  else if(inherits(data, "xts") || inherits(data, "zoo") ){
    sigma_t <- xts(sigma_t, order.by = time(data))
    H_t <-  xts(var_process$sigma_t, order.by = time(data))
    e_t <-  xts(var_process$e_t, order.by = time(data))
  }else{
    H_t <-  var_process$sigma_t
    e_t <- var_process$e_t
  }

  # Final check if BEKK is valid
  BEKK_valid <- valid_asymm_sbekk(param_mat$c0, param_mat$a, param_mat$b, param_mat$g, data, spec$model$signs)

  expected_signs=expected_indicator_value(data,spec$model$signs)

  params$likelihood_iter <- params$likelihood_iter[params$likelihood_iter != 0]

  result <- list(C0 =  param_mat$c0,
                 a = param_mat$a,
                 b = param_mat$b,
                 g = param_mat$g,
                 C0_t = tratios_mat$c0,
                 a_t = tratios_mat$a,
                 b_t = tratios_mat$b,
                 g_t = tratios_mat$g,
                 C0_sd = sd_mat$c0,
                 a_sd = sd_mat$a,
                 b_sd = sd_mat$b,
                 g_sd = sd_mat$g,
                 theta = params$theta,
                 signs = spec$model$signs,
                 log_likelihood = params$likelihood,
                 BEKK_valid = BEKK_valid,
                 sigma_t = sigma_t,
                 H_t = H_t,
                 e_t = e_t,
                 Second_moments_of_residuals = cov(var_process$e_t),
                 iter = params$iter,
                 likelihood_iter = params$likelihood_iter,
                 asymmetric = TRUE,
                 expected_signs = expected_signs,
                 data = data,
                 spec = spec,
                 QML_t_ratios = QML_t_ratios)
  class(result) <- c('bekkFit', 'sbekka')

  result$AIC <- AIC(result)$AIC
  result$BIC <- BIC(result)$BIC
  result$Portmanteau.test <- portmanteau.test(result, lags = 5)

  return(result)
}

test_that("virf works with 2-dimensional test data set", {
  spec = bekk_spec()
  x = bekk_fit(spec, StocksBonds)
  s = virf(x)
  expect_equal(round(sum(s$VIRF),3), -0.727)
  expect_equal(round(sum(s$VIRF_upper),3), -0.661)
  expect_equal(round(sum(s$VIRF_lower),3), -0.794)
})
# test_that("virf works with 3-dimensional test data set", {
#   spec = bekk_spec()
#   x = bekk_fit(spec, GoldStocksBonds)
#   s = virf(x, n.ahead = 50)
#   expect_equal(round(sum(s$VIRF_upper),3), 0)
#   expect_equal(round(sum(s$VIRF_lower),3), 0)
#   expect_equal(round(sum(s$VIRF),3), 0)
# })

test_that("Simulating 2 dim symmetric BEKK based on estimated ts object", {
  obj_spec <- bekk_spec()
  x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)

  x2 <- simulate(x1, nsim = 100)
  expect_equal(nrow(x2), 100)
  expect_equal(ncol(x2), 2)
})

test_that("Simulating 2 dim symmetric BEKK based on bekk_spec", {
  obj_spec <- bekk_spec(model = list(type = "bekk", asymmetric = FALSE), N = 2,
                        init_values = as.matrix(c(0.019783677, -0.014962718, 0.085265736, 0.178859630, -0.007691659,
                                         -0.037152006, 0.298147337, 0.981184605, 0.001337367, 0.013260377, 0.951431611)))

  x2 <- simulate(obj_spec, nsim = 100)
  expect_equal(nrow(x2), 100)
  expect_equal(ncol(x2), 2)
})

test_that("Simulating 3 dim symmetric BEKK based on estimated xts object", {
  obj_spec <- bekk_spec()
  x1 <- bekk_fit(obj_spec, GoldStocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)

  x2 <- simulate(x1, nsim = 100)
  expect_equal(nrow(x2), 100)
  expect_equal(ncol(x2), 3)
})

test_that("Simulating 3 dim symmetric BEKK based on bekk_spec", {
  obj_spec <- bekk_spec(model = list(type = "bekk", asymmetric = FALSE), N = 3,
                        init_values = as.matrix(c(0.0010122220, -0.0002678085, 0.0001905036, 0.0008937830, 0.0003111280,
                                        0.0003893194, 0.2453276126, 0.0222127631, -0.0434456548, -0.0366629406,
                                        0.2637613982, -0.0248105673, 0.0624407934, -0.0016446789, 0.1733957064,
                                        0.9651528236, -0.0060973318, 0.0155928988, 0.0125157899, 0.9606411455,
                                        0.0008724460, -0.0196494968, -0.0009070321, 0.9801993180)))

  x2 <- simulate(obj_spec, nsim = 100)
  expect_equal(nrow(x2), 100)
  expect_equal(ncol(x2), 3)
})

test_that("Simulating 2 dim asymmetric BEKK based on estimated ts object", {
  obj_spec <- bekk_spec(model = list(type = 'bekk', asymmetric = TRUE))
  x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)

  x2 <- simulate(x1, nsim = 100)
  expect_equal(nrow(x2), 100)
  expect_equal(ncol(x2), 2)
})

test_that("Simulating 2 dim asymmetric BEKK based on bekk_spec", {
  obj_spec <- bekk_spec(model = list(type = "bekk", asymmetric = TRUE), N = 2,
                        init_values = as.matrix(c(0.019783677, -0.014962718, 0.085265736, 0.178859630, -0.007691659,
                                        -0.037152006, 0.298147337, 0.05234,-0.02,0.032,0.1032, 0.981184605, 0.001337367, 0.013260377, 0.951431611)))

  x2 <- simulate(obj_spec, nsim = 100)
  expect_equal(nrow(x2), 100)
  expect_equal(ncol(x2), 2)
})

test_that("Symmetric BEKK 2-dims works, ts object and QML_t_ratios = FALSE, n.ahead = 1", {
  obj_spec <- bekk_spec()
  x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)

  x2 <- predict(x1, n.ahead = 1)

  expect_equal(round(as.numeric(x2$volatility_forecast[1]), 2), 0.19)
  expect_equal(round(as.numeric(x2$volatility_forecast[2]), 2), -0.1)
  expect_equal(round(as.numeric(x2$volatility_forecast[3]), 2), 0.61)

  expect_equal(round(sum(x2$H_t_forecast), 2), 0.38)
  expect_equal(round(sum(x2$volatility_lower_conf_band), 2), 0.68)
  expect_equal(round(sum(x2$volatility_upper_conf_band), 2), 0.71)
})

test_that("Symmetric BEKK 2-dims works, ts object and QML_t_ratios = FALSE, n.ahead = 5", {
  obj_spec <- bekk_spec()
  x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)

  x2 <- predict(x1, n.ahead = 5)

  expect_equal(round(sum(x2$volatility_forecast[,1]), 2), 0.94)
  expect_equal(round(sum(x2$volatility_forecast[,2]), 2), -0.51)
  expect_equal(round(sum(x2$volatility_forecast[,3]), 2), 3.08)

  expect_equal(round(sum(x2$H_t_forecast), 2), 1.96)
  expect_equal(round(sum(x2$volatility_lower_conf_band), 2), 3.42)
  expect_equal(round(sum(x2$volatility_upper_conf_band), 2), 3.62)
})

#
# test_that("Symmetric BEKK 3-dims works, xts object and QML_t_ratios = FALSE, n.ahead = 1", {
#   obj_spec <- bekk_spec()
#   x1 <- bekk_fit(obj_spec, GoldStocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)
#
#   x2 <- predict(x1, n.ahead = 1)
#
#   expect_equal(round(as.numeric(x2$volatility_forecast[,1]), 2), 0.01)
#   expect_equal(round(as.numeric(x2$volatility_forecast[,2]), 2), 0.12)
#   expect_equal(round(as.numeric(x2$volatility_forecast[,3]), 2), 0.22)
#   expect_equal(round(as.numeric(x2$volatility_forecast[,4]), 2), 0.01)
#   expect_equal(round(as.numeric(x2$volatility_forecast[,5]), 2), -0.16)
#
#   expect_equal(round(sum(x2$H_t_forecast), 2), 0)
#   expect_equal(round(sum(x2$volatility_lower_conf_band), 2), 0.19)
#   expect_equal(round(sum(x2$volatility_upper_conf_band), 2), 0.23)
# })
#
# test_that("Symmetric BEKK 3-dims works, xts object and QML_t_ratios = FALSE, n.ahead = 5", {
#   obj_spec <- bekk_spec()
#   x1 <- bekk_fit(obj_spec, GoldStocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)
#
#   x2 <- predict(x1, n.ahead = 5)
#
#   expect_equal(round(sum(as.numeric(x2$volatility_forecast[,1])), 2), 0.05)
#   expect_equal(round(sum(as.numeric(x2$volatility_forecast[,2])), 2), 0.58)
#   expect_equal(round(sum(as.numeric(x2$volatility_forecast[,3])), 2), 1.09)
#   expect_equal(round(sum(as.numeric(x2$volatility_forecast[,4])), 2), 0.04)
#   expect_equal(round(sum(as.numeric(x2$volatility_forecast[,5])), 2), -0.8)
#
#   expect_equal(round(sum(x2$H_t_forecast), 2), 0)
#   expect_equal(round(sum(x2$volatility_lower_conf_band), 2), 0.85)
#   expect_equal(round(sum(x2$volatility_upper_conf_band), 2), 1.15)
# })


test_that("Asymmetric BEKK 2-dims works, ts object and QML_t_ratios = FALSE, n.ahead = 1", {
  obj_spec <- bekk_spec(model = list(type = "bekk", asymmetric = TRUE))
  x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)

  x2 <- predict(x1, n.ahead = 1)

  expect_equal(round(as.numeric(x2$volatility_forecast[1]), 2), 0.19)
  expect_equal(round(as.numeric(x2$volatility_forecast[2]), 2), -0.13)
  expect_equal(round(as.numeric(x2$volatility_forecast[3]), 2), 0.59)

  expect_equal(round(sum(x2$H_t_forecast), 2), 0.36)
  expect_equal(round(sum(x2$volatility_lower_conf_band), 2), 0.61)
  expect_equal(round(sum(x2$volatility_upper_conf_band), 2), 0.67)
})

test_that("Symmetric BEKK 2-dims works, ts object and QML_t_ratios = FALSE, n.ahead = 5", {
  obj_spec <- bekk_spec(model = list(type = "bekk", asymmetric = TRUE))
  x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)

  x2 <- predict(x1, n.ahead = 5)

  expect_equal(round(sum(as.numeric(x2$volatility_forecast[,1])), 2), 0.93)
  expect_equal(round(sum(as.numeric(x2$volatility_forecast[,2])), 2), -0.65)
  expect_equal(round(sum(as.numeric(x2$volatility_forecast[,3])), 2), 2.88)

  expect_equal(round(sum(x2$H_t_forecast), 2), 1.69)
  expect_equal(round(sum(x2$volatility_lower_conf_band), 2), 3.04)
  expect_equal(round(sum(x2$volatility_upper_conf_band), 2), 3.33)
})

#
# test_that("Symmetric BEKK 3-dims works, xts object and QML_t_ratios = FALSE, n.ahead = 1", {
#   obj_spec <- bekk_spec(model = list(type = "bekk", asymmetric = TRUE))
#   x1 <- bekk_fit(obj_spec, GoldStocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)
#
#   x2 <- predict(x1, n.ahead = 1)
#
#   expect_equal(round(as.numeric(x2$volatility_forecast[,1]), 2), 0.01)
#   expect_equal(round(as.numeric(x2$volatility_forecast[,2]), 2), 0.01)
#   expect_equal(round(as.numeric(x2$volatility_forecast[,3]), 2), 0.35)
#   expect_equal(round(as.numeric(x2$volatility_forecast[,4]), 2), 0.01)
#   expect_equal(round(as.numeric(x2$volatility_forecast[,5]), 2), -0.11)
#
#   expect_equal(round(sum(x2$H_t_forecast), 2), 0)
#   expect_equal(round(sum(x2$volatility_lower_conf_band), 2), 0.25)
#   expect_equal(round(sum(x2$volatility_upper_conf_band), 2), 0.31)
# })
#
# test_that("Symmetric BEKK 3-dims works, xts object and QML_t_ratios = FALSE, n.ahead = 5", {
#   obj_spec <- bekk_spec(model = list(type = "bekk", asymmetric = TRUE))
#   x1 <- bekk_fit(obj_spec, GoldStocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)
#
#   x2 <- predict(x1, n.ahead = 5)
#
#   expect_equal(round(sum(as.numeric(x2$volatility_forecast[,1])), 2), 0.05)
#   expect_equal(round(sum(as.numeric(x2$volatility_forecast[,2])), 2), 0.09)
#   expect_equal(round(sum(as.numeric(x2$volatility_forecast[,3])), 2), 1.64)
#   expect_equal(round(sum(as.numeric(x2$volatility_forecast[,4])), 2), 0.05)
#   expect_equal(round(sum(as.numeric(x2$volatility_forecast[,5])), 2), -0.63)
#
#   expect_equal(round(sum(x2$H_t_forecast), 2), 0)
#   expect_equal(round(sum(x2$volatility_lower_conf_band), 2), 1.04)
#   expect_equal(round(sum(x2$volatility_upper_conf_band), 2), 1.45)
# })



test_that("Symmetric scalar BEKK 2-dims works, ts object and QML_t_ratios = FALSE, n.ahead = 1", {
  obj_spec <- bekk_spec(model = list(type="sbekk", asymmetric =F))
  x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)

  x2 <- predict(x1, n.ahead = 10)

  expect_equal(round(as.numeric(x2$volatility_forecast[1,1]), 2), 0.19)
  expect_equal(round(as.numeric(x2$volatility_forecast[2,2]), 2), -0.09)
  expect_equal(round(as.numeric(x2$volatility_forecast[2,3]), 2), 0.64)

  expect_equal(round(sum(x2$H_t_forecast), 2), 4.31)
  expect_equal(round(sum(x2$volatility_lower_conf_band), 2), 7.48)
  expect_equal(round(sum(x2$volatility_upper_conf_band), 2), 7.6)
})

# test_that("Symmetric BEKK 2-dims works, ts object and QML_t_ratios = FALSE", {
#   obj_spec <- bekk_spec()
#   x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)
#
#   expect_equal(round(x1$log_likelihood), -7382)
#   expect_equal(round(sum(x1$theta)),  2)
#   expect_equal(round(sum(c(x1$C0_t, x1$A_t, x1$G_t))), 1208)
# })
#
# test_that("Symmetric BEKK 2-dims works, ts object and QML_t_ratios = TRUE", {
#   obj_spec <- bekk_spec()
#   x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = TRUE, max_iter = 50, crit = 1e-9)
#
#   expect_equal(round(x1$log_likelihood), -7382)
#   expect_equal(round(sum(x1$theta)),  2)
#   expect_lt(sum(c(x1$C0_t, x1$A_t, x1$G_t)), 1208)
# })
#
# test_that("Symmetric BEKK 3-dims works, xts object and QML_t_ratios = FALSE", {
#   obj_spec <- bekk_spec()
#   x1 <- bekk_fit(obj_spec, GoldStocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)
#
#   expect_equal(round(x1$log_likelihood), 75249)
#   expect_equal(round(sum(x1$theta)),  4)
#   expect_equal(round(sum(c(x1$C0_t, x1$A_t, x1$G_t))), 2384)
# })
#
# test_that("Symmetric BEKK 3-dims works, xts object and QML_t_ratios = TRUE", {
#   obj_spec <- bekk_spec()
#   x1 <- bekk_fit(obj_spec, GoldStocksBonds, QML_t_ratios = TRUE, max_iter = 50, crit = 1e-9)
#
#   expect_equal(round(x1$log_likelihood), 75249)
#   expect_equal(round(sum(x1$theta)),  4)
#   expect_lt(sum(c(x1$C0_t, x1$A_t, x1$G_t)), 2384)
# })
#
# test_that("Asymmetric BEKK 2-dims works, ts object and QML_t_ratios = FALSE", {
#   obj_spec <- bekk_spec(model = list(type = "bekk", asymmetric = TRUE))
#   x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)
#
#   expect_equal(round(x1$log_likelihood), -7340)
#   expect_equal(round(sum(x1$theta)),  3)
#   expect_equal(round(sum(c(x1$C0_t, x1$A_t, x1$G_t))), 1086)
# })
#
# test_that("Asymmetric BEKK 2-dims works, ts object and QML_t_ratios = TRUE", {
#   obj_spec <- bekk_spec(model = list(type = "bekk", asymmetric = TRUE))
#   x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = TRUE, max_iter = 50, crit = 1e-9)
#
#   expect_equal(round(x1$log_likelihood), -7340)
#   expect_equal(round(sum(x1$theta)),  3)
#   expect_lt(sum(c(x1$C0_t, x1$A_t, x1$G_t)), 1086)
# })
#
#
# test_that("Asymmetric BEKK 3-dims works, xts object and QML_t_ratios = TRUE", {
#   obj_spec <- bekk_spec(model = list(type = "bekk", asymmetric = TRUE))
#   x1 <- bekk_fit(obj_spec, GoldStocksBonds, QML_t_ratios = TRUE, max_iter = 50, crit = 1e-9)
#
#   expect_equal(round(x1$log_likelihood), 75327)
#   expect_equal(round(sum(x1$theta)),  4)
#   expect_lt(sum(c(x1$C0_t, x1$A_t, x1$G_t)), 1797)
# })
test_that("Symmetric dBEKK 2-dims works, ts object and QML_t_ratios = FALSE", {
  obj_spec <- bekk_spec(model = list(type = "dbekk", asymmetric = F))
  x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)

  expect_equal(round(x1$log_likelihood), -7388)
  expect_equal(round(sum(x1$theta)),  3)
  expect_equal(round(sum(c(x1$C0_t, x1$A_t, x1$G_t))), 1231)
})
test_that("Asymmetric dBEKK 3-dims works, xts object and QML_t_ratios = TRUE", {
  obj_spec <- bekk_spec(model = list(type = "dbekk", asymmetric =T))
  x1 <- bekk_fit(obj_spec, GoldStocksBonds, QML_t_ratios = TRUE, max_iter = 50, crit = 1e-9)

  expect_equal(round(x1$log_likelihood), 75068)
  expect_equal(round(sum(x1$theta)),  4)
  expect_lt(sum(c(x1$C0_t, x1$A_t, x1$G_t)), 100)
})

test_that("Symmetric sBEKK 2-dims works, ts object and QML_t_ratios = FALSE", {
  obj_spec <- bekk_spec(model = list(type = "sbekk", asymmetric = F))
  x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)

  expect_equal(round(x1$log_likelihood), -7417)
  expect_equal(round(sum(x1$theta)),  1)
  expect_equal(round(sum(c(x1$C0_t, x1$A_t, x1$G_t))), 35)
})
test_that("Symmetric sBEKK 3-dims works, xts object and QML_t_ratios = TRUE", {
  obj_spec <- bekk_spec(model = list(type = "sbekk", asymmetric = F))
  x1 <- bekk_fit(obj_spec, GoldStocksBonds, QML_t_ratios = TRUE, max_iter = 50, crit = 1e-9)

  expect_equal(round(x1$log_likelihood), 74997)
  expect_equal(round(sum(x1$theta)),  1)
  expect_lt(sum(c(x1$C0_t, x1$A_t, x1$G_t)), 399.75)
})

gridSearch_BEKK <- function(r) {
  N  <- ncol(r)

  uncond_var <- crossprod(r)/nrow(r)
  A <- matrix(0, ncol = N, nrow = N)
  G <- matrix(0, ncol = N, nrow = N)
  C <- matrix(0, ncol = N, nrow = N)

  diag(A) <- 0.3
  diag(G) <- 0.92
  diag(C) <- 0.05*diag(uncond_var)


  for (i in 1:N){
    for (j in seq(i,N)){

      cij <- uncond_var[i, j]/sqrt(uncond_var[i, i]*uncond_var[j, j])
      C[i,j] <- cij*sqrt(C[i, i]*C[j, j])
      C[j,i] <- C[i, j]

    }
  }

  C = t(chol(C))
  C0 = C[,1]

  if (N > 2) {
    for (i in 2:(N-1)){
      C0 = c(C0, C[i:N, i])
    }
  }

  C0 = c(C0, C[N, N])

  for (i in 1:N) {
    for (j in 1:N) {
      if (i < j) {
        A[i, j] <-  0.03
        if (j == N & i == 1) {
          A[i, j] <-  -0.03
        }
        G[i, j] <- -0.03
      } else if (i > j) {
        A[i, j] <- 0.03
        G[i, j] <- 0.03
      }
    }
  }

  th0 = c(C0, c(A), c(G))
  lik = loglike_bekk(th0, r)

  return(list(th0, lik))

}

gridSearch_asymmetricBEKK <- function(r, signs) {
  N  <- ncol(r)

  uncond_var <- crossprod(r)/nrow(r)
  A <- matrix(0, ncol = N, nrow = N)
  B <- matrix(0, ncol = N, nrow = N)
  G <- matrix(0, ncol = N, nrow = N)
  C <- matrix(0, ncol = N, nrow = N)
  #th0=numeric(2*n^2+n*(n+1)/2)

  diag(A) <- 0.3
  diag(B) <- 0.3
  diag(G) <- 0.92
  diag(C) <- 0.05*diag(uncond_var)


  for (i in 1:N){
    for (j in seq(i,N)){

      cij <- uncond_var[i, j]/sqrt(uncond_var[i, i]*uncond_var[j, j])
      C[i,j] <- cij*sqrt(C[i, i]*C[j, j])
      C[j,i] <- C[i, j]

    }
  }

  C = t(chol(C))
  C0 = C[,1]

  if (N > 2) {
    for (i in 2:(N-1)){
      C0 = c(C0, C[i:N, i])
    }
  }

  C0 = c(C0, C[N, N])

  for (i in 1:N) {
    for (j in 1:N) {
      if (i < j) {
        A[i, j] <-  0.03
        B[i, j] <-  0.03
        if (j == N & i == 1) {
          A[i, j] <-  -0.03
          B[i, j] <-  -0.03
        }
        G[i, j] <- -0.03
      } else if (i > j) {
        A[i, j] <- 0.03
        B[i, j] <- 0.03
        G[i, j] <- 0.03
      }
    }
  }

  th0 = c(C0, c(A), c(B), c(G))
  lik = loglike_asymm_bekk(th0, r, signs)

  return(list(th0, lik))

}

gridSearch_dBEKK <- function(r) {
  N  <- ncol(r)

  uncond_var <- crossprod(r)/nrow(r)
  A <- matrix(0, ncol = N, nrow = N)
  G <- matrix(0, ncol = N, nrow = N)
  C <- matrix(0, ncol = N, nrow = N)

  diag(A) <- 0.3
  diag(G) <- 0.92
  diag(C) <- 0.05*diag(uncond_var)


  for (i in 1:N){
    for (j in seq(i,N)){

      cij <- uncond_var[i, j]/sqrt(uncond_var[i, i]*uncond_var[j, j])
      C[i,j] <- cij*sqrt(C[i, i]*C[j, j])
      C[j,i] <- C[i, j]

    }
  }

  C = t(chol(C))
  C0 = C[,1]

  if (N > 2) {
    for (i in 2:(N-1)){
      C0 = c(C0, C[i:N, i])
    }
  }

  C0 = c(C0, C[N, N])

  for (i in 1:N) {
    for (j in 1:N) {
      if (i < j) {
        A[i, j] <-  0.03
        if (j == N & i == 1) {
          A[i, j] <-  -0.03
        }
        G[i, j] <- -0.03
      } else if (i > j) {
        A[i, j] <- 0.03
        G[i, j] <- 0.03
      }
    }
  }

  th0 = c(C0, diag(A), diag(G))
  lik = loglike_dbekk(th0, r)

  return(list(th0, lik))

}

gridSearch_asymmetricdBEKK <- function(r, signs) {
  N  <- ncol(r)

  uncond_var <- crossprod(r)/nrow(r)
  A <- matrix(0, ncol = N, nrow = N)
  B <- matrix(0, ncol = N, nrow = N)
  G <- matrix(0, ncol = N, nrow = N)
  C <- matrix(0, ncol = N, nrow = N)
  #th0=numeric(2*n^2+n*(n+1)/2)

  diag(A) <- 0.3
  diag(B) <- 0.3
  diag(G) <- 0.92
  diag(C) <- 0.05*diag(uncond_var)


  for (i in 1:N){
    for (j in seq(i,N)){

      cij <- uncond_var[i, j]/sqrt(uncond_var[i, i]*uncond_var[j, j])
      C[i,j] <- cij*sqrt(C[i, i]*C[j, j])
      C[j,i] <- C[i, j]

    }
  }

  C = t(chol(C))
  C0 = C[,1]

  if (N > 2) {
    for (i in 2:(N-1)){
      C0 = c(C0, C[i:N, i])
    }
  }

  C0 = c(C0, C[N, N])


  th0 = c(C0, diag(A), diag(B), diag(G))
  lik = loglike_asymm_dbekk(th0, r, signs)

  return(list(th0, lik))

}

gridSearch_sBEKK <- function(r) {
  N  <- ncol(r)
  C <- matrix(0, ncol = N, nrow = N)
  uncond_var <- crossprod(r)/nrow(r)


  a <- 0.2
  g <- 0.7
  diag(C) <- 0.05*diag(uncond_var)


  for (i in 1:N){
    for (j in seq(i,N)){

      cij <- uncond_var[i, j]/sqrt(uncond_var[i, i]*uncond_var[j, j])
      C[i,j] <- cij*sqrt(C[i, i]*C[j, j])
      C[j,i] <- C[i, j]

    }
  }

  C = t(chol(C))
  C0 = C[,1]

  if (N > 2) {
    for (i in 2:(N-1)){
      C0 = c(C0, C[i:N, i])
    }
  }

  C0 = c(C0, C[N, N])



  th0 = c(C0, a, g)
  lik = loglike_sbekk(th0, r)

  return(list(th0, lik))

}

gridSearch_asymmetricsBEKK <- function(r, signs) {
  N  <- ncol(r)
  C <- matrix(0, ncol = N, nrow = N)
  uncond_var <- crossprod(r)/nrow(r)

  #th0=numeric(2*n^2+n*(n+1)/2)

  a <- 0.2
  b <- 0.1
  g <- 0.6
  diag(C) <- 0.05*diag(uncond_var)


  for (i in 1:N){
    for (j in seq(i,N)){

      cij <- uncond_var[i, j]/sqrt(uncond_var[i, i]*uncond_var[j, j])
      C[i,j] <- cij*sqrt(C[i, i]*C[j, j])
      C[j,i] <- C[i, j]

    }
  }

  C = t(chol(C))
  C0 = C[,1]

  if (N > 2) {
    for (i in 2:(N-1)){
      C0 = c(C0, C[i:N, i])
    }
  }

  C0 = c(C0, C[N, N])



  th0 = c(C0, a, b, g)
  lik = loglike_asymm_sbekk(th0, r, signs)

  return(list(th0, lik))

}
#H2 Code
# gridSearch_BEKK <- function(r){
#   N  <- ncol(r)
#
#   uncond_var <- crossprod(r)/nrow(r)
#   A <- matrix(0, ncol = N, nrow = N)
#   G <- matrix(0, ncol = N, nrow = N)
#   C <- matrix(0, ncol = N, nrow = N)
#   #th0=numeric(2*n^2+n*(n+1)/2)
#
#   diag(A) <- 0.3
#   diag(G) <- 0.92
#   diag(C) <- 0.05*diag(uncond_var)
#
#
#   for (i in 1:N){
#     for (j in seq(i,N)){
#
#       cij <- uncond_var[i, j]/sqrt(uncond_var[i, i]*uncond_var[j, j])
#       C[i,j] <- cij*sqrt(C[i, i]*C[j, j])
#       C[j,i] <- C[i, j]
#
#     }
#   }
#
#   C = t(chol(C))
#   C0 = C[,1]
#
#   if (N > 2) {
#     for (i in 2:(N-1)){
#       C0 = c(C0, C[i:N, i])
#     }
#   }
#
#   C0 = c(C0, C[N, N])
#
#   # deterministic variance components
#
#   th0 = c(C0, c(A), c(G))
#
#   # change elements of A and G and compute likelihood in each step
#
#   likmax = -1e25
#
#   #print th0
#   result= recursiveSearch_BEKK(r, C0, c(A), c(G), 1, th0, likmax)
#   th0=result[[1]]
#   likmax=result[[2]]
#   return(list(th0,likmax))
#
# }
#
#
# recursiveSearch_BEKK=function(r, c0, avec, gvec, index, thetaopt, likmax){
#
#   n=ncol(r)
#   start= -3
#   endr = 3
#   step = 6
#   indextest=0
#
#   if (index == n^2){
#     index = index + 2;
#   } else if (index < n^2){
#     indextest = (index-1)/(n+1)
#   } else{
#     indextest = (index-n^2-1)/(n+1)
#   }
#   # we have a diagonal element
#   if (indextest - floor(indextest) == 0){
#     index = index + 1
#   }
#
#   for (i in seq(start, endr, step)){
#     val = i/100
#
#     #set a and g respectively according to index, exclude diagonal elements
#     if (index <= n^2){
#       avec[index] = val
#     } else{
#       gvec[index-n^2] = val
#     }
#     #last element is excluded
#     if (index < (2*n^2-1)){
#       # recursive step
#       result= recursiveSearch_BEKK(r, c0, avec, gvec, index+1, thetaopt, likmax)
#       thetaopt=result[[1]]
#       likmax=result[[2]]
#     } else{
#       #final step
#       theta = c(c0,avec,gvec)
#       #likelihood
#       lik = loglike_bekk(theta,r)
#       if (lik > likmax){
#         thetaopt = theta
#         likmax = lik
#       }
#
#     }
#   }
#
#   return(list(thetaopt,likmax))
# }

#' BEKKs: Volatility modelling
#'
#' @name BEKKs
#' @author \itemize{
#' \item Markus J. Fülle  \email{fuelle@uni-goettingen.de}
#' \item Helmut Herwartz \email{hherwartz@uni-goettingen.de}
#' \item Alexander Lange \email{alexander.lange@uni-goettingen.de}
#' \item Christian M. Hafner \email{christian.hafner@uclouvain.be}
#' }
#'
#' @import mathjaxr
#' @description
#' \loadmathjax
#' This package implements estimation, simulation and forecasting techniques for conditional volatility modelling using the BEKK model. We refer the reader to Fülle et al. (2024) for a package overview.
#'  The full BEKK(1,1,1) model of Engle and Kroner (1995)
#' \mjdeqn{H_t = CC' + A' r_{t-1} r_{t-1}'A + G' H_{t-1}G ,}{H_t = CC' + A' r_{t-1} r_{t-1}'A + G' H_{t-1}G ,}  the asymmetric extensions of Kroner and Ng (1998) and Grier et. al. (2004)
#' \mjdeqn{H_t = CC' + A' r_{t-1} r_{t-1}'A +B'\gamma_{t-1} \gamma_{t-1}' B+G'H_{t-1}G}{H_t = CC' +A'r_{t-1} r_{t-1}'A +B'\gamma_{t-1} \gamma_{t-1}' B+G'H_{t-1}G,,}
#' with \mjdeqn{\gamma_t = r_t I\left(r_t lower than 0 \right)}{\gamma_t = r_t I(r_t &lt; 0 )} are implemented.
#' Moreover, the diagonal BEKK, where the parameter matrices A, B and G are reduced to diagonal matrices and
#' the scalar BEKK model of Ding and Engle (2001)
#' \mjdeqn{H_t = CC' + a r_{t-1} r_{t-1}' + g H_{t-1},}{H_t = CC' + a r_{t-1} r_{t-1}' + g H_{t-1},}
#' where a and g are scalar parameters and are implemented to allow faster but less flexible estimation in higher dimensions.
#' @details
#' The main functions are:
#' \itemize{
#' \item \tabular{ll}{ \code{\link{bekk_spec}} \tab Specifies the model type to be estimated.}
#' \item \tabular{ll}{\code{\link{bekk_fit}} \tab Estimates a BEKK(1,1,1) model of a given series and
#'    specification object \link{bekk_spec}.}
#' \item \tabular{ll}{ \code{\link{simulate}} \tab Simulates a BEKK(1,1,1) process using either a \link{bekk_fit} or \link{bekk_spec} object.}
#' \item \tabular{ll}{ \code{\link{predict}} \tab Forecasts  conditional volatility using a \link{bekk_fit} object.}
#' \item \tabular{ll}{ \code{\link{VaR}} \tab Estimates (portfolio) Value-at-Risk using a fitted BEKK(1,1,1) model.}
#' \item \tabular{ll}{ \code{\link{backtest}} \tab Backtesting estimated (portfolio) value-at-risks of a fitted BEKK(1,1,1) model.}
#' \item \tabular{ll}{ \code{\link{virf}} \tab Calculates volatility impulse response functions for fitted symmetric BEKK(1,1,1) models.}
#' }
#' @references Engle, R. F. and K. F. Kroner (1995). Multivariate simultaneous generalized arch. Econometric Theory 11(1),122-150.
#' @references Fülle, M. J., A. Lange, C. M. Hafner, and H. Herwartz (2024). BEKKs: An R package for estimation of
#' conditional volatility of multivariate time series. Journal of Statistical Software 111 (4), 1–34. <doi:10.18637/jss.v111.i04>.
#' @references Kroner, K. F. and V. K. Ng (1998). Modeling asymmetric comovements of asset returns. Review of Financial Studies 11(4), 817-44.
#' @references Ding, Zhuanxin and Engle, Robert F (2001). Large scale conditional covariance matrix modeling, estimation and testing. NYU working paper No. Fin-01-029.
#' @references Grier, K. B., Olan T. Henry, N. Olekalns, and K. Shields (2004). The asymmetric effects of uncertainty on inflation and output growth. Journal of Applied Econometrics  19(5), 551-565.
#' @references Hafner CM, Herwartz H (2006). Volatility impulse responses for multivariate GARCH models:  An exchange rate illustration. Journal of International Money and Finance,25,719-740.
#' @useDynLib BEKKs
#' @importFrom Rcpp sourceCpp
#' @md
NULL

#' Backtesting via Value-at-Risk (VaR)
#'
#' @description Method for backtesting a model obtained from \link{bekk_fit} in terms of VaR-forcasting accuracy using a rolling window approach.
#'
#' @param x An object of class "bekkFit" from the function \link{bekk_fit}.
#' @param window_length An integer specifying the length of the rolling window.
#' @param p A numerical value that determines the confidence level. The default value is set at 0.99 in accordance with the Basel Regulation.
#' @param portfolio_weights A vector determining the portfolio weights to calculate the portfolio VaR. If set to "NULL", the univariate VaR for each series are calculated.
#' @param n.ahead Number of periods to predict conditional volatility. Default is a one-period ahead forecast.
#' @param distribution A character string determining the assumed distribution of the residuals. Implemented are "normal", "empirical" and "t". The default is assuming the empirical distribution of the residuals.
#' @param nc Number of cores to be used for parallel computation.
#' @return  Returns a S3 class "backtest" object containing the VaR forecast, out-of-sample returns and backtest statistics according to the R-package "GAS". conf
#' @examples
#' \donttest{
#' \dontshow{Sys.setenv("OMP_THREAD_LIMIT"="1")}
#' data(StocksBonds)
#' obj_spec <- bekk_spec()
#' x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)
#'
#' # backtesting
#' x2 <- backtest(x1, window_length = 6000, n.ahead = 1, nc = 1)
#' plot(x2)
#' # backtesting using 5 day-ahead forecasts
#' x3 <- backtest(x1, window_length = 6000, n.ahead = 5, nc = 1)
#' plot(x3)
#' # backtesting using 20 day-ahead forecasts and portfolio
#' x4 <- backtest(x1, window_length = 6000, portfolio_weights = c(0.5,0.5), n.ahead = 20, nc = 1)
#' plot(x4)
#' }
#'
#' @import xts
#' @import stats
#' @import pbapply
#' @import lubridate
#' @export

backtest<- function(x, window_length = 1000, p = 0.99, portfolio_weights = NULL,  n.ahead = 1, distribution = "empirical", nc = 1) {
  UseMethod('backtest')
}

#' @export
backtest.bekkFit <-  function(x, window_length = 1000, p = 0.99, portfolio_weights = NULL, n.ahead = 1, distribution = "empirical", nc = 1)
{
  data <- x$data
  n <- nrow(data)
  N <- ncol(data)
  n_out = n - window_length

  match.arg(distribution, c("empirical", "t", "normal"))


  #out_sample_returns <-  x$data[(window_length+1):n,] %*% t(portfolio_weights)
  if(window_length < 500){
    stop("The supplied 'window_length' must be larger than 500.")
  }
 if(window_length >= n){
   stop("The supplied 'window_length' exeeds the length of the data.")
 }
  if(window_length < 1000 &&  distribution == "empirical"){
    stop("Empirical distribution only available for window_length > 1000.")
  }

  if((n-window_length) < n.ahead){
    stop("The supplied 'n.ahead' exceeds the forecasting horizon.")
  }
  if(length(portfolio_weights)!= N && !is.null(portfolio_weights)){
    stop("Portfolio weights do no match number of time series")
  }

  if (is.null(portfolio_weights)) {
    hit_rate = numeric(N)
    out_sample_returns <-  x$data[(window_length+1):n,]

    #VaR <- matrix(NA, nrow = n_out, ncol = N)

    OoS_indices <- seq(1,n_out, n.ahead)

    wrapper <- function(i) {
      if(n.ahead > 1 && i > (n_out-n.ahead)){
        n.ahead = n_out-i+1
      }
      spec = x$spec
      fit <- bekk_fit(spec, data[i:(window_length-1+i),])
      predict <- predict(fit, n.ahead = n.ahead, ci = 0.5)
      #VaR[i:(i+n.ahead-1),] = as.matrix(VaR(predict, p = p, portfolio_weights = portfolio_weights)$VaR[(window_length+1):(window_length+n.ahead),])
      res = as.matrix(VaR(predict, p = p, portfolio_weights = portfolio_weights, distribution = distribution)$VaR[(window_length+1):(window_length+n.ahead),])
      return(res)


    }

    if(.Platform$OS.type == "windows") {
      cl = parallel::makeCluster(nc)
      VaR = pbapply::pblapply(X=OoS_indices, FUN=wrapper, cl = cl)
      parallel::stopCluster(cl)
    } else {
      VaR = pbapply::pblapply(X=OoS_indices, FUN=wrapper, cl = nc)
    }




    VaR = do.call(rbind,VaR)


    for(j in 1:N){
             hit_rate[j]=  sum(VaR[,j] > out_sample_returns[,j])
    }


    hit_rate = hit_rate/n_out
    #backtests = list()



    #VaR <- as.data.frame(VaR)
    for (i in 1:N) {
      #backtests[[i]] = suppressWarnings(BacktestVaR(out_sample_returns[,i], VaR[,i], alpha = 1- p))
      colnames(VaR)[i] <- paste('VaR of', colnames(x$data)[i])
    }
  } else {
    out_sample_returns = x$data[(window_length+1):n,] %*% matrix(portfolio_weights, ncol = 1, nrow = N)
    hit_rate = 0

  #  VaR <- matrix(NA, nrow = n_out, ncol = 1)
  #   i = 1
  #
  #     spec = x$spec
  #     fit <- bekk_fit(spec, data[i:(window_length-1+i),])
  #     forecast <- forecast(fit, n.ahead = n.ahead, ci = 0.5)
  #     VaR[i:(i+n.ahead-1),] = as.matrix(VaR(forecast, p = p, portfolio_weights = portfolio_weights)$VaR[(window_length+1):(window_length+n.ahead),])
  #
  #
  #     hit_rate= hit_rate  + sum(VaR[i:(i+n.ahead-1),] > out_sample_returns[i:(i+n.ahead-1),])
  #
  #
  #
  #      i = i + n.ahead
  #      if(n.ahead > 1 && i >= (n_out-n.ahead)){
  #        n.ahead = 1
  #      }
  #   }
  #   hit_rate = hit_rate/n_out
  #   backtests= suppressWarnings(GAS::BacktestVaR(out_sample_returns, VaR, alpha = 1- p))
  #   VaR <- as.data.frame(VaR)
  # }


    OoS_indices <- seq(1,(n_out), n.ahead)
    wrapper <- function(i) {
      if(n.ahead > 1 && i > (n_out-n.ahead)){
        n.ahead = n_out-i+1
      }
      spec = x$spec
      fit <- bekk_fit(spec, data[i:(window_length-1+i),])
      predict <- predict(fit, n.ahead = n.ahead, ci = 0.5)

      res = as.matrix(VaR(predict, p = p, portfolio_weights = portfolio_weights, distribution = distribution)$VaR[(window_length+1):(window_length+n.ahead),])
      return(res)


    }
    #future::plan(future::multicore(workers = n_cores))

    # cl = future::makeClusterPSOCK(nc)
    # VaR = future.apply::future_lapply(X=OoS_indices, FUN=wrapper)

    if(.Platform$OS.type == "Windows") {
      cl = parallel::makeCluster(nc)
      VaR = pbapply::pblapply(X=OoS_indices, FUN=wrapper, cl = cl)
      parallel::stopCluster(cl)
    } else {
      VaR = pbapply::pblapply(X=OoS_indices, FUN=wrapper, cl = nc)
    }

    VaR = do.call(rbind,VaR)

    hit_rate= sum(VaR > out_sample_returns)

    hit_rate = hit_rate/n_out
    #backtests= suppressWarnings(GAS::BacktestVaR(out_sample_returns, VaR, alpha = 1- p))

    out_sample_returns = as.data.frame(out_sample_returns)

    VaR = as.data.frame(VaR)
    colnames(VaR)<- "Portfolio"
    colnames(out_sample_returns)<- "Portfolio"
}

  if (inherits(x$data, "ts")) {
    VaR <- xts(VaR, order.by =  lubridate::as_date(lubridate::date_decimal(time(x$data)[(window_length+1):n])))
    out_sample_returns <- xts(out_sample_returns, order.by = lubridate::as_date(lubridate::date_decimal(time(x$data)[(window_length+1):n])))
  }else if(inherits(x$data, "xts") || inherits(x$data, "zoo") ){
    VaR <- xts(VaR, order.by = time(x$data[(window_length+1):n,]))
    out_sample_returns <- xts(out_sample_returns, order.by = time(x$data[(window_length+1):n,]))

  }



  result=list(
    VaR = VaR,
    out_sample_returns = out_sample_returns,
    hit_rate = hit_rate,
    #backtests = backtests,
    portfolio_weights = portfolio_weights,
    bekkFit = x,
    p = p,
    window_length = window_length
  )
  class(result) <- c('backtest')
  return(result)
}

#' @export
print.var <- function(x,...){
  object <- x
  bekkObject <- object$bekk
  if (any(class(bekkObject) == 'bekk')) {
    cat(paste("\n", "BEKK VaR results", "\n", sep = ""))
    underScore <- paste(rep("-", nchar("BEKK VaR results")), collapse = "")
  } else if (any(class(bekkObject) == 'bekka')) {
    cat(paste("\n", "Asymmetric BEKK VaR results", "\n", sep = ""))
    underScore <- paste(rep("-", nchar("Asymmetric BEKK VaR results")), collapse = "")
  } else if (any(class(bekkObject) == 'dbekk')) {
    cat(paste("\n", "Diagonal BEKK VaR results", "\n", sep = ""))
    underScore <- paste(rep("-", nchar("Diagonal BEKK VaR results")), collapse = "")
  } else if (any(class(bekkObject) == 'dbekka')) {
    cat(paste("\n", "Asymmetric diagonal BEKK VaR results", "\n", sep = ""))
    underScore <- paste(rep("-", nchar("Asymmetric diagonal BEKK VaR results")), collapse = "")
  } else if (any(class(bekkObject) == 'sbekk')) {
    cat(paste("\n", "Scalar BEKK VaR results", "\n", sep = ""))
    underScore <- paste(rep("-", nchar("Scalar BEKK VaR results")), collapse = "")
  } else if (any(class(bekkObject) == 'sbekka')) {
    cat(paste("\n", "Asymmetric scalar BEKK VaR results", "\n", sep = ""))
    underScore <- paste(rep("-", nchar("Asymmetric scalar BEKK VaR results")), collapse = "")
  }
  cat(underScore)
  cat("\nValue-at-risk confidence level: ")
  cat(as.character(object$p))
  if(!is.null(object$portfolio_weights)){
    cat("\nPortfolio weights: ")
    cat(object$portfolio_weights)
  }else{
    cat("\nPortfolio weights: None\n")

  }
}

#' @export

summary.backtest <- function(object, ...) {
  bekkObject <- object$bekkFit
  if (any(class(bekkObject) == 'bekk')) {
    cat(paste("\n", "BEKK backtesting results", "\n", sep = ""))
    underScore <- paste(rep("-", nchar("BEKK backtesting results")), collapse = "")
  } else if (any(class(bekkObject) == 'bekka')) {
    cat(paste("\n", "Asymmetric BEKK backtesting results", "\n", sep = ""))
    underScore <- paste(rep("-", nchar("Asymmetric BEKK backtesting results")), collapse = "")
  } else if (any(class(bekkObject) == 'dbekk')) {
    cat(paste("\n", "Diagonal BEKK backtesting results", "\n", sep = ""))
    underScore <- paste(rep("-", nchar("Diagonal BEKK backtesting results")), collapse = "")
  } else if (any(class(bekkObject) == 'dbekka')) {
    cat(paste("\n", "Asymmetric diagonal BEKK backtesting results", "\n", sep = ""))
    underScore <- paste(rep("-", nchar("Asymmetric diagonal BEKK backtesting results")), collapse = "")
  } else if (any(class(bekkObject) == 'sbekk')) {
    cat(paste("\n", "Scalar BEKK backtesting results", "\n", sep = ""))
    underScore <- paste(rep("-", nchar("Scalar BEKK backtesting results")), collapse = "")
  } else if (any(class(bekkObject) == 'sbekka')) {
    cat(paste("\n", "Asymmetric scalar BEKK backtesting results", "\n", sep = ""))
    underScore <- paste(rep("-", nchar("Asymmetric scalar BEKK backtesting results")), collapse = "")
  }
  cat(underScore)
  cat("\nValue-at-risk confidence level: ")
  cat(as.character(object$p))
  cat("\nWindow length: ")
  cat(as.character(object$window_length))
  if(!is.null(object$portfolio_weights)){
    res_hit <- data.frame(matrix(NA, ncol = ncol(object$VaR), nrow=1))
    res_Kupiec <- data.frame(matrix(NA, ncol = ncol(object$VaR), nrow=2))
    res_Christoffersen <- data.frame(matrix(NA, ncol = ncol(object$VaR), nrow=2))
    colnames(res_hit)=c("")
    #colnames(res_Kupiec)=c("")
    #colnames(res_Christoffersen)=c("")
    row.names(res_hit)=c("")
    #row.names(res_Kupiec)=c("Test:", "p-value:")
    #row.names(res_Christoffersen)=c("Test:", "p-value:")

      #res_Kupiec[,1]=object$backtests$LRuc
      #res_Christoffersen[,1]=object$backtests$LRcc


    cat("\nPortfolio weights: ")
    cat(object$portfolio_weights)
    cat("\n")
    cat(underScore)
    res_hit[1,]=object$hit_rate
    cat(paste("\nHit rate:", round(res_hit,3), "\n", sep = " " ))

    cat("\nUnconditional coverage test of Kupiec: \n")
    print(res_Kupiec)
    cat("\nconditional coverage test of Christoffersen: \n")
    print(res_Christoffersen)
  }else{
    res_hit <- data.frame(matrix(NA, ncol = ncol(object$VaR), nrow=1))
    res_Kupiec <- data.frame(matrix(NA, ncol = ncol(object$VaR), nrow=2))
    res_Christoffersen <- data.frame(matrix(NA, ncol = ncol(object$VaR), nrow=2))
    colnames(res_hit)=colnames(object$out_sample_returns)
    colnames(res_Kupiec)=colnames(res_hit)
    colnames(res_Christoffersen)=colnames(res_hit)
    row.names(res_hit)=c("")
    row.names(res_Kupiec)=c("Test:", "p-value:")
    row.names(res_Christoffersen)=c("Test:", "p-value:")
    # for(i in 1:ncol(object$VaR)){
    #  res_Kupiec[,i]=object$backtests[[i]]$LRuc
    #  res_Christoffersen[,i]=object$backtests[[i]]$LRcc
    # }

    cat("\nPortfolio weights: None\n")
    cat(underScore)
    res_hit[1,]=object$hit_rate
    cat("\nHit rates: \n")
    cat(underScore)
    print(res_hit)
    cat("\nUnconditional coverage test of Kupiec: \n")
    print(res_Kupiec)
    cat("\nconditional coverage test of Christoffersen: \n")
    print(res_Christoffersen)
  }
  cat("\n")

}

#' @export
print.backtest <- function(x,...){
  object <- x
  bekkObject <- object$bekkFit
  if (any(class(bekkObject) == 'bekk')) {
    cat(paste("\n", "BEKK backtesting results", "\n", sep = ""))
    underScore <- paste(rep("-", nchar("BEKK backtesting results")), collapse = "")
  } else if (any(class(bekkObject) == 'bekka')) {
    cat(paste("\n", "Asymmetric BEKK backtesting results", "\n", sep = ""))
    underScore <- paste(rep("-", nchar("Asymmetric BEKK backtesting results")), collapse = "")
  } else if (any(class(bekkObject) == 'dbekk')) {
    cat(paste("\n", "Diagonal BEKK backtesting results", "\n", sep = ""))
    underScore <- paste(rep("-", nchar("Diagonal BEKK backtesting results")), collapse = "")
  } else if (any(class(bekkObject) == 'dbekka')) {
    cat(paste("\n", "Asymmetric diagonal BEKK backtesting results", "\n", sep = ""))
    underScore <- paste(rep("-", nchar("Asymmetric diagonal BEKK backtesting results")), collapse = "")
  } else if (any(class(bekkObject) == 'sbekk')) {
    cat(paste("\n", "Scalar BEKK backtesting results", "\n", sep = ""))
    underScore <- paste(rep("-", nchar("Scalar BEKK backtesting results")), collapse = "")
  } else if (any(class(bekkObject) == 'sbekka')) {
    cat(paste("\n", "Asymmetric scalar BEKK backtesting results", "\n", sep = ""))
    underScore <- paste(rep("-", nchar("Asymmetric scalar BEKK backtesting results")), collapse = "")
  }
  cat(underScore)
  cat("\nValue-at-risk confidence level: ")
  cat(as.character(object$p))
  cat("\nWindow length: ")
  cat(as.character(object$window_length))
  if(!is.null(object$portfolio_weights)){
    cat("\nPortfolio weights: ")
    cat(object$portfolio_weights)
    cat("\n")
    cat(underScore)
    cat("\nHit rate: ")
    cat(round(object$hit_rate,3))
  }else{
    cat("\nPortfolio weights: None\n")
    res_hit <- data.frame(matrix(NA, ncol = ncol(object$VaR), nrow=1))
    colnames(res_hit)=colnames(object$out_sample_returns)
    row.names(res_hit)=c("")

    res_hit[1,]=round(object$hit_rate,3)
    cat(underScore)
    cat("\nHit rates: \n")
    print(res_hit)

  }
  cat("\n")
}

#' Estimating multivariate volatility impulse response functions (VIRF) for BEKK models
#'
#' @description Method for estimating VIRFs of N-dimensional BEKK models. Currently, only VIRFs for symmetric BEKK models are implemented.
#'
#' @param x An object of class "bekkfit" from function \link{bekk_fit}.
#' @param time Time instance to calculate VIRFs for.
#' @param q A number specifying the quantile to be considered for a shock on which basis the VIRFs are generated.
#' @param index_series An integer defining the number of series for which a shock is assumed.
#' @param n.ahead An integer defining the number periods for which the VIRFs are generated.
#' @param ci A number defining the confidence level for the confidence bands.
#' @param time_shock Boolean indicating if the estimated residuals at date specified by "time" shall be used as a shock.
#' @return  Returns an object of class "virf".
#' @references Hafner CM, Herwartz H (2006). Volatility impulse responses for multivariate GARCH models:  An exchange rate illustration. Journal of International Money and Finance,25,719–740.
#' @examples
#' \donttest{
#'
#' data(StocksBonds)
#' obj_spec <- bekk_spec()
#' x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)
#'
#' # 250 day ahead VIRFs and 90% CI for a Shock in the 1% quantile  of Bonds (i.e. series=2)
#' # shock is supposed to occur at day 500
#' x2 <- virf(x1, time = 500, q = 0.01, index_series=2, n.ahead = 500, ci = 0.90)
#' plot(x2)
#' }
#' @import xts
#' @import stats
#' @import numDeriv
#' @export

virf <- function(x ,time = 1, q = 0.05, index_series = 1, n.ahead = 10, ci = 0.9, time_shock = FALSE){

  if (!inherits(x, 'bekkFit')) {
    stop('Please provide and object of class "bekkFit" for x')
  }

  if (x$spec$model$asymmetric==T) {
    stop('VIRFs are implemented only for symmetric BEKK models.')
  }


  if(!( n.ahead%%1==0) || n.ahead < 1){
    stop('Please provide a posive integer for periods')
  }

  if(!(index_series%%1==0) || index_series < 1){
    stop('Please provide a posive integer for index_series')
  }
  if(index_series > ncol(x$data)){
    stop('Total number of indices in the data is lower than index_series')
  }
  if(!is.numeric(time)){
    if( !is.numeric(x$data[time]) || nrow(x$data[time])==0){
  stop('Provided date object is not included in data')
    }
  }else{
  if((time%%1!=0 || time < 1) ){
    stop('Please provide a posive integer or a date object for time')
  }else if(!(time%%1!=0 || time < 1) && time > nrow(x$data)){
    stop('Total number of observations is exeded by time')
  }
  }

  if(!is.numeric(q) || q < 0 || q > 1 || length(q)>1 || length(q) == 0){
    stop('Please provide a number in the interval (0,1) for q.')

  }
  UseMethod('virf')

}

#' @export
virf.bekk <- function(x, time = 1, q = 0.05, index_series=1, n.ahead = 10, ci = 0.9, time_shock = FALSE) {

  N <- ncol(x$data)
  data <- x$data
  H <- matrix(x$H_t[time,],N,N)
  #get quantiles of returns
  residuals = x$e_t
  if(!time_shock ){
  shocks = matrix(0, nrow = 1, ncol = N)

  shocks[index_series] = sapply(q,FUN=quantile,x=as.matrix(residuals[,index_series]))
  for(i in 1: N){
    if(i==index_series){
      shocks[index_series] = sapply(q,FUN=quantile,x=as.matrix(residuals[,index_series]))
    }else{
      shocks[i] = sapply(0.5,FUN=quantile,x=as.matrix(residuals[,i])) *0

    }
  }
  }else{
    shocks = matrix(x$e_t[time,],nrow = 1, ncol = N)
  }

  VIRF = virf_bekk(H, x$theta, matrix(shocks, ncol=N, nrow = 1), n.ahead)
  #dupl <- duplication_mat(N)
  #elim <- elimination_mat(N)

  score_final = score_bekk(x$theta, x$data)
  score_outer = t(score_final) %*% score_final
  # s1 = eigen_value_decomposition(s1_temp)
  hesse_final = solve(hesse_bekk(x$theta, x$data))
  #s1_temp = solve(hesse_final)
  if(x$QML_t_ratios==TRUE){
    Sigma_temp=hesse_final%*%score_outer%*%hesse_final
  }else{
    Sigma_temp=solve(score_outer)
  }

  s1_temp = function(th){
    virf_bekk(H, th, matrix(shocks, ncol=N, nrow = 1), n.ahead)
  }

  th<-x$theta
  d_virf = jacobian(s1_temp,th)
  s1_temp=d_virf%*%Sigma_temp%*%t(d_virf)


#   s1 = s1_temp*0
#   counter = 1
#   while(counter < nrow(s1)){
#     s1[counter:(counter+n.ahead-1),counter:(counter+n.ahead-1)]=s1_temp[counter:(counter+n.ahead-1),counter:(counter+n.ahead-1)]
#   counter = counter + n.ahead
# }

  s1 = sqrt(abs(diag(s1_temp))) * qnorm(ci)
  #return(s1)
  #print(det(d_virf%*%hesse_final%*%t(d_virf)))
  VIRF_lower = VIRF  - matrix(s1, nrow = n.ahead, ncol = N*(N+1)/2)

  VIRF_upper = VIRF + matrix(s1, nrow = n.ahead, ncol = N*(N+1)/2)

  # for (i in 1:nrow(VIRF)) {
  #   tm <- matrix((dupl%*%VIRF[i,]), N, N, byrow = T)
  #   tm2 <- sqrt(solve(diag(abs(diag(tm)))))%*%tm%*%sqrt(solve(diag(abs(diag(tm)))))
  #   diag(tm2) <- sqrt(abs(diag(tm)))%*%solve(diag(abs(diag(tm))))%*%diag(diag(tm))
  #   VIRF[i,] <- elim%*%c(tm2)
  # }




  VIRF <- as.data.frame(VIRF)
  VIRF_lower <- as.data.frame(VIRF_lower)
  VIRF_upper <- as.data.frame(VIRF_upper)
  for(i in 1:ncol(VIRF)){
    colnames(VIRF)[i] <- paste("VIRF for", colnames(x$sigma_t)[i], sep=" ")
         }

  colnames(VIRF) <- gsub("Conditional","conditional",  colnames(VIRF))
  colnames(VIRF) <- gsub("correlation","covariance",  colnames(VIRF))
  colnames(VIRF) <- gsub("standard deviation","variance",  colnames(VIRF))

  result <- list(VIRF=VIRF,
                 VIRF_upper=VIRF_upper,
                 VIRF_lower=VIRF_lower,
                 N=N,
                 time=time,
                 q=q,
                 index_series=index_series,
                 x=x)
  class(result) <- c('virf','bekkFit', 'bekk')
  return(result)
}
#' @export
virf.dbekk <- function(x, time = 1, q = 0.05, index_series=1, n.ahead = 10, ci = 0.9, time_shock = FALSE) {

  N <- ncol(x$data)
  data <- x$data
  H <- matrix(x$H_t[time,],N,N)
  #get quantiles of returns
  residuals = x$e_t
  shocks = matrix(0, nrow = 1, ncol = N)
  if(!time_shock ){
  shocks[index_series] = sapply(q,FUN=quantile,x=as.matrix(residuals[,index_series]))
  for(i in 1: N){
    if(i==index_series){
      shocks[index_series] = sapply(q,FUN=quantile,x=as.matrix(residuals[,index_series]))
    }else{
      shocks[i] = sapply(0.5,FUN=quantile,x=as.matrix(residuals[,i])) * 0

    }
  }
  }else{
    shocks = matrix(x$e_t[time,],nrow = 1, ncol = N)
  }
  VIRF = virf_dbekk(H, x$theta, matrix(shocks, ncol=N, nrow = 1), n.ahead)

  hesse_final = solve(hesse_dbekk(x$theta, x$data))
  score_final = score_dbekk(x$theta, x$data)
  score_outer = t(score_final) %*% score_final
  if(x$QML_t_ratios==TRUE){
    Sigma_temp=hesse_final%*%score_outer%*%hesse_final
  }else{
    Sigma_temp=solve(score_outer)
  }

  s1_temp = function(th){
    virf_dbekk(H, th, matrix(shocks, ncol=N, nrow = 1), n.ahead)
  }

  th<-x$theta
  d_virf = jacobian(s1_temp,th)
  s1_temp=d_virf%*%Sigma_temp%*%t(d_virf)
  #   s1 = s1_temp*0
  #   counter = 1
  #   while(counter < nrow(s1)){
  #     s1[counter:(counter+n.ahead-1),counter:(counter+n.ahead-1)]=s1_temp[counter:(counter+n.ahead-1),counter:(counter+n.ahead-1)]
  #   counter = counter + n.ahead
  # }

  s1 = sqrt(diag(s1_temp)) * qnorm(ci)
  #return(s1)
  #print(det(d_virf%*%hesse_final%*%t(d_virf)))
  VIRF_lower = VIRF  - matrix(s1, nrow = n.ahead, ncol = N*(N+1)/2)

  VIRF_upper = VIRF + matrix(s1, nrow = n.ahead, ncol = N*(N+1)/2)

  # for (i in 1:nrow(VIRF)) {
  #   tm <- matrix((dupl%*%VIRF[i,]), N, N, byrow = T)
  #   tm2 <- sqrt(solve(diag(abs(diag(tm)))))%*%tm%*%sqrt(solve(diag(abs(diag(tm)))))
  #   diag(tm2) <- sqrt(abs(diag(tm)))%*%solve(diag(abs(diag(tm))))%*%diag(diag(tm))
  #   VIRF[i,] <- elim%*%c(tm2)
  # }




  VIRF <- as.data.frame(VIRF)
  VIRF_lower <- as.data.frame(VIRF_lower)
  VIRF_upper <- as.data.frame(VIRF_upper)
  for(i in 1:ncol(VIRF)){
    colnames(VIRF)[i] <- paste("VIRF for", colnames(x$sigma_t)[i], sep=" ")

  }
  colnames(VIRF) <- gsub("Conditional","conditional",  colnames(VIRF))
  colnames(VIRF) <- gsub("correlation","covariance",  colnames(VIRF))
  colnames(VIRF) <- gsub("standard deviation","variance",  colnames(VIRF))



  result <- list(VIRF=VIRF,
                 VIRF_upper=VIRF_upper,
                 VIRF_lower=VIRF_lower,
                 N=N,
                 time=time,
                 q=q,
                 index_series=index_series,
                 x=x)
  class(result) <- c('virf','bekkFit', 'dbekk')
  return(result)
}
#' @export
virf.sbekk <- function(x, time = 1, q = 0.05, index_series=1, n.ahead = 10, ci = 0.9, time_shock = FALSE) {

  N <- ncol(x$data)
  data <- x$data
  H <- matrix(x$H_t[time,],N,N)
  #get quantiles of returns
  residuals = x$e_t
  shocks = matrix(0, nrow = 1, ncol = N)
  if(!time_shock ){
  shocks[index_series] = sapply(q,FUN=quantile,x=as.matrix(residuals[,index_series]))
  for(i in 1: N){
    if(i==index_series){
      shocks[index_series] = sapply(q,FUN=quantile,x=as.matrix(residuals[,index_series]))
    }else{
      shocks[i] = sapply(0.5,FUN=quantile,x=as.matrix(residuals[,i])) * 0

    }
  }
  }else{
    shocks = matrix(x$e_t[time,],nrow = 1, ncol = N)
  }

  VIRF = virf_sbekk(H, x$theta, matrix(shocks, ncol=N, nrow = 1), n.ahead)
  #dupl <- duplication_mat(N)
  #elim <- elimination_mat(N)

  # score_final = score_bekk(x$theta, x$data)
  # s1_temp = solve(t(score_final) %*% score_final)
  # s1 = eigen_value_decomposition(s1_temp)
  #hesse_final = solve(hesse_sbekk(x$theta, x$data))
  #s1_temp = solve(hesse_final)

  hesse_final = solve(hesse_sbekk(x$theta, x$data))
  score_final = score_sbekk(x$theta, x$data)
  score_outer = t(score_final) %*% score_final
  if(x$QML_t_ratios==TRUE){
    Sigma_temp=hesse_final%*%score_outer%*%hesse_final
  }else{
    Sigma_temp=solve(score_outer)
  }

  s1_temp = function(th){
    virf_dbekk(H, th, matrix(shocks, ncol=N, nrow = 1), n.ahead)
  }

  s1_temp = function(th){
    virf_sbekk(H, th, matrix(shocks, ncol=N, nrow = 1), n.ahead)
  }

  th<-x$theta
  d_virf = jacobian(s1_temp,th)
  s1_temp=d_virf%*%Sigma_temp%*%t(d_virf)
  #   s1 = s1_temp*0
  #   counter = 1
  #   while(counter < nrow(s1)){
  #     s1[counter:(counter+n.ahead-1),counter:(counter+n.ahead-1)]=s1_temp[counter:(counter+n.ahead-1),counter:(counter+n.ahead-1)]
  #   counter = counter + n.ahead
  # }

  s1 = sqrt(diag(s1_temp)) * qnorm(ci)  #return(s1)
  #print(det(d_virf%*%hesse_final%*%t(d_virf)))
  VIRF_lower = VIRF  - matrix(s1, nrow = n.ahead, ncol = N*(N+1)/2)

  VIRF_upper = VIRF + matrix(s1, nrow = n.ahead, ncol = N*(N+1)/2)

  # for (i in 1:nrow(VIRF)) {
  #   tm <- matrix((dupl%*%VIRF[i,]), N, N, byrow = T)
  #   tm2 <- sqrt(solve(diag(abs(diag(tm)))))%*%tm%*%sqrt(solve(diag(abs(diag(tm)))))
  #   diag(tm2) <- sqrt(abs(diag(tm)))%*%solve(diag(abs(diag(tm))))%*%diag(diag(tm))
  #   VIRF[i,] <- elim%*%c(tm2)
  # }




  VIRF <- as.data.frame(VIRF)
  VIRF_lower <- as.data.frame(VIRF_lower)
  VIRF_upper <- as.data.frame(VIRF_upper)
  for(i in 1:ncol(VIRF)){
    colnames(VIRF)[i] <- paste("VIRF for", colnames(x$sigma_t)[i], sep=" ")

  }

  colnames(VIRF) <- gsub("Conditional","conditional",  colnames(VIRF))
  colnames(VIRF) <- gsub("correlation","covariance",  colnames(VIRF))
  colnames(VIRF) <- gsub("standard deviation","variance",  colnames(VIRF))


  result <- list(VIRF=VIRF,
                 VIRF_upper=VIRF_upper,
                 VIRF_lower=VIRF_lower,
                 N=N,
                 time=time,
                 q=q,
                 index_series=index_series,
                 x=x)
  class(result) <- c('virf','bekkFit', 'sbekk')
  return(result)
}

#' bekkFit method
#'
#' @description Generic 'bekkFit' methods. More details on 'bekkFit' are described in \link{bekk_fit}
#'
#' @param x An object of class "bekkFit" from function \link{bekk_fit}.
#' @param object An object of class "bekkFit" from function \link{bekk_fit}.
#' @param k Numeric value, the penalty per parameter for AIC to be used; the default k = 2 is the classical AIC.
#' @param ...	Further arguments to be passed to and from other methods.
#'
#' @examples
#' \donttest{
#'
#' data(StocksBonds)
#'
#' # Fitting a symmetric BEKK model
#' obj_spec <- bekk_spec()
#' x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)
#'
#' logLik(x1)
#' }
#' @import xts
#' @import stats
#' @rdname bekk_fit_methods
#' @export
print.bekkFit <- function(x,...){
  bekkObject <- x

  if (all(class(bekkObject) != 'var')) {
    if (any(class(bekkObject) == 'bekk')) {
      cat(paste("\n", "BEKK estimation results", "\n", sep = ""))
      underScore <- paste(rep("-", nchar("BEKK estimation results")), collapse = "")
    } else if (any(class(bekkObject) == 'bekka')) {
      cat(paste("\n", "Asymmetric BEKK estimation results", "\n", sep = ""))
      underScore <- paste(rep("-", nchar("Asymmetric BEKK estimation results")), collapse = "")
    } else if (any(class(bekkObject) == 'dbekk')) {
      cat(paste("\n", "Diagonal BEKK estimation results", "\n", sep = ""))
      underScore <- paste(rep("-", nchar("Diagonal BEKK estimation results")), collapse = "")
    } else if (any(class(bekkObject) == 'dbekka')) {
      cat(paste("\n", "Asymmetric diagonal BEKK estimation results", "\n", sep = ""))
      underScore <- paste(rep("-", nchar("Asymmetric diagonal BEKK estimation results")), collapse = "")
    } else if (any(class(bekkObject) == 'sbekk')) {
      cat(paste("\n", "Scalar BEKK estimation results", "\n", sep = ""))
      underScore <- paste(rep("-", nchar("Scalar BEKK estimation results")), collapse = "")
    } else if (any(class(bekkObject) == 'sbekka')) {
      cat(paste("\n", "Asymmetric scalar BEKK estimation results", "\n", sep = ""))
      underScore <- paste(rep("-", nchar("Asymmetric scalar BEKK estimation results")), collapse = "")
    }

    cat(underScore)
    cat("\nLog-likelihood: ")
    cat(bekkObject$log_likelihood)
    cat("\nBEKK model stationary: ")
    cat(bekkObject$BEKK_valid)
    cat("\nNumber of BHHH iterations: ")
    cat(bekkObject$iter)
    cat("\nAIC: ")
    cat(bekkObject$AIC)
    cat("\nBIC: ")
    cat(bekkObject$BIC)
    cat("\n")
  }
  bekkObject
}

#' @rdname bekk_fit_methods
#' @export
residuals.bekkFit <- function(object, ...) {
  object$e_t
}




#' @noRd
AIC.bekkFit <- function(object, ..., k = 2) {

  x <- object

  AICinner <- function(e) {
    N <- ncol(e$data)
    if (any(class(e) == 'bekk')) {
      aic <- k * (2 * N^2 + N * (N + 1)/2) - 2 * llv(e)
    } else if (any(class(e) == 'bekka')) {
      aic <- k * (3 * N^2 + N * (N + 1)/2) - 2 * llv(e)
    } else if (any(class(e) == 'sbekk')) {
      aic <- k * (2  + N * (N + 1)/2) - 2 * llv(e)
    } else if (any(class(e) == 'sbekka')) {
      aic <- k * (3 + N * (N + 1)/2) - 2 * llv(e)
    }else if (any(class(e) == 'dbekk')) {
      aic <- k * (2  + N * (N + 1)/2) - 2 * llv(e)
    } else if (any(class(e) == 'dbekka')) {
      aic <- k * (3 + N * (N + 1)/2) - 2 * llv(e)
    }
    return(aic)
  }

  if(!missing(...)) {# several objects: produce data.frame
    lls <- sapply(list(x, ...), AICinner)
    vals <- sapply(list(x, ...), function(e1){length(e1$theta)})
    aic <- data.frame(df = vals, AIC = lls)
  } else {
    aic <- AICinner(x)
    aic <- data.frame(df = length(x$theta), AIC = aic)
  }

  return(aic)
}
#' @noRd
BIC.bekkFit <- function(object, ...) {

  x <- object

  BICinner <- function(e) {
    N <- ncol(e$data)
    if (any(class(e) == 'bekk')) {
      bic <- (2* N^2 + N * (N + 1)/2) * log(nrow(e$data)) - 2 * llv(e)
    } else if (any(class(e) == 'bekka')) {
      bic <- (3 * N^2 + N * (N + 1)/2) * log(nrow(e$data)) - 2 * llv(e)
    } else if (any(class(e) == 'sbekk')) {
      bic <- (2 + N * (N + 1)/2) * log(nrow(e$data)) - 2 * llv(e)
    } else if (any(class(e) == 'sbekka')) {
      bic <- (3 + N * (N + 1)/2) * log(nrow(e$data)) - 2 * llv(e)
    } else if (any(class(e) == 'dbekk')) {
      bic <- (2 * N + N * (N + 1)/2)*  log(nrow(e$data)) - 2 * llv(e)
    } else if (any(class(e) == 'dbekka')) {
      bic <- (3 * N + N * (N + 1)/2) * log(nrow(e$data)) - 2 * llv(e)
    }
    return(bic)
  }

  if(!missing(...)) {# several objects: produce data.frame
    lls <- sapply(list(x, ...), BICinner)
    vals <- sapply(list(x, ...), function(e1){length(e1$theta)})
    bic <- data.frame(df = vals, BIC = lls)
  } else {
    bic <- data.frame(df = length(x$theta), BIC = BICinner(x))
  }

  return(bic)
}

#' @noRd
llv <- function(object) {


  llv_inner <- function(x){
    if (any(class(x) == 'bekk')) {
      logl <- loglike_bekk(x$theta, x$data)
    } else if (any(class(x) == 'bekka')) {
      logl <- loglike_asymm_bekk(x$theta, x$data, x$signs)
    } else if (any(class(x) == 'sbekk')) {
      logl <- loglike_sbekk(x$theta, x$data)
    } else if (any(class(x) == 'sbekka')) {
      logl <- loglike_asymm_sbekk(x$theta, x$data, x$signs)
    }else if (any(class(x) == 'dbekk')) {
      logl <- loglike_dbekk(x$theta, x$data)
    } else if (any(class(x) == 'dbekka')) {
      logl <- loglike_asymm_dbekk(x$theta, x$data, x$signs)
    }

    return(logl)

  }


    lls <- llv_inner(object)



  return(lls)
}

#' @rdname bekk_fit_methods
#' @export
logLik.bekkFit <- function(object, ..., k = 2) {


  llv_inner <- function(x){
  if (any(class(x) == 'bekk')) {
    logl <- loglike_bekk(x$theta, x$data)
  } else if (any(class(x) == 'bekka')) {
    logl <- loglike_asymm_bekk(x$theta, x$data, x$signs)
  } else if (any(class(x) == 'sbekk')) {
    logl <- loglike_sbekk(x$theta, x$data)
  } else if (any(class(x) == 'sbekka')) {
    logl <- loglike_asymm_sbekk(x$theta, x$data, x$signs)
  }else if (any(class(x) == 'dbekk')) {
    logl <- loglike_dbekk(x$theta, x$data)
  } else if (any(class(x) == 'dbekka')) {
    logl <- loglike_asymm_dbekk(x$theta, x$data, x$signs)
  }

  return(logl)

  }

  if(!missing(...)) {# several objects: produce data.frame
    lls <- sapply(list(object, ...), llv_inner)
    vals <- sapply(list(object, ...), function(e1){length(e1$theta)})
    aic <- data.frame(df = vals, LLV = lls, AIC = AIC(object, ..., k = k)$AIC, BIC = BIC(object, ...)$BIC)
  } else {
    lls <- llv_inner(object)
    aic <- data.frame(df = length(object$theta), LLV = lls, AIC = AIC(object, k = k)$AIC, BIC = BIC(object)$BIC)


     }
  return(aic)
}

#' @name simulate
#' @rdname simulate
#' @title Simulating BEKK models
#'
#' @description Method for simulating a N-dimensional BEKK model.
#'
#' @param object A spec object of class "bekkSpec" from the function \link{bekk_spec} or a fitted bekk model of class "bekkFit" from the \link{bekk_fit} function
#' @param nsim Number of observations of the simulated sample
#' @param ... Further parameters to be passed on to the function.
#' @return Returns a simulated time series S3 class object using the parameters of passed "bekkSpec" or "bekkFit".
#'
#' @examples
#' \donttest{
#'
#' # simulate a BEKK with estimated parameter
#' obj_spec <- bekk_spec()
#' x1 <- bekk_fit(obj_spec, StocksBonds)
#'
#' x2 <- simulate(x1, nsim = 3000)
#'
#' plot(x2)
#'
#' }
#'

#' @export
simulate.bekk <- function(object, nsim, ...) {
  spec <- object
  if(is.null(nsim) || !is.numeric(nsim) || nsim < 1){
    stop("Please provide an integer specifying the number of observations")
  }
  if(!inherits(spec,c("bekkSpec", "bekkFit"))){
    stop("Please provide an object of class bekk_fit or 'bekk_spec'.")
  }
  xx <- process_object(spec)
  par <- coef_mat(xx$theta,xx$N)
  if(xx$BEKK_valid==FALSE){
    stop("Please provide a stationary BEKK model.")

  }

  sim_dat <- simulate_bekk_c(c(xx$theta), nsim, xx$N)

  return(ts(sim_dat))
}
#' @rdname simulate

#' @export
simulate.bekka <- function(object, ..., nsim) {
  spec <- object
  if(is.null(nsim) || !is.numeric(nsim) || nsim < 1){
    stop("Please provide an integer specifying the number of observations")
  }
  if(!inherits(spec,c("bekkSpec", "bekkFit"))){
    stop("Please provide an object of class bekk_fit or 'bekk_spec'.")
  }
  xx <- process_object(spec)
  par <- coef_mat_asymm(xx$theta,xx$N)
  if(xx$BEKK_valid==FALSE){
    stop("Please provide a stationary BEKK model.")

  }
  #expected_signs
  sim_dat <- simulate_bekka_c(c(xx$theta), nsim, xx$N, xx$signs, xx$expected_signs)

  return(ts(sim_dat))
}

#' @rdname simulate

#' @export
simulate.dbekk <- function(object, ..., nsim) {
  spec <- object
  if(is.null(nsim) || !is.numeric(nsim) || nsim < 1){
    stop("Please provide an integer specifying the number of observations")
  }
  if(!inherits(spec,c("bekkSpec", "bekkFit"))){
    stop("Please provide an object of class bekk_fit or 'bekk_spec'.")
  }
  xx <- process_object(spec)
  par <- coef_mat_diagonal(xx$theta,xx$N)
  if(xx$BEKK_valid==FALSE){
    stop("Please provide a stationary BEKK model.")

  }

  sim_dat <- simulate_dbekk_c(c(xx$theta), nsim, xx$N)

  return(ts(sim_dat))
}
#' @rdname simulate
#' @export
simulate.dbekka <- function(object, ..., nsim) {
  spec <- object
  if(is.null(nsim) || !is.numeric(nsim) || nsim < 1){
    stop("Please provide an integer specifying the number of observations")
  }
  if(!inherits(spec,c("bekkSpec", "bekkFit"))){
    stop("Please provide an object of class bekk_fit or 'bekk_spec'.")
  }
  xx <- process_object(spec)
  par <- coef_mat_asymm_diagonal(xx$theta,xx$N)
  if(xx$BEKK_valid==FALSE){
    stop("Please provide a stationary BEKK model.")
  }
  #expected_signs
  sim_dat <- simulate_dbekka_c(c(xx$theta), nsim, xx$N, xx$signs, xx$expected_signs)

  return(ts(sim_dat))
}

#' @rdname simulate
#' @export
simulate.sbekk <- function(object, ..., nsim) {
  spec <- object
  if(is.null(nsim) || !is.numeric(nsim) || nsim < 1){
    stop("Please provide an integer specifying the number of observations")
  }
  if(!inherits(spec,c("bekkSpec", "bekkFit"))){
    stop("Please provide an object of class bekk_fit or 'bekk_spec'.")
  }
  xx <- process_object(spec)
  par <- coef_mat_scalar(xx$theta,xx$N)
  if(xx$BEKK_valid==FALSE){
    stop("Please provide a stationary BEKK model.")
  }

  sim_dat <- simulate_sbekk_c(c(xx$theta), nsim, xx$N)

  return(ts(sim_dat))
}
#' @rdname simulate
#' @export
simulate.sbekka <- function(object, ..., nsim) {
  spec <- object
  if(is.null(nsim) || !is.numeric(nsim) || nsim < 1){
    stop("Please provide an integer specifying the number of observations")
  }
  if(!inherits(spec,c("bekkSpec", "bekkFit"))){
    stop("Please provide an object of class bekk_fit or 'bekk_spec'.")
  }
  xx <- process_object(spec)
  par <- coef_mat_asymm_scalar(xx$theta,xx$N)
  if(xx$BEKK_valid==FALSE){
    stop("Please provide a stationary BEKK model.")
      }
  #expected_signs
  sim_dat <- simulate_sbekka_c(c(xx$theta), nsim, xx$N, xx$signs, xx$expected_signs)

  return(ts(sim_dat))
}

#' @name predict
#' @rdname predict
#' @title Forecasting conditional volatilities with BEKK models
#'
#' @description Method for predicting a N-dimensional BEKK covariances.
#'
#' @param object A fitted bekk model of class "bekkFit" from the \link{bekk_fit} function
#' @param n.ahead Number of periods to forecast conditional volatility. Default is a one-period ahead forecast.
#' @param ci Floating point in [0,1] defining the niveau for confidence bands of the conditional volatility forecast. Default is 95 per cent niveau confidence bands.
#' @param ... Further parameters to be passed on to the function.
#' @return Returns a S3 class "bekkForecast" object containing the conditional volatility forecasts and respective confindence bands.
#' @examples
#' \donttest{#'
#' data(StocksBonds)
#' obj_spec <- bekk_spec()
#' x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)
#'
#' x2 <- predict(x1, n.ahead = 1)
#'
#' }



#' @export
predict.bekk <- function(object, n.ahead = 1, ci = 0.95, ...) {
  x <- object
  N <- ncol(x$data)
  NoBs <- nrow(x$data)
  #var_process <- sigma_bekk(xx$data, xx$C0, xx$A, xx$G)
  H_t <- vector(mode = "list",length = n.ahead+1)
  H_t[[1]] <- matrix(x$H_t[NoBs,],nrow = N, ncol = N)
  current_returns <- matrix(c(x$data[NoBs,]), nrow = 1)

  for(i in 1:n.ahead){
    H_t[[i+1]] <- x$C0 %*% t(x$C0) + t(x$A) %*% t(current_returns) %*% current_returns %*% x$A + t(x$G) %*% H_t[[i]] %*% x$G
    current_returns <- eigen_value_decomposition(H_t[[i+1]])
  }

  sigma_t <- matrix(NA, nrow = n.ahead, ncol = N^2)
  for (i in 2:(n.ahead+1)){
      tm2 <- sqrt(solve(diag(diag(H_t[[i]]))))%*%H_t[[i]]%*%sqrt(solve(diag(diag(H_t[[i]]))))
      diag(tm2) <- sqrt(diag(H_t[[i]]))
      sigma_t[i-1,] <- c(tm2)
  }

  colnames(sigma_t) <- rep(1, N^2)

  k <- 1
  k2 <- 1
  for (i in 1:N) {
    for (j in 1:N) {
      if (i == j) {
        colnames(sigma_t)[k2] <- paste('Conditional standard deviation of \n', colnames(x$data)[k])
        k <- k + 1
        k2 <- k2 +1
      } else {
        colnames(sigma_t)[k2] <- paste('Conditional correlation of \n', colnames(x$data)[i], ' and ', colnames(x$data)[j])
        k2 <- k2 +1
      }
    }
  }

  elim <- elimination_mat(N)
  sigma_t <- sigma_t[, which(colSums(elim) == 1)]

  H_t_f <- matrix(NA, nrow = n.ahead, ncol = ncol(x$data)^2)

  for (i in 2:(n.ahead+1)){
    H_t_f[i-1, ] <- c(H_t[[i]])
  }

  # Generating confidence intervals

  score_final = score_bekk(x$theta, x$data)
  s1_temp = diag(inv_gen(t(score_final) %*% score_final),names=T )
  s1 = sqrt(s1_temp)

  lower_theta = x$theta - qnorm(ci)*s1
  upper_theta = x$theta + qnorm(ci)*s1

  H_t_lower <- vector(mode = "list",length = n.ahead+1)
  H_t_upper <- vector(mode = "list",length = n.ahead+1)
  x_lower <- coef_mat(lower_theta, N)
  x_upper <- coef_mat(upper_theta, N)
  if(!any(c(valid_bekk(x_lower$c0,x_lower$a,x_lower$g), valid_bekk(x_upper$c0,x_upper$a,x_upper$g)))){
    stop("Lower and/or upper CI are not leading to a stationary BEKK model. Please decrease the cofidence level ci.")
  }
  #check if t(x_lower$C0) or x_lower$C0 for sigma_bekk
  H_t_lower[[1]] <- matrix(sigma_bekk(x$data, t(x_lower$c0), x_lower$a, x_lower$g)$sigma_t[NoBs,], nrow = N, ncol = N)
  H_t_upper[[1]] <- matrix(sigma_bekk(x$data,t(x_upper$c0),x_upper$a, x_upper$g)$sigma_t[NoBs,], nrow = N, ncol = N)
  current_returns <- matrix(c(x$data[NoBs,]), nrow = 1)


  for(i in 1:n.ahead){
    H_t_lower[[i+1]] <- x_lower$c0 %*% t(x_lower$c0) + t(x_lower$a) %*% t(current_returns) %*% current_returns %*% x_lower$a + t(x_lower$g) %*% H_t[[i]] %*% x_lower$g
    current_returns <- eigen_value_decomposition(H_t_lower[[i+1]])
  }

  current_returns <- matrix(c(x$data[NoBs,]), nrow = 1)

  for(i in 1:n.ahead){
    H_t_upper[[i+1]] <- x_upper$c0 %*% t(x_upper$c0) + t(x_upper$a) %*% t(current_returns) %*% current_returns %*% x_upper$a + t(x_upper$g) %*% H_t[[i]] %*% x_upper$g
    current_returns <- eigen_value_decomposition(H_t_upper[[i+1]])
  }
  sigma_t_lower <- matrix(NA, nrow = n.ahead, ncol = N^2)

  for (i in 2:(n.ahead+1)){
    tm2 <- sqrt(solve(diag(diag(H_t_lower[[i]]))))%*%H_t_lower[[i]]%*%sqrt(solve(diag(diag(H_t_lower[[i]]))))
    diag(tm2) <- sqrt(diag(H_t_lower[[i]]))
    sigma_t_lower[i-1,] <- c(tm2)
  }
  sigma_t_upper <- matrix(NA, nrow = n.ahead, ncol = N^2)
  for (i in 2:(n.ahead+1)){
    tm2 <- sqrt(solve(diag(diag(H_t_upper[[i]]))))%*%H_t_upper[[i]]%*%sqrt(solve(diag(diag(H_t_upper[[i]]))))
    diag(tm2) <- sqrt(diag(H_t_upper[[i]]))
    sigma_t_upper[i-1,] <- c(tm2)
  }


  colnames(sigma_t_lower) <- rep(1, N^2)
  colnames(sigma_t_upper) <- rep(1, N^2)

  k <- 1
  k2 <- 1
  for (i in 1:N) {
    for (j in 1:N) {
      if (i == j) {
        colnames(sigma_t_lower)[k2] <- paste('Lower CI conditional standard deviation of \n', colnames(x$data)[k])
        colnames(sigma_t_upper)[k2] <- paste('Upper CI conditional standard deviation of \n', colnames(x$data)[k])
        k <- k + 1
        k2 <- k2 +1
      } else {
        colnames(sigma_t_lower)[k2] <- paste('Lower CI conditional correlation of \n', colnames(x$data)[i], ' and ', colnames(x$data)[j])
        colnames(sigma_t_upper)[k2] <- paste('Upper CI conditional correlation of \n', colnames(x$data)[i], ' and ', colnames(x$data)[j])

        k2 <- k2 +1
      }
    }
  }

  elim <- elimination_mat(N)
  sigma_t_lower <- sigma_t_lower[, which(colSums(elim) == 1)]
  sigma_t_upper <- sigma_t_upper[, which(colSums(elim) == 1)]

  H_t_f_lower <- H_t_f_upper <- matrix(NA, nrow = n.ahead+1 , ncol = ncol(x$data)^2)

  for (i in 1:(n.ahead+1)){
    H_t_f_lower[i, ] <- c(H_t_lower[[i]])
    H_t_f_upper[i, ] <- c(H_t_upper[[i]])
  }

  if (inherits(x$data, "ts")) {
    sigma_t <- ts(sigma_t, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    sigma_t_lower <- ts(sigma_t_lower, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    sigma_t_upper <- ts(sigma_t_upper, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    H_t_f <- ts(H_t_f, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    H_t_f_lower <- ts(H_t_f_lower, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    H_t_f_upper <- ts(H_t_f_upper, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))

    }
  else if(inherits(x$data, "xts") || inherits(x$data, "zoo") ){
    sigma_t <- xts(matrix(sigma_t, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                   by = periodicity(x$data)$units))
    sigma_t_lower <- xts(matrix(sigma_t_lower, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                         by = periodicity(x$data)$units))
    sigma_t_upper <- xts(matrix(sigma_t_upper, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                         by = periodicity(x$data)$units))
    H_t_f <- xts(matrix(H_t_f, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                   by = periodicity(x$data)$units))
    H_t_f_lower <- xts(matrix(H_t_f_lower, nrow = n.ahead+1), order.by = seq((time(x$data)[nrow(x$data)]), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                               by = periodicity(x$data)$units))
    H_t_f_upper <- xts(matrix(H_t_f_upper, nrow = n.ahead+1), order.by = seq((time(x$data)[nrow(x$data)]), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                               by = periodicity(x$data)$units))

    }


  result <- list(
    volatility_forecast = sigma_t,
    volatility_lower_conf_band = sigma_t_lower,
    volatility_upper_conf_band = sigma_t_upper,
    H_t_forecast = H_t_f,
    H_t_lower_conf_band = H_t_f_lower,
    H_t_upper_conf_band = H_t_f_upper,
    n.ahead = n.ahead,
    bekkfit = x
  )
  class(result) <- c('bekkForecast', 'bekk')
  return(result)
}
#' @rdname predict
#' @export
predict.bekka <- function(object, n.ahead = 1, ci = 0.95, ...) {
  x <- object
  N <- ncol(x$data)
  NoBs <- nrow(x$data)
  #var_process <- sigma_bekk(xx$data, xx$C0, xx$A, xx$G)
  H_t = vector(mode = "list",length=n.ahead+1)
  H_t[[1]] = matrix(x$H_t[NoBs,],nrow = N, ncol = N)
  current_returns <- matrix(c(x$data[NoBs,]), nrow = 1)

  H_t[[2]]=x$C0 %*% t(x$C0) + t(x$A) %*% t(current_returns) %*% current_returns %*% x$A + indicatorFunction(as.matrix(current_returns), x$signs) * t(x$B) %*% t(current_returns) %*% current_returns %*% x$B + t(x$G) %*% H_t[[1]] %*% x$G

  expected_signs=expected_indicator_value(x$data,x$signs)
  for(i in 2:n.ahead){
    H_t[[i+1]]=x$C0 %*% t(x$C0) + t(x$A) %*% H_t[[i]] %*% x$A + expected_signs * t(x$B) %*% H_t[[i]] %*% x$B + t(x$G) %*% H_t[[i]] %*% x$G

  }

  sigma_t = matrix(NA, nrow = n.ahead, ncol = N^2)
  for (i in 1:(n.ahead+1)){
    tm2 <- sqrt(solve(diag(diag(H_t[[i]]))))%*%H_t[[i]]%*%sqrt(solve(diag(diag(H_t[[i]]))))
    diag(tm2) <- sqrt(diag(H_t[[i]]))
    sigma_t[i-1,] <- c(tm2)
  }

  colnames(sigma_t) <- rep(1, N^2)

  k <- 1
  k2 <- 1
  for (i in 1:N) {
    for (j in 1:N) {
      if (i == j) {
        colnames(sigma_t)[k2] <- paste('Conditional standard deviation of \n', colnames(x$data)[k])
        k <- k + 1
        k2 <- k2 +1
      } else {
        colnames(sigma_t)[k2] <- paste('Conditional correlation of \n', colnames(x$data)[i], ' and ', colnames(x$data)[j])
        k2 <- k2 +1
      }
    }
  }

  elim <- elimination_mat(N)
  sigma_t <- sigma_t[, which(colSums(elim) == 1)]


  H_t_f <- matrix(NA, nrow = n.ahead, ncol = ncol(x$data)^2)

  for (i in 1:(n.ahead+1)){
    H_t_f[i-1, ] <- c(H_t[[i]])
  }


  #95% confidence interval

  score_final = score_asymm_bekk(x$theta, x$data, x$signs)
  s1_temp = diag(inv_gen(t(score_final) %*% score_final),names=T)
  s1 = sqrt(s1_temp)

  lower_theta = x$theta - qnorm(ci)*s1
  upper_theta = x$theta + qnorm(ci)*s1

  H_t_lower <- vector(mode = "list",length = n.ahead+1)
  H_t_upper <- vector(mode = "list",length = n.ahead+1)
  x_lower <- coef_mat_asymm(lower_theta, N)
  x_upper <- coef_mat_asymm(upper_theta, N)
  #check if t(x_lower$C0) or x_lower$C0 for sigma_bekk
  H_t_lower[[1]] <- matrix(sigma_bekk_asymm(x$data, t(x_lower$c0), x_lower$a, x_lower$b, x_lower$g, x$signs)$sigma_t[NoBs,],nrow = N, ncol = N)
  H_t_upper[[1]] <- matrix(sigma_bekk_asymm(x$data, t(x_upper$c0), x_upper$a, x_upper$b, x_upper$g, x$signs)$sigma_t[NoBs,],nrow = N, ncol = N)
  current_returns <- matrix(c(x$data[NoBs,]), nrow = 1)


  for(i in 1:n.ahead){
    H_t_lower[[i+1]] <- x_lower$c0 %*% t(x_lower$c0) + t(x_lower$a) %*% t(current_returns) %*% current_returns %*% x_lower$a + expected_signs * t(x_lower$b) %*% H_t_lower[[i]] %*% x_lower$b + t(x_lower$g) %*% H_t[[i]] %*% x_lower$g
    current_returns <- eigen_value_decomposition(H_t_lower[[i+1]])
  }

  current_returns <- matrix(c(x$data[NoBs,]), nrow = 1)

  for(i in 1:n.ahead){
    H_t_upper[[i+1]] <- x_upper$c0 %*% t(x_upper$c0) + t(x_upper$a) %*% t(current_returns) %*% current_returns %*% x_upper$a + expected_signs * t(x_upper$b) %*% H_t_upper[[i]] %*% x_upper$b + t(x_upper$g) %*% H_t[[i]] %*% x_upper$g
    current_returns <- eigen_value_decomposition(H_t_upper[[i+1]])
  }
  sigma_t_lower <- matrix(NA, nrow = n.ahead, ncol = N^2)

  for (i in 2:(n.ahead+1)){
    tm2 <- sqrt(solve(diag(diag(H_t_lower[[i]]))))%*%H_t_lower[[i]]%*%sqrt(solve(diag(diag(H_t_lower[[i]]))))
    diag(tm2) <- sqrt(diag(H_t_lower[[i]]))
    sigma_t_lower[i-1,] <- c(tm2)
  }
  sigma_t_upper <- matrix(NA, nrow = n.ahead, ncol = N^2)
  for (i in 2:(n.ahead+1)){
    tm2 <- sqrt(solve(diag(diag(H_t_upper[[i]]))))%*%H_t_upper[[i]]%*%sqrt(solve(diag(diag(H_t_upper[[i]]))))
    diag(tm2) <- sqrt(diag(H_t_upper[[i]]))
    sigma_t_upper[i-1,] <- c(tm2)
  }


  colnames(sigma_t_lower) <- rep(1, N^2)
  colnames(sigma_t_upper) <- rep(1, N^2)

  k <- 1
  k2 <- 1
  for (i in 1:N) {
    for (j in 1:N) {
      if (i == j) {
        colnames(sigma_t_lower)[k2] <- paste('Lower CI conditional standard deviation of \n', colnames(x$data)[k])
        colnames(sigma_t_upper)[k2] <- paste('Upper CI conditional standard deviation of \n', colnames(x$data)[k])
        k <- k + 1
        k2 <- k2 +1
      } else {
        colnames(sigma_t_lower)[k2] <- paste('Lower CI conditional correlation of \n', colnames(x$data)[i], ' and ', colnames(x$data)[j])
        colnames(sigma_t_upper)[k2] <- paste('Upper CI conditional correlation of \n', colnames(x$data)[i], ' and ', colnames(x$data)[j])

        k2 <- k2 +1
      }
    }
  }

  elim <- elimination_mat(N)
  sigma_t_lower <- sigma_t_lower[, which(colSums(elim) == 1)]
  sigma_t_upper <- sigma_t_upper[, which(colSums(elim) == 1)]

  H_t_f_lower <- H_t_f_upper <- matrix(NA, nrow = n.ahead + 1, ncol = ncol(x$data)^2)

  for (i in 1:(n.ahead+1)){
    H_t_f_lower[i, ] <- c(H_t_lower[[i]])
    H_t_f_upper[i, ] <- c(H_t_upper[[i]])
  }

  if (inherits(x$data, "ts")) {
    sigma_t <- ts(sigma_t, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    sigma_t_lower <- ts(sigma_t_lower, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    sigma_t_upper <- ts(sigma_t_upper, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    H_t_f <- ts(H_t_f, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    H_t_f_lower <- ts(H_t_f_lower, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    H_t_f_upper <- ts(H_t_f_upper, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))

  }
  else if(inherits(x$data, "xts") || inherits(x$data, "zoo") ){
    sigma_t <- xts(matrix(sigma_t, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                   by = periodicity(x$data)$units))
    sigma_t_lower <- xts(matrix(sigma_t_lower, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                               by = periodicity(x$data)$units))
    sigma_t_upper <- xts(matrix(sigma_t_upper, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                               by = periodicity(x$data)$units))
    H_t_f <- xts(matrix(H_t_f, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                               by = periodicity(x$data)$units))
    H_t_f_lower <- xts(matrix(H_t_f_lower, nrow = n.ahead+1), order.by = seq((time(x$data)[nrow(x$data)]), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                             by = periodicity(x$data)$units))
    H_t_f_upper <- xts(matrix(H_t_f_upper, nrow = n.ahead+1), order.by = seq((time(x$data)[nrow(x$data)]), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                             by = periodicity(x$data)$units))

  }


  result <- list(
    volatility_forecast = sigma_t,
    volatility_lower_conf_band = sigma_t_lower,
    volatility_upper_conf_band = sigma_t_upper,
    H_t_forecast = H_t_f,
    H_t_lower_conf_band = H_t_f_lower,
    H_t_upper_conf_band = H_t_f_upper,
    n.ahead = n.ahead,
    bekkfit = x
  )
  class(result) <-  c('bekkForecast', 'bekka')
  return(result)
}

#' @rdname predict
#' @export
predict.dbekk <- function(object, n.ahead = 1, ci = 0.95, ...) {
  x <- object
  N <- ncol(x$data)
  NoBs <- nrow(x$data)
  #var_process <- sigma_bekk(xx$data, xx$C0, xx$A, xx$G)
  H_t <- vector(mode = "list",length = n.ahead+1)
  H_t[[1]] <- matrix(x$H_t[NoBs,],nrow = N, ncol = N)
  current_returns <- matrix(c(x$data[NoBs,]), nrow = 1)

  for(i in 1:n.ahead){
    H_t[[i+1]] <- x$C0 %*% t(x$C0) + t(x$A) %*% t(current_returns) %*% current_returns %*% x$A + t(x$G) %*% H_t[[i]] %*% x$G
    current_returns <- eigen_value_decomposition(H_t[[i+1]])
  }

  sigma_t <- matrix(NA, nrow = n.ahead, ncol = N^2)
  for (i in 2:(n.ahead+1)){
    tm2 <- sqrt(solve(diag(diag(H_t[[i]]))))%*%H_t[[i]]%*%sqrt(solve(diag(diag(H_t[[i]]))))
    diag(tm2) <- sqrt(diag(H_t[[i]]))
    sigma_t[i-1,] <- c(tm2)
  }

  colnames(sigma_t) <- rep(1, N^2)

  k <- 1
  k2 <- 1
  for (i in 1:N) {
    for (j in 1:N) {
      if (i == j) {
        colnames(sigma_t)[k2] <- paste('Conditional standard deviation of \n', colnames(x$data)[k])
        k <- k + 1
        k2 <- k2 +1
      } else {
        colnames(sigma_t)[k2] <- paste('Conditional correlation of \n', colnames(x$data)[i], ' and ', colnames(x$data)[j])
        k2 <- k2 +1
      }
    }
  }

  elim <- elimination_mat(N)
  sigma_t <- sigma_t[, which(colSums(elim) == 1)]

  H_t_f <- matrix(NA, nrow = n.ahead, ncol = ncol(x$data)^2)

  for (i in 2:(n.ahead+1)){
    H_t_f[i-1, ] <- c(H_t[[i]])
  }

  # Generating confidence intervals

  score_final = score_dbekk(x$theta, x$data)
  s1_temp = diag(inv_gen(t(score_final) %*% score_final),names=T)
  s1 = sqrt(s1_temp)

  lower_theta = x$theta - qnorm(ci)*s1
  upper_theta = x$theta + qnorm(ci)*s1

  H_t_lower <- vector(mode = "list",length = n.ahead+1)
  H_t_upper <- vector(mode = "list",length = n.ahead+1)
  x_lower <- coef_mat_diagonal(lower_theta, N)
  x_upper <- coef_mat_diagonal(upper_theta, N)
  #check if t(x_lower$C0) or x_lower$C0 for sigma_bekk
  H_t_lower[[1]] <- matrix(sigma_bekk(x$data, t(x_lower$c0), x_lower$a, x_lower$g)$sigma_t[NoBs,], nrow = N, ncol = N)
  H_t_upper[[1]] <- matrix(sigma_bekk(x$data,t(x_upper$c0),x_upper$a, x_upper$g)$sigma_t[NoBs,], nrow = N, ncol = N)
  current_returns <- matrix(c(x$data[NoBs,]), nrow = 1)


  for(i in 1:n.ahead){
    H_t_lower[[i+1]] <- x_lower$c0 %*% t(x_lower$c0) + t(x_lower$a) %*% t(current_returns) %*% current_returns %*% x_lower$a + t(x_lower$g) %*% H_t[[i]] %*% x_lower$g
    current_returns <- eigen_value_decomposition(H_t_lower[[i+1]])
  }

  current_returns <- matrix(c(x$data[NoBs,]), nrow = 1)

  for(i in 1:n.ahead){
    H_t_upper[[i+1]] <- x_upper$c0 %*% t(x_upper$c0) + t(x_upper$a) %*% t(current_returns) %*% current_returns %*% x_upper$a + t(x_upper$g) %*% H_t[[i]] %*% x_upper$g
    current_returns <- eigen_value_decomposition(H_t_upper[[i+1]])
  }
  sigma_t_lower <- matrix(NA, nrow = n.ahead, ncol = N^2)

  for (i in 2:(n.ahead+1)){
    tm2 <- sqrt(solve(diag(diag(H_t_lower[[i]]))))%*%H_t_lower[[i]]%*%sqrt(solve(diag(diag(H_t_lower[[i]]))))
    diag(tm2) <- sqrt(diag(H_t_lower[[i]]))
    sigma_t_lower[i-1,] <- c(tm2)
  }
  sigma_t_upper <- matrix(NA, nrow = n.ahead, ncol = N^2)
  for (i in 2:(n.ahead+1)){
    tm2 <- sqrt(solve(diag(diag(H_t_upper[[i]]))))%*%H_t_upper[[i]]%*%sqrt(solve(diag(diag(H_t_upper[[i]]))))
    diag(tm2) <- sqrt(diag(H_t_upper[[i]]))
    sigma_t_upper[i-1,] <- c(tm2)
  }


  colnames(sigma_t_lower) <- rep(1, N^2)
  colnames(sigma_t_upper) <- rep(1, N^2)

  k <- 1
  k2 <- 1
  for (i in 1:N) {
    for (j in 1:N) {
      if (i == j) {
        colnames(sigma_t_lower)[k2] <- paste('Lower CI conditional standard deviation of \n', colnames(x$data)[k])
        colnames(sigma_t_upper)[k2] <- paste('Upper CI conditional standard deviation of \n', colnames(x$data)[k])
        k <- k + 1
        k2 <- k2 +1
      } else {
        colnames(sigma_t_lower)[k2] <- paste('Lower CI conditional correlation of \n', colnames(x$data)[i], ' and ', colnames(x$data)[j])
        colnames(sigma_t_upper)[k2] <- paste('Upper CI conditional correlation of \n', colnames(x$data)[i], ' and ', colnames(x$data)[j])

        k2 <- k2 +1
      }
    }
  }

  elim <- elimination_mat(N)
  sigma_t_lower <- sigma_t_lower[, which(colSums(elim) == 1)]
  sigma_t_upper <- sigma_t_upper[, which(colSums(elim) == 1)]

  H_t_f_lower <- H_t_f_upper <- matrix(NA, nrow = n.ahead + 1, ncol = ncol(x$data)^2)

  for (i in 1:(n.ahead+1)){
    H_t_f_lower[i, ] <- c(H_t_lower[[i]])
    H_t_f_upper[i, ] <- c(H_t_upper[[i]])
  }

  if (inherits(x$data, "ts")) {
    sigma_t <- ts(sigma_t, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    sigma_t_lower <- ts(sigma_t_lower, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    sigma_t_upper <- ts(sigma_t_upper, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    H_t_f <- ts(H_t_f, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    H_t_f_lower <- ts(H_t_f_lower, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    H_t_f_upper <- ts(H_t_f_upper, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))

  }
  else if(inherits(x$data, "xts") || inherits(x$data, "zoo") ){
    sigma_t <- xts(matrix(sigma_t, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                   by = periodicity(x$data)$units))
    sigma_t_lower <- xts(matrix(sigma_t_lower, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                               by = periodicity(x$data)$units))
    sigma_t_upper <- xts(matrix(sigma_t_upper, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                               by = periodicity(x$data)$units))
    H_t_f <- xts(matrix(H_t_f, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                               by = periodicity(x$data)$units))
    H_t_f_lower <- xts(matrix(H_t_f_lower, nrow = n.ahead+1), order.by = seq((time(x$data)[nrow(x$data)]), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                             by = periodicity(x$data)$units))
    H_t_f_upper <- xts(matrix(H_t_f_upper, nrow = n.ahead+1), order.by = seq((time(x$data)[nrow(x$data)]), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                             by = periodicity(x$data)$units))

  }


  result <- list(
    volatility_forecast = sigma_t,
    volatility_lower_conf_band = sigma_t_lower,
    volatility_upper_conf_band = sigma_t_upper,
    H_t_forecast = H_t_f,
    H_t_lower_conf_band = H_t_f_lower,
    H_t_upper_conf_band = H_t_f_upper,
    n.ahead = n.ahead,
    bekkfit = x
  )
  class(result) <- c('bekkForecast', 'dbekk')
  return(result)
}
#' @rdname predict
#' @export
predict.dbekka <- function(object, n.ahead = 1, ci = 0.95, ...) {
  x <- object
  N <- ncol(x$data)
  NoBs <- nrow(x$data)
  #var_process <- sigma_bekk(xx$data, xx$C0, xx$A, xx$G)
  H_t = vector(mode = "list",length=n.ahead+1)
  H_t[[1]] = matrix(x$H_t[NoBs,],nrow = N, ncol = N)
  current_returns <- matrix(c(x$data[NoBs,]), nrow = 1)

  H_t[[2]]=x$C0 %*% t(x$C0) + t(x$A) %*% t(current_returns) %*% current_returns %*% x$A + indicatorFunction(as.matrix(current_returns), x$signs) * t(x$B) %*% t(current_returns) %*% current_returns %*% x$B + t(x$G) %*% H_t[[1]] %*% x$G

  expected_signs=expected_indicator_value(x$data,x$signs)
  for(i in 2:n.ahead){
    H_t[[i+1]]=x$C0 %*% t(x$C0) + t(x$A) %*% H_t[[i]] %*% x$A + expected_signs * t(x$B) %*% H_t[[i]] %*% x$B + t(x$G) %*% H_t[[i]] %*% x$G

  }

  sigma_t = matrix(NA, nrow = n.ahead, ncol = N^2)
  for (i in 1:(n.ahead+1)){
    tm2 <- sqrt(solve(diag(diag(H_t[[i]]))))%*%H_t[[i]]%*%sqrt(solve(diag(diag(H_t[[i]]))))
    diag(tm2) <- sqrt(diag(H_t[[i]]))
    sigma_t[i-1,] <- c(tm2)
  }

  colnames(sigma_t) <- rep(1, N^2)

  k <- 1
  k2 <- 1
  for (i in 1:N) {
    for (j in 1:N) {
      if (i == j) {
        colnames(sigma_t)[k2] <- paste('Conditional standard deviation of \n', colnames(x$data)[k])
        k <- k + 1
        k2 <- k2 +1
      } else {
        colnames(sigma_t)[k2] <- paste('Conditional correlation of \n', colnames(x$data)[i], ' and ', colnames(x$data)[j])
        k2 <- k2 +1
      }
    }
  }

  elim <- elimination_mat(N)
  sigma_t <- sigma_t[, which(colSums(elim) == 1)]


  H_t_f <- matrix(NA, nrow = n.ahead, ncol = ncol(x$data)^2)

  for (i in 1:(n.ahead+1)){
    H_t_f[i-1, ] <- c(H_t[[i]])
  }


  #95% confidence interval

  score_final = score_asymm_dbekk(x$theta, x$data, x$signs)
  s1_temp = diag(inv_gen(t(score_final) %*% score_final),names=T)
  s1 = sqrt(s1_temp)

  lower_theta = x$theta - qnorm(ci)*s1
  upper_theta = x$theta + qnorm(ci)*s1

  H_t_lower <- vector(mode = "list",length = n.ahead+1)
  H_t_upper <- vector(mode = "list",length = n.ahead+1)
  x_lower <- coef_mat_asymm_diagonal(lower_theta, N)
  x_upper <- coef_mat_asymm_diagonal(upper_theta, N)
  #check if t(x_lower$C0) or x_lower$C0 for sigma_bekk
  H_t_lower[[1]] <- matrix(sigma_bekk_asymm(x$data, t(x_lower$c0), x_lower$a, x_lower$b, x_lower$g, x$signs)$sigma_t[NoBs,],nrow = N, ncol = N)
  H_t_upper[[1]] <- matrix(sigma_bekk_asymm(x$data, t(x_upper$c0), x_upper$a, x_upper$b, x_upper$g, x$signs)$sigma_t[NoBs,],nrow = N, ncol = N)
  current_returns <- matrix(c(x$data[NoBs,]), nrow = 1)


  for(i in 1:n.ahead){
    H_t_lower[[i+1]] <- x_lower$c0 %*% t(x_lower$c0) + t(x_lower$a) %*% t(current_returns) %*% current_returns %*% x_lower$a + expected_signs * t(x_lower$b) %*% H_t_lower[[i]] %*% x_lower$b + t(x_lower$g) %*% H_t[[i]] %*% x_lower$g
    current_returns <- eigen_value_decomposition(H_t_lower[[i+1]])
  }

  current_returns <- matrix(c(x$data[NoBs,]), nrow = 1)

  for(i in 1:n.ahead){
    H_t_upper[[i+1]] <- x_upper$c0 %*% t(x_upper$c0) + t(x_upper$a) %*% t(current_returns) %*% current_returns %*% x_upper$a + expected_signs * t(x_upper$b) %*% H_t_upper[[i]] %*% x_upper$b + t(x_upper$g) %*% H_t[[i]] %*% x_upper$g
    current_returns <- eigen_value_decomposition(H_t_upper[[i+1]])
  }
  sigma_t_lower <- matrix(NA, nrow = n.ahead, ncol = N^2)

  for (i in 2:(n.ahead+1)){
    tm2 <- sqrt(solve(diag(diag(H_t_lower[[i]]))))%*%H_t_lower[[i]]%*%sqrt(solve(diag(diag(H_t_lower[[i]]))))
    diag(tm2) <- sqrt(diag(H_t_lower[[i]]))
    sigma_t_lower[i-1,] <- c(tm2)
  }
  sigma_t_upper <- matrix(NA, nrow = n.ahead, ncol = N^2)
  for (i in 2:(n.ahead+1)){
    tm2 <- sqrt(solve(diag(diag(H_t_upper[[i]]))))%*%H_t_upper[[i]]%*%sqrt(solve(diag(diag(H_t_upper[[i]]))))
    diag(tm2) <- sqrt(diag(H_t_upper[[i]]))
    sigma_t_upper[i-1,] <- c(tm2)
  }


  colnames(sigma_t_lower) <- rep(1, N^2)
  colnames(sigma_t_upper) <- rep(1, N^2)

  k <- 1
  k2 <- 1
  for (i in 1:N) {
    for (j in 1:N) {
      if (i == j) {
        colnames(sigma_t_lower)[k2] <- paste('Lower CI conditional standard deviation of \n', colnames(x$data)[k])
        colnames(sigma_t_upper)[k2] <- paste('Upper CI conditional standard deviation of \n', colnames(x$data)[k])
        k <- k + 1
        k2 <- k2 +1
      } else {
        colnames(sigma_t_lower)[k2] <- paste('Lower CI conditional correlation of \n', colnames(x$data)[i], ' and ', colnames(x$data)[j])
        colnames(sigma_t_upper)[k2] <- paste('Upper CI conditional correlation of \n', colnames(x$data)[i], ' and ', colnames(x$data)[j])

        k2 <- k2 +1
      }
    }
  }

  elim <- elimination_mat(N)
  sigma_t_lower <- sigma_t_lower[, which(colSums(elim) == 1)]
  sigma_t_upper <- sigma_t_upper[, which(colSums(elim) == 1)]

  H_t_f_lower <- H_t_f_upper <- matrix(NA, nrow = n.ahead + 1, ncol = ncol(x$data)^2)

  for (i in 1:(n.ahead+1)){
    H_t_f_lower[i, ] <- c(H_t_lower[[i]])
    H_t_f_upper[i, ] <- c(H_t_upper[[i]])
  }

  if (inherits(x$data, "ts")) {
    sigma_t <- ts(sigma_t, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    sigma_t_lower <- ts(sigma_t_lower, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    sigma_t_upper <- ts(sigma_t_upper, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    H_t_f <- ts(H_t_f, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    H_t_f_lower <- ts(H_t_f_lower, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    H_t_f_upper <- ts(H_t_f_upper, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))

  }
  else if(inherits(x$data, "xts") || inherits(x$data, "zoo") ){
    sigma_t <- xts(matrix(sigma_t, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                   by = periodicity(x$data)$units))
    sigma_t_lower <- xts(matrix(sigma_t_lower, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                               by = periodicity(x$data)$units))
    sigma_t_upper <- xts(matrix(sigma_t_upper, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                               by = periodicity(x$data)$units))
    H_t_f <- xts(matrix(H_t_f, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                               by = periodicity(x$data)$units))
    H_t_f_lower <- xts(matrix(H_t_f_lower, nrow = n.ahead+1), order.by = seq((time(x$data)[nrow(x$data)]), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                             by = periodicity(x$data)$units))
    H_t_f_upper <- xts(matrix(H_t_f_upper, nrow = n.ahead+1), order.by = seq((time(x$data)[nrow(x$data)]), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                             by = periodicity(x$data)$units))

  }


  result <- list(
    volatility_forecast = sigma_t,
    volatility_lower_conf_band = sigma_t_lower,
    volatility_upper_conf_band = sigma_t_upper,
    H_t_forecast = H_t_f,
    H_t_lower_conf_band = H_t_f_lower,
    H_t_upper_conf_band = H_t_f_upper,
    n.ahead = n.ahead,
    bekkfit = x
  )
  class(result) <-  c('bekkForecast', 'dbekka')
  return(result)
}
#' @rdname predict
#' @export
predict.sbekk <- function(object, n.ahead = 1, ci = 0.95, ...) {
  x <- object
  N <- ncol(x$data)
  NoBs <- nrow(x$data)
  #var_process <- sigma_bekk(xx$data, xx$C0, xx$A, xx$G)
  H_t <- vector(mode = "list",length = n.ahead+1)
  H_t[[1]] <- matrix(x$H_t[NoBs,],nrow = N, ncol = N)
  current_returns <- matrix(c(x$data[NoBs,]), nrow = 1)

  for(i in 1:n.ahead){
    H_t[[i+1]] <- x$C0 %*% t(x$C0) + x$a * t(current_returns) %*% current_returns+ x$g * H_t[[i]]
    current_returns <- eigen_value_decomposition(H_t[[i+1]])
  }

  sigma_t <- matrix(NA, nrow = n.ahead, ncol = N^2)
  for (i in 2:(n.ahead+1)){
    tm2 <- sqrt(solve(diag(diag(H_t[[i]]))))%*%H_t[[i]]%*%sqrt(solve(diag(diag(H_t[[i]]))))
    diag(tm2) <- sqrt(diag(H_t[[i]]))
    sigma_t[i-1,] <- c(tm2)
  }

  colnames(sigma_t) <- rep(1, N^2)

  k <- 1
  k2 <- 1
  for (i in 1:N) {
    for (j in 1:N) {
      if (i == j) {
        colnames(sigma_t)[k2] <- paste('Conditional standard deviation of \n', colnames(x$data)[k])
        k <- k + 1
        k2 <- k2 +1
      } else {
        colnames(sigma_t)[k2] <- paste('Conditional correlation of \n', colnames(x$data)[i], ' and ', colnames(x$data)[j])
        k2 <- k2 +1
      }
    }
  }

  elim <- elimination_mat(N)
  sigma_t <- sigma_t[, which(colSums(elim) == 1)]

  H_t_f <- matrix(NA, nrow = n.ahead, ncol = ncol(x$data)^2)

  for (i in 2:(n.ahead+1)){
    H_t_f[i-1, ] <- c(H_t[[i]])
  }

  # Generating confidence intervals

  score_final = score_sbekk(x$theta, x$data)
  s1_temp = diag(inv_gen(t(score_final) %*% score_final),names=T)
  s1 = sqrt(s1_temp)

  lower_theta = x$theta - qnorm(ci)*s1
  upper_theta = x$theta + qnorm(ci)*s1

  H_t_lower <- vector(mode = "list",length = n.ahead+1)
  H_t_upper <- vector(mode = "list",length = n.ahead+1)
  x_lower <- coef_mat_scalar(lower_theta, N)
  x_upper <- coef_mat_scalar(upper_theta, N)
  #check if t(x_lower$C0) or x_lower$C0 for sigma_bekk
  H_t_lower[[1]] <- matrix(sigma_sbekk(x$data, t(x_lower$c0), x_lower$a, x_lower$g)$sigma_t[NoBs,], nrow = N, ncol = N)
  H_t_upper[[1]] <- matrix(sigma_sbekk(x$data,t(x_upper$c0),x_upper$a, x_upper$g)$sigma_t[NoBs,], nrow = N, ncol = N)
  current_returns <- matrix(c(x$data[NoBs,]), nrow = 1)


  for(i in 1:n.ahead){
    H_t_lower[[i+1]] <- x_lower$c0 %*% t(x_lower$c0) + x_lower$a * t(current_returns) %*% current_returns + x_lower$g * H_t[[i]]
    current_returns <- eigen_value_decomposition(H_t_lower[[i+1]])
  }

  current_returns <- matrix(c(x$data[NoBs,]), nrow = 1)

  for(i in 1:n.ahead){
    H_t_upper[[i+1]] <- x_upper$c0 %*% t(x_upper$c0) + x_upper$a * t(current_returns) %*% current_returns + x_upper$g * H_t[[i]]
    current_returns <- eigen_value_decomposition(H_t_upper[[i+1]])
  }
  sigma_t_lower <- matrix(NA, nrow = n.ahead, ncol = N^2)

  for (i in 2:(n.ahead+1)){
    tm2 <- sqrt(solve(diag(diag(H_t_lower[[i]]))))%*%H_t_lower[[i]]%*%sqrt(solve(diag(diag(H_t_lower[[i]]))))
    diag(tm2) <- sqrt(diag(H_t_lower[[i]]))
    sigma_t_lower[i-1,] <- c(tm2)
  }
  sigma_t_upper <- matrix(NA, nrow = n.ahead, ncol = N^2)
  for (i in 2:(n.ahead+1)){
    tm2 <- sqrt(solve(diag(diag(H_t_upper[[i]]))))%*%H_t_upper[[i]]%*%sqrt(solve(diag(diag(H_t_upper[[i]]))))
    diag(tm2) <- sqrt(diag(H_t_upper[[i]]))
    sigma_t_upper[i-1,] <- c(tm2)
  }


  colnames(sigma_t_lower) <- rep(1, N^2)
  colnames(sigma_t_upper) <- rep(1, N^2)

  k <- 1
  k2 <- 1
  for (i in 1:N) {
    for (j in 1:N) {
      if (i == j) {
        colnames(sigma_t_lower)[k2] <- paste('Lower CI conditional standard deviation of \n', colnames(x$data)[k])
        colnames(sigma_t_upper)[k2] <- paste('Upper CI conditional standard deviation of \n', colnames(x$data)[k])
        k <- k + 1
        k2 <- k2 +1
      } else {
        colnames(sigma_t_lower)[k2] <- paste('Lower CI conditional correlation of \n', colnames(x$data)[i], ' and ', colnames(x$data)[j])
        colnames(sigma_t_upper)[k2] <- paste('Upper CI conditional correlation of \n', colnames(x$data)[i], ' and ', colnames(x$data)[j])

        k2 <- k2 +1
      }
    }
  }

  elim <- elimination_mat(N)
  sigma_t_lower <- sigma_t_lower[, which(colSums(elim) == 1)]
  sigma_t_upper <- sigma_t_upper[, which(colSums(elim) == 1)]

  H_t_f_lower <- H_t_f_upper <- matrix(NA, nrow = n.ahead + 1, ncol = ncol(x$data)^2)

  for (i in 1:(n.ahead+1)){
    H_t_f_lower[i, ] <- c(H_t_lower[[i]])
    H_t_f_upper[i, ] <- c(H_t_upper[[i]])
  }

  if (inherits(x$data, "ts")) {
    sigma_t <- ts(sigma_t, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    sigma_t_lower <- ts(sigma_t_lower, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    sigma_t_upper <- ts(sigma_t_upper, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    H_t_f <- ts(H_t_f, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    H_t_f_lower <- ts(H_t_f_lower, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    H_t_f_upper <- ts(H_t_f_upper, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))

  }
  else if(inherits(x$data, "xts") || inherits(x$data, "zoo") ){
    sigma_t <- xts(matrix(sigma_t, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                   by = periodicity(x$data)$units))
    sigma_t_lower <- xts(matrix(sigma_t_lower, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                               by = periodicity(x$data)$units))
    sigma_t_upper <- xts(matrix(sigma_t_upper, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                               by = periodicity(x$data)$units))
    H_t_f <- xts(matrix(H_t_f, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                               by = periodicity(x$data)$units))
    H_t_f_lower <- xts(matrix(H_t_f_lower, nrow = n.ahead+1), order.by = seq((time(x$data)[nrow(x$data)]), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                             by = periodicity(x$data)$units))
    H_t_f_upper <- xts(matrix(H_t_f_upper, nrow = n.ahead+1), order.by = seq((time(x$data)[nrow(x$data)]), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                             by = periodicity(x$data)$units))

  }


  result <- list(
    volatility_forecast = sigma_t,
    volatility_lower_conf_band = sigma_t_lower,
    volatility_upper_conf_band = sigma_t_upper,
    H_t_forecast = H_t_f,
    H_t_lower_conf_band = H_t_f_lower,
    H_t_upper_conf_band = H_t_f_upper,
    n.ahead = n.ahead,
    bekkfit = x
  )
  class(result) <- c('bekkForecast', 'sbekk')
  return(result)
}

#' @rdname predict
#' @export
predict.sbekka <- function(object, n.ahead = 1, ci = 0.95, ...) {
  x <- object
  N <- ncol(x$data)
  NoBs <- nrow(x$data)
  #var_process <- sigma_bekk(xx$data, xx$C0, xx$A, xx$G)
  H_t = vector(mode = "list",length=n.ahead+1)
  H_t[[1]] = matrix(x$H_t[NoBs,],nrow = N, ncol = N)
  current_returns <- matrix(c(x$data[NoBs,]), nrow = 1)

  H_t[[2]]=x$C0 %*% t(x$C0) + x$a * t(current_returns) %*% current_returns + indicatorFunction(as.matrix(current_returns), x$signs) * x$b * t(current_returns) %*% current_returns + x$g * H_t[[1]]

  expected_signs=expected_indicator_value(x$data,x$signs)
  for(i in 2:n.ahead){
    H_t[[i+1]]=x$C0 %*% t(x$C0) + x$a * H_t[[i]]  + expected_signs * x$b * H_t[[i]]  + x$g * H_t[[i]]

  }

  sigma_t = matrix(NA, nrow = n.ahead, ncol = N^2)
  for (i in 1:(n.ahead+1)){
    tm2 <- sqrt(solve(diag(diag(H_t[[i]]))))%*%H_t[[i]]%*%sqrt(solve(diag(diag(H_t[[i]]))))
    diag(tm2) <- sqrt(diag(H_t[[i]]))
    sigma_t[i-1,] <- c(tm2)
  }

  colnames(sigma_t) <- rep(1, N^2)

  k <- 1
  k2 <- 1
  for (i in 1:N) {
    for (j in 1:N) {
      if (i == j) {
        colnames(sigma_t)[k2] <- paste('Conditional standard deviation of \n', colnames(x$data)[k])
        k <- k + 1
        k2 <- k2 +1
      } else {
        colnames(sigma_t)[k2] <- paste('Conditional correlation of \n', colnames(x$data)[i], ' and ', colnames(x$data)[j])
        k2 <- k2 +1
      }
    }
  }

  elim <- elimination_mat(N)
  sigma_t <- sigma_t[, which(colSums(elim) == 1)]


  H_t_f <- matrix(NA, nrow = n.ahead, ncol = ncol(x$data)^2)

  for (i in 1:(n.ahead+1)){
    H_t_f[i-1, ] <- c(H_t[[i]])
  }


  #95% confidence interval

  score_final = score_asymm_sbekk(x$theta, x$data, x$signs)
  s1_temp = diag(inv_gen(t(score_final) %*% score_final),names=T)
  s1 = sqrt(s1_temp)

  lower_theta = x$theta - qnorm(ci)*s1
  upper_theta = x$theta + qnorm(ci)*s1

  H_t_lower <- vector(mode = "list",length = n.ahead+1)
  H_t_upper <- vector(mode = "list",length = n.ahead+1)
  x_lower <- coef_mat_asymm_scalar(lower_theta, N)
  x_upper <- coef_mat_asymm_scalar(upper_theta, N)
  #check if t(x_lower$C0) or x_lower$C0 for sigma_bekk
  H_t_lower[[1]] <- matrix(sigma_sbekk_asymm(x$data, t(x_lower$c0), x_lower$a, x_lower$b, x_lower$g, x$signs)$sigma_t[NoBs,],nrow = N, ncol = N)
  H_t_upper[[1]] <- matrix(sigma_sbekk_asymm(x$data, t(x_upper$c0), x_upper$a, x_upper$b, x_upper$g, x$signs)$sigma_t[NoBs,],nrow = N, ncol = N)
  current_returns <- matrix(c(x$data[NoBs,]), nrow = 1)


  for(i in 1:n.ahead){
    H_t_lower[[i+1]] <- x_lower$c0 %*% t(x_lower$c0) + x_lower$a * t(current_returns) %*% current_returns + expected_signs * x_lower$b * H_t_lower[[i]]  + x_lower$g * H_t[[i]]
    current_returns <- eigen_value_decomposition(H_t_lower[[i+1]])
  }

  current_returns <- matrix(c(x$data[NoBs,]), nrow = 1)

  for(i in 1:n.ahead){
    H_t_upper[[i+1]] <- x_upper$c0 %*% t(x_upper$c0) + x_upper$a * t(current_returns) %*% current_returns + expected_signs * x_upper$b * H_t_upper[[i]] + x_upper$g * H_t[[i]]
    current_returns <- eigen_value_decomposition(H_t_upper[[i+1]])
  }
  sigma_t_lower <- matrix(NA, nrow = n.ahead, ncol = N^2)

  for (i in 2:(n.ahead+1)){
    tm2 <- sqrt(solve(diag(diag(H_t_lower[[i]]))))%*%H_t_lower[[i]]%*%sqrt(solve(diag(diag(H_t_lower[[i]]))))
    diag(tm2) <- sqrt(diag(H_t_lower[[i]]))
    sigma_t_lower[i-1,] <- c(tm2)
  }
  sigma_t_upper <- matrix(NA, nrow = n.ahead, ncol = N^2)
  for (i in 2:(n.ahead+1)){
    tm2 <- sqrt(solve(diag(diag(H_t_upper[[i]]))))%*%H_t_upper[[i]]%*%sqrt(solve(diag(diag(H_t_upper[[i]]))))
    diag(tm2) <- sqrt(diag(H_t_upper[[i]]))
    sigma_t_upper[i-1,] <- c(tm2)
  }


  colnames(sigma_t_lower) <- rep(1, N^2)
  colnames(sigma_t_upper) <- rep(1, N^2)

  k <- 1
  k2 <- 1
  for (i in 1:N) {
    for (j in 1:N) {
      if (i == j) {
        colnames(sigma_t_lower)[k2] <- paste('Lower CI conditional standard deviation of \n', colnames(x$data)[k])
        colnames(sigma_t_upper)[k2] <- paste('Upper CI conditional standard deviation of \n', colnames(x$data)[k])
        k <- k + 1
        k2 <- k2 +1
      } else {
        colnames(sigma_t_lower)[k2] <- paste('Lower CI conditional correlation of \n', colnames(x$data)[i], ' and ', colnames(x$data)[j])
        colnames(sigma_t_upper)[k2] <- paste('Upper CI conditional correlation of \n', colnames(x$data)[i], ' and ', colnames(x$data)[j])

        k2 <- k2 +1
      }
    }
  }

  elim <- elimination_mat(N)
  sigma_t_lower <- sigma_t_lower[, which(colSums(elim) == 1)]
  sigma_t_upper <- sigma_t_upper[, which(colSums(elim) == 1)]

  H_t_f_lower <- H_t_f_upper <- matrix(NA, nrow = n.ahead + 1, ncol = ncol(x$data)^2)

  for (i in 1:(n.ahead+1)){
    H_t_f_lower[i, ] <- c(H_t_lower[[i]])
    H_t_f_upper[i, ] <- c(H_t_upper[[i]])
  }

  if (inherits(x$data, "ts")) {
    sigma_t <- ts(sigma_t, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    sigma_t_lower <- ts(sigma_t_lower, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    sigma_t_upper <- ts(sigma_t_upper, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    H_t_f <- ts(H_t_f, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    H_t_f_lower <- ts(H_t_f_lower, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))
    H_t_f_upper <- ts(H_t_f_upper, start = (time(x$data)[nrow(x$data)]+1), frequency = frequency(x$data))

  }
  else if(inherits(x$data, "xts") || inherits(x$data, "zoo") ){
    sigma_t <- xts(matrix(sigma_t, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                   by = periodicity(x$data)$units))
    sigma_t_lower <- xts(matrix(sigma_t_lower, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                               by = periodicity(x$data)$units))
    sigma_t_upper <- xts(matrix(sigma_t_upper, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                               by = periodicity(x$data)$units))
    H_t_f <- xts(matrix(H_t_f, nrow = n.ahead), order.by = seq((time(x$data)[nrow(x$data)]+1), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                               by = periodicity(x$data)$units))
    H_t_f_lower <- xts(matrix(H_t_f_lower, nrow = n.ahead+1), order.by = seq((time(x$data)[nrow(x$data)]), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                             by = periodicity(x$data)$units))
    H_t_f_upper <- xts(matrix(H_t_f_upper, nrow = n.ahead+1), order.by = seq((time(x$data)[nrow(x$data)]), (time(x$data)[nrow(x$data)] +  n.ahead),
                                                                             by = periodicity(x$data)$units))

  }


  result <- list(
    volatility_forecast = sigma_t,
    volatility_lower_conf_band = sigma_t_lower,
    volatility_upper_conf_band = sigma_t_upper,
    H_t_forecast = H_t_f,
    H_t_lower_conf_band = H_t_f_lower,
    H_t_upper_conf_band = H_t_f_upper,
    n.ahead = n.ahead,
    bekkfit = x
  )
  class(result) <-  c('bekkForecast', 'sbekka')
  return(result)
}

#' BEKK specification method
#'
#' @description Method for creating a N-dimensional BEKK model specification object prior to fitting and/or simulating.
#'
#' @param model A list containing the model type specification: Either "bekk" "dbekk" or "sbekk".
#' Moreover it can be specified whether the model should be estimated allowing for asymmetric volatility structure.
#' @param init_values initial values for \link{bekk_fit} during BHHH algorithm. It can be either a numerical vector of suitable dimension, 'NULL' (default) to use a simple grid search algorithm, or a character vector i.e. "random" to use a random starting value generator (set a seed in advance for reproducible results), or
#'  "simple" for relying on a simple initial values generator based on typical values for BEKK parameter found in the literature. If the object from this function is passed to \link{simulate}, "init_values" are used as parameters for data generating process.
#' @param signs An N-dimensional vector consisting of "1" or "-1" to indicate the asymmetric effects to be considered.
#' Setting the i-th element of the vector to "1" or "-1" means that the model takes into account additional volatility if the returns of the i-th column in the data matrix are either positive or negative.
#' If "asymmetric = TRUE", the default is set to "rep(-1, N)" i.e. it is assumed that excess volatility occurs for all series if the returns are negative.
#' @param N Integer specifying the dimension of the BEKK model. Only relevant when this object of class "bekkSpec"" is used for simulating BEKK processes by applying it to \link{simulate}.
#' @return Returns a S3 class "bekkSpec"  object containing the specifications of the model to be estimated.
#'
#'
#' @examples
#' \donttest{
#'
#' data(StocksBonds)
#'
#' # Fitting a symmetric BEKK model using default starting values
#' # - i.e. fixed values
#' obj_spec_fixed <- bekk_spec(init_values = NULL)
#' x1 <- bekk_fit(obj_spec_fixed, StocksBonds, QML_t_ratios = FALSE,
#' max_iter = 50, crit = 1e-9)
#'# Fitting a symmetric BEKK model using initial values originating from a
#'# random grid search algorithm
#' obj_spec_random <- bekk_spec(init_values = "random")
#' x2 <- bekk_fit(obj_spec_random, StocksBonds, QML_t_ratios = FALSE,
#' max_iter = 50, crit = 1e-9)
#' summary(x1)
#' summary(x2)
#' plot(x1)
#' plot(x2)
#' # Fitting an asymmetric BEKK model with default starting values
#' obj_spec_fix <- bekk_spec(model = list(type = "bekk", asymmetric = TRUE),
#' init_values = NULL)
#' x1 <- bekk_fit(obj_spec_fix, StocksBonds)
#' obj_spec_random <- bekk_spec(model = list(type = "bekk", asymmetric = TRUE),
#' init_values = "random")
#' x2 <- bekk_fit(obj_spec_random, StocksBonds)
#' summary(x1)
#' summary(x2)
#' }
#' @export
bekk_spec <- function(model = list(type = "bekk", asymmetric = FALSE),
                      init_values = NULL, signs = NULL, N = NULL) {

  if(!is.logical(model$asymmetric) || is.null(model$asymmetric)){
    stop('Please specify whether the model to be estimated is asymmetric or not.')

  }
  # Checking type
  match.arg(model$type, c("bekk", "dbekk", "sbekk"))
  # Checking inputs
  if(!is.null(N) & is.numeric(init_values) ) {
    if(nrow(init_values) != 2 * N^2 + N * (N + 1)/2 & model$type == "bekk"  & model$asymmetric == FALSE) {
      stop('Number of initial parameter does not match dimension of data and model.')
    }
    if(nrow(init_values) != 3 * N^2 + N * (N + 1)/2 & model$type == "bekk"  & model$asymmetric == TRUE) {
      stop('Number of initial parameter does not match dimension of data and model.')
    }
    if(nrow(init_values) != 2 * N + N * (N + 1)/2 & model$type == "dbekk"  & model$asymmetric == FALSE) {
      stop('Number of initial parameter does not match dimension of data and model.')
    }
    if(nrow(init_values) != 3 * N + N * (N + 1)/2 & model$type == "dbekk"  & model$asymmetric == TRUE) {
      stop('Number of initial parameter does not match dimension of data and model.')
    }
    if(nrow(init_values) != 2 + N * (N + 1)/2 & model$type == "sbekk"  & model$asymmetric == FALSE) {
      stop('Number of initial parameter does not match dimension of data and model.')
    }
    if(nrow(init_values) != 3 + N * (N + 1)/2 & model$type == "sbekk"  & model$asymmetric == TRUE) {
      stop('Number of initial parameter does not match dimension of data and model.')
    }
  }

  specification <- list(model = model, init_values = init_values, signs = signs, N = N)
  class(specification) <- 'bekkSpec'
  class(specification)[2] <- model$type

  if (model$asymmetric == TRUE) {

    if (!is.null(signs)) {
      signs <- matrix(signs, ncol = 1)
      if(any(abs(signs) != 1)){
        stop('Elements of "signs" must be either "1" or "-1".')
      }
    }

    specification$model$signs <- signs
    class(specification)[2] <- paste(class(specification)[2], 'a', sep = "")
  }

  return(specification)
}

#' @import ggplot2
#' @import ggfortify
#' @import reshape2
#' @export
#'
plot.backtest <- function(x, ...) {

  obs <- NULL
  type <- NULL
  time <- NULL
  V1 <- NULL
  Portfolio <- NULL


    if(is.null(x$portfolio_weights)) {
      if (inherits(x$VaR, c("ts"))) {
        colnames(x$VaR) = colnames(x$out_sample_returns)

        t_series <- as.data.frame(x$VaR)
        t_series$time <- time(x$VaR)
        t_series <- melt(t_series, id ="time")

        out_sample_returns <- as.data.frame(x$out_sample_returns)
        out_sample_returns$time <- time(x$out_sample_returns)
        out_sample_returns <- melt(out_sample_returns, id ="time")

        ggplot(t_series) + geom_point(data = out_sample_returns, mapping = aes(x = time, y = value, colour = "Returns"),  show.legend = TRUE) + theme_bw() + xlab('') + ylab('Returns/VaR') +
          geom_line(aes(x = time, y = value, colour = "Estimated VaR")) + scale_color_manual(values = c('black', 'blue'), "") +
          facet_wrap(~variable, scales = 'free_y', ncol = 1)+
          theme(legend.position="bottom", legend.title = element_blank())



      }else if (inherits(x$VaR, c("xts","zoo"))) {
        names(x$VaR) = names(x$out_sample_returns)

        t_series <- as.data.frame(x$VaR)
        t_series$time <- time(x$VaR)
        t_series <- melt(t_series, id ="time")

        out_sample_returns <- as.data.frame(x$out_sample_returns)
        out_sample_returns$time <- time(x$out_sample_returns)
        out_sample_returns <- melt(out_sample_returns, id ="time")


        ggplot(t_series) + geom_point(data = out_sample_returns, mapping = aes(x = time, y = value, colour = "Returns"),  show.legend = TRUE) + theme_bw() + xlab('') + ylab('Returns/VaR') +
          geom_line(aes(x = time, y = value, colour = "Estimated VaR")) + scale_color_manual(values = c('black', 'blue'), "") +
          facet_wrap(~variable, scales = 'free_y', ncol = 1)+
          theme(legend.position="bottom", legend.title = element_blank())

        } else {

        names(x$VaR) = names(x$out_sample_returns)

        x$VaR$obs <- 1:nrow(x$VaR)
        VaR <- melt(x$VaR, id = 'obs')
        x$out_sample_returns$obs <- 1:nrow(x$out_sample_returns)



        out_sample_returns <- melt(x$out_sample_returns, id = 'obs')


        ggplot(VaR) + geom_point(data = out_sample_returns, mapping = aes(x = obs, y = value, colour = "Returns"),  show.legend = TRUE) + theme_bw() + xlab('') + ylab('VaR') +
          geom_line(aes(x = obs, y = value, colour = "Estimated VaR")) + scale_color_manual(values = c('black', 'blue'), "") +
          facet_wrap(~variable, scales = 'free_y', ncol = 1)+
          theme(legend.position="bottom", legend.title = element_blank())
    }
      }else {
      if (inherits(x$VaR, c("xts","zoo"))) {

        ggplot(data=x$VaR) + geom_point(data=x$out_sample_returns, mapping = aes(x = time(x$VaR), y = Portfolio, colour="Returns")) +
          geom_line(x$VaR , mapping = aes(x = time(x$VaR), y = Portfolio, colour="Estimated VaR")) + theme_bw() + xlab("") + ylab('Portfolio returns/VaR') + ggtitle('Portfolio Backtest')+ scale_color_manual(values = c('blue', 'black'), "") +
          theme(legend.position="bottom", legend.title = element_blank())
      } else {
        ggplot(x$VaR) + geom_point(data = x$out_sample_returns,
                                                                              mapping = aes(x = 1:nrow(x$VaR), y = Portfolio, colour="Returns")) + theme_bw() + xlab('') + ylab('Returns/VaR') +
        geom_line(aes(x = 1:nrow(x$VaR), y = Portfolio, colour="Estimated VaR")) + ggtitle('Portfolio Backtest')+ scale_color_manual(values = c('blue', 'black'), "") +
          theme(legend.position="bottom", legend.title = element_blank())

      }
    }



}

#' @import ggplot2
#' @import grid
#' @import gridExtra
#' @export

plot.virf <- function(x, ...){

  V1 <- NULL
  l <- NULL
  lower <- NULL
  upper <- NULL

    trianglePlotGrid <- function(plots){
      #take a list of plots and returns a single plot where the elements in the list arranged in a triangular grid
      #plots should be a list of 1 or 3 or 6... plots to be arranged in a trianglular structure with 1 plot in the top row
      ncols <- (-1 + sqrt(1 + 8*length(plots)))/2

      grobs <- list()

      k <-1

      for (i in 1:ncols) {
        for (j in 1:ncols) {
          if (i <= j) {
            grobs[[length(grobs)+1]] <- plots[[k]]
            k <- k+1
          } else {
            grobs[[length(grobs)+1]] <- nullGrob()
          }
        }
      }
      do.call("grid.arrange", c(grobs, ncol=ncols))
    }





      xxc <- colnames(x$VIRF)

      plist <- vector(mode = "list", length = ncol(x$VIRF))

      for (i in 1:ncol(x$VIRF)) {
        xx1 <- data.frame(x$VIRF[,i])
        xxci <- data.frame(x$VIRF_lower[,i],x$VIRF_upper[,i])
        colnames(xx1) <- 'V1'
        colnames(xxci) <- c("lower","upper")

        if (grepl('correlation', xxc[i])) {
          plist[[i]] <- ggplot(xx1, aes(x = 1:nrow(x$VIRF), y = V1)) + geom_line() + geom_ribbon(data=xxci,aes(ymin=lower,ymax=upper),alpha=0.3) + theme_bw() + xlab('') + ylab('') + geom_hline(yintercept = 0, col = 'red')
        } else {
          plist[[i]] <- ggplot(xx1, aes(x = 1:nrow(x$VIRF), y = V1)) + geom_line() + geom_ribbon(data=xxci,aes(ymin=lower,ymax=upper),alpha=0.3) + theme_bw()+ xlab('') + ylab('') + geom_hline(yintercept = 0, col = 'red')
        }
      }



    for (i in 1:ncol(x$VIRF)) {
      plist[[i]] <- plist[[i]] + ggtitle(xxc[i])
    }

    trianglePlotGrid(plist)


}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

diag_selection_mat <- function(n) {
    .Call('_BEKKs_diag_selection_mat', PACKAGE = 'BEKKs', n)
}

cut_mat_symmetric <- function(n) {
    .Call('_BEKKs_cut_mat_symmetric', PACKAGE = 'BEKKs', n)
}

cut_mat_asymmetric <- function(n) {
    .Call('_BEKKs_cut_mat_asymmetric', PACKAGE = 'BEKKs', n)
}

set_seed <- function(seed) {
    invisible(.Call('_BEKKs_set_seed', PACKAGE = 'BEKKs', seed))
}

valid_bekk <- function(C, A, G) {
    .Call('_BEKKs_valid_bekk', PACKAGE = 'BEKKs', C, A, G)
}

valid_asymm_bekk <- function(C, A, B, G, r, signs) {
    .Call('_BEKKs_valid_asymm_bekk', PACKAGE = 'BEKKs', C, A, B, G, r, signs)
}

valid_asymm_bekk_sim <- function(C, A, B, G, exp_indicator_value, signs) {
    .Call('_BEKKs_valid_asymm_bekk_sim', PACKAGE = 'BEKKs', C, A, B, G, exp_indicator_value, signs)
}

loglike_bekk <- function(theta, r) {
    .Call('_BEKKs_loglike_bekk', PACKAGE = 'BEKKs', theta, r)
}

loglike_asymm_bekk <- function(theta, r, signs) {
    .Call('_BEKKs_loglike_asymm_bekk', PACKAGE = 'BEKKs', theta, r, signs)
}

loglike_dbekk <- function(theta, r) {
    .Call('_BEKKs_loglike_dbekk', PACKAGE = 'BEKKs', theta, r)
}

loglike_asymm_dbekk <- function(theta, r, signs) {
    .Call('_BEKKs_loglike_asymm_dbekk', PACKAGE = 'BEKKs', theta, r, signs)
}

score_bekk <- function(theta, r) {
    .Call('_BEKKs_score_bekk', PACKAGE = 'BEKKs', theta, r)
}

score_dbekk <- function(theta, r) {
    .Call('_BEKKs_score_dbekk', PACKAGE = 'BEKKs', theta, r)
}

score_asymm_bekk <- function(theta, r, signs) {
    .Call('_BEKKs_score_asymm_bekk', PACKAGE = 'BEKKs', theta, r, signs)
}

score_asymm_dbekk <- function(theta, r, signs) {
    .Call('_BEKKs_score_asymm_dbekk', PACKAGE = 'BEKKs', theta, r, signs)
}

bhh_bekk <- function(r, theta, max_iter, crit) {
    .Call('_BEKKs_bhh_bekk', PACKAGE = 'BEKKs', r, theta, max_iter, crit)
}

bhh_asymm_bekk <- function(r, theta, max_iter, crit, signs) {
    .Call('_BEKKs_bhh_asymm_bekk', PACKAGE = 'BEKKs', r, theta, max_iter, crit, signs)
}

bhh_dbekk <- function(r, theta, max_iter, crit) {
    .Call('_BEKKs_bhh_dbekk', PACKAGE = 'BEKKs', r, theta, max_iter, crit)
}

bhh_asymm_dbekk <- function(r, theta, max_iter, crit, signs) {
    .Call('_BEKKs_bhh_asymm_dbekk', PACKAGE = 'BEKKs', r, theta, max_iter, crit, signs)
}

random_grid_search_BEKK <- function(r) {
    .Call('_BEKKs_random_grid_search_BEKK', PACKAGE = 'BEKKs', r)
}

random_grid_search_asymmetric_BEKK <- function(r, signs) {
    .Call('_BEKKs_random_grid_search_asymmetric_BEKK', PACKAGE = 'BEKKs', r, signs)
}

random_grid_search_dBEKK <- function(r) {
    .Call('_BEKKs_random_grid_search_dBEKK', PACKAGE = 'BEKKs', r)
}

random_grid_search_asymmetric_dBEKK <- function(r, signs) {
    .Call('_BEKKs_random_grid_search_asymmetric_dBEKK', PACKAGE = 'BEKKs', r, signs)
}

sigma_bekk <- function(r, C, A, G) {
    .Call('_BEKKs_sigma_bekk', PACKAGE = 'BEKKs', r, C, A, G)
}

sigma_bekk_asymm <- function(r, C, A, B, G, signs) {
    .Call('_BEKKs_sigma_bekk_asymm', PACKAGE = 'BEKKs', r, C, A, B, G, signs)
}

hesse_bekk <- function(theta, r) {
    .Call('_BEKKs_hesse_bekk', PACKAGE = 'BEKKs', theta, r)
}

hesse_dbekk <- function(theta, r) {
    .Call('_BEKKs_hesse_dbekk', PACKAGE = 'BEKKs', theta, r)
}

hesse_asymm_bekk <- function(theta, r, signs) {
    .Call('_BEKKs_hesse_asymm_bekk', PACKAGE = 'BEKKs', theta, r, signs)
}

hesse_asymm_dbekk <- function(theta, r, signs) {
    .Call('_BEKKs_hesse_asymm_dbekk', PACKAGE = 'BEKKs', theta, r, signs)
}

eigen_value_decomposition <- function(A) {
    .Call('_BEKKs_eigen_value_decomposition', PACKAGE = 'BEKKs', A)
}

virf_bekk <- function(H_t, theta, shocks, periods) {
    .Call('_BEKKs_virf_bekk', PACKAGE = 'BEKKs', H_t, theta, shocks, periods)
}

virf_dbekk <- function(H_t, theta, shocks, periods) {
    .Call('_BEKKs_virf_dbekk', PACKAGE = 'BEKKs', H_t, theta, shocks, periods)
}

virf_bekka <- function(H_t, C, A, B, G, signs, expected_signs, shocks, periods) {
    .Call('_BEKKs_virf_bekka', PACKAGE = 'BEKKs', H_t, C, A, B, G, signs, expected_signs, shocks, periods)
}

virf_sbekk <- function(H_t, theta, shocks, periods) {
    .Call('_BEKKs_virf_sbekk', PACKAGE = 'BEKKs', H_t, theta, shocks, periods)
}

simulate_bekk_c <- function(theta, NoObs, n) {
    .Call('_BEKKs_simulate_bekk_c', PACKAGE = 'BEKKs', theta, NoObs, n)
}

simulate_bekka_c <- function(theta, NoObs, n, signs, expected_signs) {
    .Call('_BEKKs_simulate_bekka_c', PACKAGE = 'BEKKs', theta, NoObs, n, signs, expected_signs)
}

simulate_dbekk_c <- function(theta, NoObs, N) {
    .Call('_BEKKs_simulate_dbekk_c', PACKAGE = 'BEKKs', theta, NoObs, N)
}

simulate_dbekka_c <- function(theta, NoObs, N, signs, expected_signs) {
    .Call('_BEKKs_simulate_dbekka_c', PACKAGE = 'BEKKs', theta, NoObs, N, signs, expected_signs)
}

simulate_sbekk_c <- function(theta, NoObs, N) {
    .Call('_BEKKs_simulate_sbekk_c', PACKAGE = 'BEKKs', theta, NoObs, N)
}

simulate_sbekka_c <- function(theta, NoObs, N, signs, expected_signs) {
    .Call('_BEKKs_simulate_sbekka_c', PACKAGE = 'BEKKs', theta, NoObs, N, signs, expected_signs)
}

indicatorFunction <- function(r, signs) {
    .Call('_BEKKs_indicatorFunction', PACKAGE = 'BEKKs', r, signs)
}

expected_indicator_value <- function(r, signs) {
    .Call('_BEKKs_expected_indicator_value', PACKAGE = 'BEKKs', r, signs)
}

elimination_mat <- function(n) {
    .Call('_BEKKs_elimination_mat', PACKAGE = 'BEKKs', n)
}

commutation_mat <- function(n) {
    .Call('_BEKKs_commutation_mat', PACKAGE = 'BEKKs', n)
}

duplication_mat <- function(n) {
    .Call('_BEKKs_duplication_mat', PACKAGE = 'BEKKs', n)
}

inv_gen <- function(m) {
    .Call('_BEKKs_inv_gen', PACKAGE = 'BEKKs', m)
}

valid_sbekk <- function(C, a, g) {
    .Call('_BEKKs_valid_sbekk', PACKAGE = 'BEKKs', C, a, g)
}

valid_asymm_sbekk <- function(C, a, b, g, r, signs) {
    .Call('_BEKKs_valid_asymm_sbekk', PACKAGE = 'BEKKs', C, a, b, g, r, signs)
}

loglike_sbekk <- function(theta, r) {
    .Call('_BEKKs_loglike_sbekk', PACKAGE = 'BEKKs', theta, r)
}

loglike_asymm_sbekk <- function(theta, r, signs) {
    .Call('_BEKKs_loglike_asymm_sbekk', PACKAGE = 'BEKKs', theta, r, signs)
}

score_sbekk <- function(theta, r) {
    .Call('_BEKKs_score_sbekk', PACKAGE = 'BEKKs', theta, r)
}

score_asymm_sbekk <- function(theta, r, signs) {
    .Call('_BEKKs_score_asymm_sbekk', PACKAGE = 'BEKKs', theta, r, signs)
}

bhh_sbekk <- function(r, theta, max_iter, crit) {
    .Call('_BEKKs_bhh_sbekk', PACKAGE = 'BEKKs', r, theta, max_iter, crit)
}

bhh_asymm_sbekk <- function(r, theta, max_iter, crit, signs) {
    .Call('_BEKKs_bhh_asymm_sbekk', PACKAGE = 'BEKKs', r, theta, max_iter, crit, signs)
}

hesse_sbekk <- function(theta, r) {
    .Call('_BEKKs_hesse_sbekk', PACKAGE = 'BEKKs', theta, r)
}

hesse_asymm_sbekk <- function(theta, r, signs) {
    .Call('_BEKKs_hesse_asymm_sbekk', PACKAGE = 'BEKKs', theta, r, signs)
}

sigma_sbekk <- function(r, C, a, g) {
    .Call('_BEKKs_sigma_sbekk', PACKAGE = 'BEKKs', r, C, a, g)
}

sigma_sbekk_asymm <- function(r, C, a, b, g, signs) {
    .Call('_BEKKs_sigma_sbekk_asymm', PACKAGE = 'BEKKs', r, C, a, b, g, signs)
}

random_grid_search_sBEKK <- function(r) {
    .Call('_BEKKs_random_grid_search_sBEKK', PACKAGE = 'BEKKs', r)
}

random_grid_search_asymmetric_sBEKK <- function(r, signs) {
    .Call('_BEKKs_random_grid_search_asymmetric_sBEKK', PACKAGE = 'BEKKs', r, signs)
}

YLagCr <- function(y, p) {
    .Call('_BEKKs_YLagCr', PACKAGE = 'BEKKs', y, p)
}

test_that("inv_gen works with nonsingular positive definite 2x2 matrix", {
  testm <- matrix(c(0.07188, -0.0124, -0.0124, 0.833),2,2)

  expect_equal(inv_gen(testm), solve(testm))
})

test_that("inv_gen works with singular not positive definite 2x2 matrix", {
  testm <- matrix(c(0.07188, -0.01241, 0, 0),2,2)

  expect_equal(round(inv_gen(testm), 2), matrix(c(13.51, 0, -2.331, 0), 2, 2), tolerance =0.1)
})

test_that("bhh works with 2-dimensional test set", {
  theta <- c(0.062, -0.02689, 0.210177, 0.2236, -0.03, -0.03,
             0.2236, 0.94868, 0.03, -0.03, 0.9487)
  theta <- matrix(theta, nrow = 11)
  res <- bhh_bekk(data.matrix(StocksBonds), theta, max_iter = 50, crit = 1e-9)

  expect_equal(trunc(res$likelihood), -7382)
  expect_equal(res$iter, 13)
})

# test_that("Testing Asymmetric GridSearch works with StocksBonds", {
#   set.seed(56285)
#   signs <- as.matrix(c(-1,1))
#   gs1 <- random_grid_search_asymmetric_BEKK(StocksBonds ,signs)
#
#   # Starting values should be equal for same seed
#
#   expect_equal(trunc(sum(gs1[[1]])), 3)
# })
#
#
# test_that("Testing Asymmetric GridSearch works with GoldStocksBonds", {
#   set.seed(5463782)
#   signs <- as.matrix(c(-1,1,-1))
#   gs1 <- random_grid_search_asymmetric_BEKK(GoldStocksBonds ,signs)
#
#   # Starting values should be equal for same seed
#
#   expect_equal(trunc(sum(gs1[[1]])), 4)
# })

test_that("score_sbekk works with 2-dimensional test data set", {
  init <- c(0.062, -0.02689, 0.210177, 0.1346, 0.74868)
  theta <- matrix(init, nrow = 5)

  score <- score_sbekk(theta, data.matrix(StocksBonds))
  score <- sum(score)
  expect_equal(trunc(score), 55899)
})

test_that("score_sbekk works with 3-dimensional test data set", {

  init=c(0.0015876334,  0.0002258635,  0.0006750149,  0.0014719950, -0.0005589923,  0.0012654071,   0.082,  0.8599385388)
  theta <- matrix(init, nrow = 8)
  score <- score_sbekk(theta, GoldStocksBonds)
  score <- sum(score)
  expect_equal(trunc(score), 2000803)
})

test_that("score_asymm_sbekk works with 2-dimensional test data set", {
  init <- c(0.062, -0.02689, 0.210177, 0.2236, 0.12, 0.94868)
  signs = as.matrix(c(-1, -1))
  theta <- matrix(init, nrow = 6)

  score <- score_asymm_sbekk(theta, data.matrix(StocksBonds),signs)
  score <- sum(score)
  expect_equal(trunc(score), -146558)
})

test_that("score_asymm_sbekk works with 3-dimensional test data set", {
  signs = as.matrix(c(-1, -1, -1))

  init=c(0.0015876334,  0.0002258635,  0.0006750149,  0.0014719950, -0.0005589923,  0.0012654071,  0.2610727800,  0.082,  0.9599385388)
  theta <- matrix(init, nrow = 9)
  score <- score_asymm_sbekk(theta, GoldStocksBonds, signs)
  score <- sum(score)
  expect_equal(trunc(score), -2393928)
})

test_that("Testing GridSearch works with StocksBonds", {
  set.seed(6723217)
  gs1 <- random_grid_search_BEKK(data.matrix(StocksBonds))

  # Starting values should be equal for same seed

  expect_equal(trunc(sum(gs1[[1]])), 0 )
})

# Funtion to extract conditional standard deviations from bekk_fit object
extract_csd <- function(x) {

  csd <- matrix(NA, nrow = nrow(x$sigma_t), ncol = ncol(x$data))
  csd_names <- rep(NA, ncol(x$data))

  counter <- 1
  for (i in 1:ncol(x$data)) {
    csd[, i] <- x$sigma_t[, counter]
    csd_names[i] <- colnames(x$sigma_t)[counter]
    counter <- counter + ncol(x$data) - (i - 1)
  }

  colnames(csd) <- csd_names
  return(csd)
}

process_object <- function(x) {
  UseMethod('process_object')
}

#'@exportS3Method
process_object.bekkFit <- function(x) {
  theta <- x$theta
  N <- ncol(x$C0)
  signs <- x$signs
  BEKK_valid <- x$BEKK_valid
  expected_signs <- x$expected_signs

  return(list(theta = theta,
              N = N, signs=signs, expected_signs = expected_signs, BEKK_valid = BEKK_valid
              ))
}
#'@exportS3Method
process_object.bekkSpec <- function(x) {
  if (is.null(x$init_values)) {
    stop('Please provide "initial_values" in "bekk_spec" as paramater for simulation.')
  }
  if (is.null(x$N)) {
    stop('Please provide "N" in "bekk_spec" as dimension for simulation.')
  }

  theta <- x$init_values
  N <- x$N

  if(is.null(x$signs) && x$model$asymmetric == TRUE){
    signs=as.matrix(rep(-1,N))

  }else{
    signs=x$signs
  }

  if(x$model$asymmetric == FALSE){
    par = coef_mat(theta,N)
    BEKK_valid = valid_bekk(par$c0, par$a, par$g)
  } else{
    par = coef_mat_asymm(theta,N)
    BEKK_valid = valid_asymm_bekk_sim(par$c0, par$a, par$b, par$g, 1/(N^2),signs)
  }

  expected_signs=1/(N^2)
  return(list(theta = theta,
              N = N, signs = signs, expected_signs = expected_signs, BEKK_valid = BEKK_valid))
}

# Obtaining QML t-ratios
QML_t_ratios <- function(theta, r) {
  s1 <- score_bekk(theta, r)
  s1 <- crossprod(s1)

  s2 <- hesse_bekk(theta, r)
  s2 <- solve(s2) %*% s1 %*% solve(s2)

  s2 <- sqrt(abs(diag(s2)))

  return(theta/s2)
}

QML_t_ratios_asymm <- function(theta, r, signs) {
  s1 <- score_asymm_bekk(theta, r, signs)
  s1 <- crossprod(s1)

  s2 <- hesse_asymm_bekk(theta, r, signs)
  s2 <- solve(s2) %*% s1 %*% solve(s2)

  s2 <- sqrt(abs(diag(s2)))

  return(theta/s2)
}

QML_t_ratios_dbekk <- function(theta, r) {
  s1 <- score_dbekk(theta, r)
  s1 <- crossprod(s1)

  s2 <- hesse_dbekk(theta, r)
  s2 <- solve(s2) %*% s1 %*% solve(s2)

  s2 <- sqrt(abs(diag(s2)))

  return(theta/s2)
}

QML_t_ratios_dbekka <- function(theta, r, signs) {
  s1 <- score_asymm_dbekk(theta, r, signs)
  s1 <- crossprod(s1)

  s2 <- hesse_asymm_dbekk(theta, r, signs)
  s2 <- solve(s2) %*% s1 %*% solve(s2)

  s2 <- sqrt(abs(diag(s2)))

  return(theta/s2)
}

QML_t_ratios_sbekk <- function(theta, r) {
  s1 <- score_sbekk(theta, r)
  s1 <- crossprod(s1)

  s2 <- hesse_sbekk(theta, r)
  s2 <- solve(s2) %*% s1 %*% solve(s2)

  s2 <- sqrt(abs(diag(s2)))

  return(theta/s2)
}

QML_t_ratios_sbekk_asymm <- function(theta, r, signs) {
  s1 <- score_asymm_sbekk(theta, r, signs)
  s1 <- crossprod(s1)

  s2 <- hesse_asymm_sbekk(theta, r, signs)
  s2 <- solve(s2) %*% s1 %*% solve(s2)

  s2 <- sqrt(abs(diag(s2)))

  return(theta/s2)
}

QML_sd <- function(theta, r) {
  s1 <- score_bekk(theta, r)
  s1 <- crossprod(s1)

  s2 <- hesse_bekk(theta, r)
  s2 <- solve(s2) %*% s1 %*% solve(s2)

  s2 <- sqrt(abs(diag(s2)))

  return(as.matrix(s2))
}

QML_sd_asymm <- function(theta, r, signs) {
  s1 <- score_asymm_bekk(theta, r, signs)
  s1 <- crossprod(s1)

  s2 <- hesse_asymm_bekk(theta, r, signs)
  s2 <- solve(s2) %*% s1 %*% solve(s2)

  s2 <- sqrt(abs(diag(s2)))

  return(as.matrix(s2))
}

QML_sd_dbekk <- function(theta, r) {
  s1 <- score_dbekk(theta, r)
  s1 <- crossprod(s1)

  s2 <- hesse_dbekk(theta, r)
  s2 <- solve(s2) %*% s1 %*% solve(s2)

  s2 <- sqrt(abs(diag(s2)))

  return(as.matrix(s2))
}

QML_sd_dbekka <- function(theta, r, signs) {
  s1 <- score_asymm_dbekk(theta, r, signs)
  s1 <- crossprod(s1)

  s2 <- hesse_asymm_dbekk(theta, r, signs)
  s2 <- solve(s2) %*% s1 %*% solve(s2)

  s2 <- sqrt(abs(diag(s2)))
  return(as.matrix(s2))
}

QML_sd_sbekk <- function(theta, r) {
  s1 <- score_sbekk(theta, r)
  s1 <- crossprod(s1)

  s2 <- hesse_sbekk(theta, r)
  s2 <- solve(s2) %*% s1 %*% solve(s2)

  s2 <- sqrt(abs(diag(s2)))

  return(as.matrix(s2))
}

QML_sd_sbekk_asymm <- function(theta, r, signs) {
  s1 <- score_asymm_sbekk(theta, r, signs)
  s1 <- crossprod(s1)

  s2 <- hesse_asymm_sbekk(theta, r, signs)
  s2 <- solve(s2) %*% s1 %*% solve(s2)

  s2 <- sqrt(abs(diag(s2)))

  return(as.matrix(s2))
}

#' Performing a Portmanteau test checking for remaining correlation in the empirical co-variances of the estimated BEKK residuals.
#'
#' @description Method for a Portmanteau test of the null hypothesis of no remaining correlation in the co-variances of the estimated BEKK residuals.
#'
#' @param x An object of class "bekkFit" from function \link{bekk_fit}.
#' @param lags An integer defining the lag length.
#' @return  Returns an Object of class "htest" containing the p-value and test statistic.
#'
#' @details Here, the multivariate Portmanteau test of Hosking (1980) is implemented.
#'
#' @references  J. R. M. Hosking (1980). The Multivariate Portmanteau Statistic, Journal of the American Statistical Association, 75:371, 602-608.


#' @import xts
#' @import stats
#' @import ks
#' @export
portmanteau.test <- function(x, lags = 5){
  if(!is.numeric(lags)){
    stop("Please provide a numeric object or vector for 'lags'.")
  }
  if(any(lags<3)){
    stop("Please provide 'lags' larger than 2.")
  }
  UseMethod("portmanteau.test")
}

#' @export
portmanteau.test.bekkFit <- function(x, lags = 5){
  e <- x$e_t
  n <- nrow(e)
  N <- ncol(e)
  e <- matrix(e,n,N)
  #e <- matrix(e, nrow = n, ncol = N)
  e2 <- matrix(NA,nrow = n, ncol = N*(N+1)/2)

  for(i in 1:n){
    e2[i,] <- ks::vech(crossprod(t(e[i,])))
  }
  e=e2

  c_hat <- function(j){
        c= t(e[(j+1):n,]) %*% e[1:(n-j),]
        return(c/n)
  }
  c_0 = c_hat(0)
  #c_0_inv = solve(c_0)
  c_0_inv = chol2inv(chol(c_0))
  Q <- function(lgs){
    q=0
   for(i in 1:lgs){
     c_temp = t(c_hat(i))
     q=q+sum(diag(c_temp%*%c_0_inv%*%c_temp%*%c_0_inv))
   }
    return(q)
  }
  p_val_q <- function(k, lgs){
    return(1-pchisq(k, df=(lgs-2)*(N^2)))
  }



  statistic = Q(lags)
  names(statistic) = "statistic"
  p.value = p_val_q(p_val_q(statistic,lags),lags)
  names(p.value) = "p.value"
  parameter = (lags-2)*(N^2)
  names(parameter) = "df"
  data = "Residuals of estimated BEKK process"



  rval <- list(statistic = statistic, parameter = parameter, p.value = p.value, data.name

               = data, method = paste("Portmanteau Test (Lags = ", as.character(lags), ")", sep = "" ))
  class(rval) <- "htest"
  return(rval)
}

#' @export

summary.bekk <- function(object, ...) {
  bekkObject <- object
  cat(paste("\n", "BEKK estimation results", "\n", sep = ""))
  underScore <- paste(rep("-", nchar("BEKK estimation results")), collapse = "")
  cat(underScore)
  cat("\nLog-likelihood: ")
  cat(bekkObject$log_likelihood)
  cat("\nBEKK model stationary: ")
  cat(bekkObject$BEKK_valid)
  cat("\nNumber of BHHH iterations: ")
  cat(bekkObject$iter)
  cat("\nAIC: ")
  cat(bekkObject$AIC)
  cat("\nBIC: ")
  cat(bekkObject$BIC)
  cat("\nEstimated parameter matrices: \n")
  cat("\nC \n")
  print(bekkObject$C0)
  cat("\nA \n")
  print(bekkObject$A)
  cat("\nG \n")
  print(bekkObject$G)
  cat("\nStandard errors of parameter matrices: \n")
  cat("\nC \n")
  print(bekkObject$C0_sd)
  cat("\nA \n")
  print(bekkObject$A_sd)
  cat("\nG \n")
  print(bekkObject$G_sd)
  cat("\n")
}

#' @export
summary.bekka <- function(object, ...) {
  bekkObject <- object
  cat(paste("\n", "Asymmetric BEKK estimation results", "\n", sep = ""))
  underScore <- paste(rep("-", nchar("Asymmetric BEKK estimation results")), collapse = "")
  cat(underScore)
  cat("\nLog-likelihood: ")
  cat(bekkObject$log_likelihood)
  cat("\nBEKK model stationary: ")
  cat(bekkObject$BEKK_valid)
  cat("\nNumber of BHHH iterations: ")
  cat(bekkObject$iter)
  cat("\nAIC: ")
  cat(bekkObject$AIC)
  cat("\nBIC: ")
  cat(bekkObject$BIC)
  cat("\nEstimated parameter matrices: \n")
  cat("\nC \n")
  print(bekkObject$C0)
  cat("\nA \n")
  print(bekkObject$A)
  cat("\nB \n")
  print(bekkObject$B)
  cat("\nG \n")
  print(bekkObject$G)
  cat("\nStandard errors of parameter matrices: \n")
  cat("\nC \n")
  print(bekkObject$C0_sd)
  cat("\nA \n")
  print(bekkObject$A_sd)
  cat("\nB \n")
  print(bekkObject$B_sd)
  cat("\nG \n")
  print(bekkObject$G_sd)
  cat("\n")
}

#' @export

summary.dbekk <- function(object, ...) {
  bekkObject <- object
  cat(paste("\n", "Diagonal BEKK estimation results", "\n", sep = ""))
  underScore <- paste(rep("-", nchar("Diagonal BEKK estimation results")), collapse = "")
  cat(underScore)
  cat("\nLog-likelihood: ")
  cat(bekkObject$log_likelihood)
  cat("\nDiagonal BEKK model stationary: ")
  cat(bekkObject$BEKK_valid)
  cat("\nNumber of BHHH iterations: ")
  cat(bekkObject$iter)
  cat("\nAIC: ")
  cat(bekkObject$AIC)
  cat("\nBIC: ")
  cat(bekkObject$BIC)
  cat("\nEstimated parameter matrices: \n")
  cat("\nC \n")
  print(bekkObject$C0)
  cat("\nA \n")
  print(bekkObject$A)
  cat("\nG \n")
  print(bekkObject$G)
  cat("\nStandard errors of parameter matrices: \n")
  cat("\nC \n")
  print(bekkObject$C0_sd)
  cat("\nA \n")
  print(bekkObject$A_sd)
  cat("\nG \n")
  print(bekkObject$G_sd)
  cat("\n")
}

#' @export
summary.dbekka <- function(object, ...) {
  bekkObject <- object
  cat(paste("\n", "Asymmetric diagonal BEKK estimation results", "\n", sep = ""))
  underScore <- paste(rep("-", nchar("Asymmetric diagonal BEKK estimation results")), collapse = "")
  cat(underScore)
  cat("\nLog-likelihood: ")
  cat(bekkObject$log_likelihood)
  cat("\nDiagonal BEKK model stationary: ")
  cat(bekkObject$BEKK_valid)
  cat("\nNumber of BHHH iterations: ")
  cat(bekkObject$iter)
  cat("\nAIC: ")
  cat(bekkObject$AIC)
  cat("\nBIC: ")
  cat(bekkObject$BIC)
  cat("\nEstimated parameter matrices: \n")
  cat("\nC \n")
  print(bekkObject$C0)
  cat("\nA \n")
  print(bekkObject$A)
  cat("\nB \n")
  print(bekkObject$B)
  cat("\nG \n")
  print(bekkObject$G)
  cat("\nStandard errors of parameter matrices: \n")
  cat("\nC \n")
  print(bekkObject$C0_sd)
  cat("\nA \n")
  print(bekkObject$A_sd)
  cat("\nB \n")
  print(bekkObject$B_sd)
  cat("\nG \n")
  print(bekkObject$G_sd)
  cat("\n")
}


#' @export

summary.sbekk <- function(object, ...) {
  bekkObject <- object
  cat(paste("\n", "Scalar BEKK estimation results", "\n", sep = ""))
  underScore <- paste(rep("-", nchar("Scalar BEKK estimation results")), collapse = "")
  cat(underScore)
  cat("\nLog-likelihood: ")
  cat(bekkObject$log_likelihood)
  cat("\nScalar BEKK model stationary: ")
  cat(bekkObject$BEKK_valid)
  cat("\nNumber of BHHH iterations: ")
  cat(bekkObject$iter)
  cat("\nAIC: ")
  cat(bekkObject$AIC)
  cat("\nBIC: ")
  cat(bekkObject$BIC)
  cat("\nEstimated parameter matrices: \n")
  cat("\nC \n")
  print(bekkObject$C0)
  cat("\na \n")
  print(bekkObject$a)
  cat("\ng \n")
  print(bekkObject$g)
  cat("\nStandard errors of parameter matrices: \n")
  cat("\nC \n")
  print(bekkObject$C0_sd)
  cat("\na \n")
  print(bekkObject$a_sd)
  cat("\ng \n")
  print(bekkObject$g_sd)
  cat("\n")
}

#' @export
summary.sbekka <- function(object, ...) {
  bekkObject <- object
  cat(paste("\n", "Asymmetric scalar BEKK estimation results", "\n", sep = ""))
  underScore <- paste(rep("-", nchar("Asymmetric scalar BEKK estimation results")), collapse = "")
  cat(underScore)
  cat("\nLog-likelihood: ")
  cat(bekkObject$log_likelihood)
  cat("\nScalar BEKK model stationary: ")
  cat(bekkObject$BEKK_valid)
  cat("\nNumber of BHHH iterations: ")
  cat(bekkObject$iter)
  cat("\nAIC: ")
  cat(bekkObject$AIC)
  cat("\nBIC: ")
  cat(bekkObject$BIC)
  cat("\nEstimated parameter matrices: \n")
  cat("\nC \n")
  print(bekkObject$C0)
  cat("\na \n")
  print(bekkObject$a)
  cat("\nb \n")
  print(bekkObject$b)
  cat("\ng \n")
  print(bekkObject$g)
  cat("\nStandard errors of parameter matrices: \n")
  cat("\nC \n")
  print(bekkObject$C0_sd)
  cat("\na \n")
  print(bekkObject$a_sd)
  cat("\nb \n")
  print(bekkObject$b_sd)
  cat("\ng \n")
  print(bekkObject$g_sd)
  cat("\n")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/virf.R
\name{virf}
\alias{virf}
\title{Estimating multivariate volatility impulse response functions (VIRF) for BEKK models}
\usage{
virf(
  x,
  time = 1,
  q = 0.05,
  index_series = 1,
  n.ahead = 10,
  ci = 0.9,
  time_shock = FALSE
)
}
\arguments{
\item{x}{An object of class "bekkfit" from function \link{bekk_fit}.}

\item{time}{Time instance to calculate VIRFs for.}

\item{q}{A number specifying the quantile to be considered for a shock on which basis the VIRFs are generated.}

\item{index_series}{An integer defining the number of series for which a shock is assumed.}

\item{n.ahead}{An integer defining the number periods for which the VIRFs are generated.}

\item{ci}{A number defining the confidence level for the confidence bands.}

\item{time_shock}{Boolean indicating if the estimated residuals at date specified by "time" shall be used as a shock.}
}
\value{
Returns an object of class "virf".
}
\description{
Method for estimating VIRFs of N-dimensional BEKK models. Currently, only VIRFs for symmetric BEKK models are implemented.
}
\examples{
\donttest{

data(StocksBonds)
obj_spec <- bekk_spec()
x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)

# 250 day ahead VIRFs and 90\% CI for a Shock in the 1\% quantile  of Bonds (i.e. series=2)
# shock is supposed to occur at day 500
x2 <- virf(x1, time = 500, q = 0.01, index_series=2, n.ahead = 500, ci = 0.90)
plot(x2)
}
}
\references{
Hafner CM, Herwartz H (2006). Volatility impulse responses for multivariate GARCH models:  An exchange rate illustration. Journal of International Money and Finance,25,719–740.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bekk_spec.R
\name{bekk_spec}
\alias{bekk_spec}
\title{BEKK specification method}
\usage{
bekk_spec(
  model = list(type = "bekk", asymmetric = FALSE),
  init_values = NULL,
  signs = NULL,
  N = NULL
)
}
\arguments{
\item{model}{A list containing the model type specification: Either "bekk" "dbekk" or "sbekk".
Moreover it can be specified whether the model should be estimated allowing for asymmetric volatility structure.}

\item{init_values}{initial values for \link{bekk_fit} during BHHH algorithm. It can be either a numerical vector of suitable dimension, 'NULL' (default) to use a simple grid search algorithm, or a character vector i.e. "random" to use a random starting value generator (set a seed in advance for reproducible results), or
"simple" for relying on a simple initial values generator based on typical values for BEKK parameter found in the literature. If the object from this function is passed to \link{simulate}, "init_values" are used as parameters for data generating process.}

\item{signs}{An N-dimensional vector consisting of "1" or "-1" to indicate the asymmetric effects to be considered.
Setting the i-th element of the vector to "1" or "-1" means that the model takes into account additional volatility if the returns of the i-th column in the data matrix are either positive or negative.
If "asymmetric = TRUE", the default is set to "rep(-1, N)" i.e. it is assumed that excess volatility occurs for all series if the returns are negative.}

\item{N}{Integer specifying the dimension of the BEKK model. Only relevant when this object of class "bekkSpec"" is used for simulating BEKK processes by applying it to \link{simulate}.}
}
\value{
Returns a S3 class "bekkSpec"  object containing the specifications of the model to be estimated.
}
\description{
Method for creating a N-dimensional BEKK model specification object prior to fitting and/or simulating.
}
\examples{
\donttest{

data(StocksBonds)

# Fitting a symmetric BEKK model using default starting values
# - i.e. fixed values
obj_spec_fixed <- bekk_spec(init_values = NULL)
x1 <- bekk_fit(obj_spec_fixed, StocksBonds, QML_t_ratios = FALSE,
max_iter = 50, crit = 1e-9)
# Fitting a symmetric BEKK model using initial values originating from a
# random grid search algorithm
obj_spec_random <- bekk_spec(init_values = "random")
x2 <- bekk_fit(obj_spec_random, StocksBonds, QML_t_ratios = FALSE,
max_iter = 50, crit = 1e-9)
summary(x1)
summary(x2)
plot(x1)
plot(x2)
# Fitting an asymmetric BEKK model with default starting values
obj_spec_fix <- bekk_spec(model = list(type = "bekk", asymmetric = TRUE),
init_values = NULL)
x1 <- bekk_fit(obj_spec_fix, StocksBonds)
obj_spec_random <- bekk_spec(model = list(type = "bekk", asymmetric = TRUE),
init_values = "random")
x2 <- bekk_fit(obj_spec_random, StocksBonds)
summary(x1)
summary(x2)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/VaR.R
\name{VaR}
\alias{VaR}
\title{Calculating Value-at-Risk (VaR)}
\usage{
VaR(x, p = 0.99, portfolio_weights = NULL, distribution = "empirical")
}
\arguments{
\item{x}{An object of class "bekkFit" from the function \link{bekk_fit} or an object of class "bekkForecast" from the function \link{predict}.}

\item{p}{A numerical value that determines the confidence level. The default value is set at 0.99 in accordance with the Basel Regulation.}

\item{portfolio_weights}{A vector determining the portfolio weights to calculate the portfolio VaR. If set to "NULL", the univariate VaR for each series are calculated.}

\item{distribution}{A character string determining the assumed distribution of the residuals. Implemented are "normal", "empirical" and "t". The default is using the empirical distribution of the residuals.}
}
\value{
Returns a S3 class "var" object containing the VaR forecast and respective confidence bands.
}
\description{
Method for calculating VaR from estimated covariance processes (\link{bekk_fit}) or predicted covariances (\link{predict}).
}
\examples{
\donttest{

data(StocksBonds)
obj_spec <- bekk_spec()
x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)

# single VaRs of series
x2 <- VaR(x1, distribution="normal")
plot(x2)

# VaR of equally-weighted portfolio
portfolio_weights <- c(0.5, 0.5)
x3 <- VaR(x1, portfolio_weights = portfolio_weights)
plot(x3)

# VaR of traditional 30/70 weighted bond and stock portfolio
portfolio_weights <- c(0.3, 0.7)
x4 <- VaR(x1, portfolio_weights = portfolio_weights)
plot(x4)

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bekk_fit_methods.R
\name{print.bekkFit}
\alias{print.bekkFit}
\alias{residuals.bekkFit}
\alias{logLik.bekkFit}
\title{bekkFit method}
\usage{
\method{print}{bekkFit}(x, ...)

\method{residuals}{bekkFit}(object, ...)

\method{logLik}{bekkFit}(object, ..., k = 2)
}
\arguments{
\item{x}{An object of class "bekkFit" from function \link{bekk_fit}.}

\item{...}{Further arguments to be passed to and from other methods.}

\item{object}{An object of class "bekkFit" from function \link{bekk_fit}.}

\item{k}{Numeric value, the penalty per parameter for AIC to be used; the default k = 2 is the classical AIC.}
}
\description{
Generic 'bekkFit' methods. More details on 'bekkFit' are described in \link{bekk_fit}
}
\examples{
\donttest{

data(StocksBonds)

# Fitting a symmetric BEKK model
obj_spec <- bekk_spec()
x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)

logLik(x1)
}
}

\name{GoldStocksBonds}
\alias{GoldStocksBonds}
\docType{data}
\title{
Gold stock and Bond returns
}
\description{
Trivariate data set consisting of daily gold, S&P 500 and U.S. Treasury Bond Future returns from October 1991 to October 2021.
}
\usage{data("GoldStocksBonds")}
\format{
  A time series matrix of class mts 7346 observations on the following 3 variables.
  \describe{
    \item{Gold}{a numeric vector}
    \item{S&P 500}{a numeric vector}
    \item{US Treasury Bond Future}{a numeric vector}
  }
}

\source{
Yahoo Finance.
}
\examples{
data(GoldStocksBonds)
## maybe str(GoldStocksBonds) ; plot(GoldStocksBonds) ...
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bekk_forecast.R
\name{predict}
\alias{predict}
\alias{predict.bekk}
\alias{predict.bekka}
\alias{predict.dbekk}
\alias{predict.dbekka}
\alias{predict.sbekk}
\alias{predict.sbekka}
\title{Forecasting conditional volatilities with BEKK models}
\usage{
\method{predict}{bekk}(object, n.ahead = 1, ci = 0.95, ...)

\method{predict}{bekka}(object, n.ahead = 1, ci = 0.95, ...)

\method{predict}{dbekk}(object, n.ahead = 1, ci = 0.95, ...)

\method{predict}{dbekka}(object, n.ahead = 1, ci = 0.95, ...)

\method{predict}{sbekk}(object, n.ahead = 1, ci = 0.95, ...)

\method{predict}{sbekka}(object, n.ahead = 1, ci = 0.95, ...)
}
\arguments{
\item{object}{A fitted bekk model of class "bekkFit" from the \link{bekk_fit} function}

\item{n.ahead}{Number of periods to forecast conditional volatility. Default is a one-period ahead forecast.}

\item{ci}{Floating point in [0,1] defining the niveau for confidence bands of the conditional volatility forecast. Default is 95 per cent niveau confidence bands.}

\item{...}{Further parameters to be passed on to the function.}
}
\value{
Returns a S3 class "bekkForecast" object containing the conditional volatility forecasts and respective confindence bands.
}
\description{
Method for predicting a N-dimensional BEKK covariances.
}
\examples{
\donttest{#'
data(StocksBonds)
obj_spec <- bekk_spec()
x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)

x2 <- predict(x1, n.ahead = 1)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Portmanteau_test.R
\name{portmanteau.test}
\alias{portmanteau.test}
\title{Performing a Portmanteau test checking for remaining correlation in the empirical co-variances of the estimated BEKK residuals.}
\usage{
portmanteau.test(x, lags = 5)
}
\arguments{
\item{x}{An object of class "bekkFit" from function \link{bekk_fit}.}

\item{lags}{An integer defining the lag length.}
}
\value{
Returns an Object of class "htest" containing the p-value and test statistic.
}
\description{
Method for a Portmanteau test of the null hypothesis of no remaining correlation in the co-variances of the estimated BEKK residuals.
}
\details{
Here, the multivariate Portmanteau test of Hosking (1980) is implemented.
}
\references{
J. R. M. Hosking (1980). The Multivariate Portmanteau Statistic, Journal of the American Statistical Association, 75:371, 602-608.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bekk_sim.R
\name{simulate}
\alias{simulate}
\alias{simulate.bekk}
\alias{simulate.bekka}
\alias{simulate.dbekk}
\alias{simulate.dbekka}
\alias{simulate.sbekk}
\alias{simulate.sbekka}
\title{Simulating BEKK models}
\usage{
\method{simulate}{bekk}(object, nsim, ...)

\method{simulate}{bekka}(object, ..., nsim)

\method{simulate}{dbekk}(object, ..., nsim)

\method{simulate}{dbekka}(object, ..., nsim)

\method{simulate}{sbekk}(object, ..., nsim)

\method{simulate}{sbekka}(object, ..., nsim)
}
\arguments{
\item{object}{A spec object of class "bekkSpec" from the function \link{bekk_spec} or a fitted bekk model of class "bekkFit" from the \link{bekk_fit} function}

\item{nsim}{Number of observations of the simulated sample}

\item{...}{Further parameters to be passed on to the function.}
}
\value{
Returns a simulated time series S3 class object using the parameters of passed "bekkSpec" or "bekkFit".
}
\description{
Method for simulating a N-dimensional BEKK model.
}
\examples{
\donttest{

# simulate a BEKK with estimated parameter
obj_spec <- bekk_spec()
x1 <- bekk_fit(obj_spec, StocksBonds)

x2 <- simulate(x1, nsim = 3000)

plot(x2)

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bekk_fit.R
\name{bekk_fit}
\alias{bekk_fit}
\title{Estimating multivariate BEKK-type volatility models}
\usage{
bekk_fit(spec, data, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-09)
}
\arguments{
\item{spec}{An object of class "bekkSpec" from function \link{bekk_spec}.}

\item{data}{A multivariate data object. Can be a numeric matrix or ts/xts/zoo object.}

\item{QML_t_ratios}{Logical. If QML_t_ratios = 'TRUE', the t-ratios of the BEKK parameter matrices
are exactly calculated via second order derivatives.}

\item{max_iter}{Maximum number of BHHH algorithm iterations.}

\item{crit}{Determines the precision of the BHHH algorithm.}
}
\value{
Returns a S3 class "bekkFit" object containing the estimated parameters, t-values, standard errors and volatility process of the model defined by the BEKK_spec object.
}
\description{
Method for fitting a variety of N-dimensional BEKK models.
}
\details{
The BEKK optimization routine is based on the Berndt–Hall–Hall–Hausman (BHHH) algorithm and is inspired by the study of Hafner and Herwartz (2008).
The authors provide analytical formulas for the score and Hessian of several MGARCH models in a QML framework and show that analytical derivations significantly outperform numerical methods.
}
\examples{
\donttest{

data(StocksBonds)

# Fitting a symmetric BEKK model
obj_spec <- bekk_spec()
x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)

summary(x1)

plot(x1)

# Fitting an asymmetric BEKK model
obj_spec <- bekk_spec(model = list(type = "bekk", asymmetric = TRUE))
x1 <- bekk_fit(obj_spec, StocksBonds)

summary(x1)

plot(x1)

# Fitting a symmetric diagonal BEKK model
obj_spec <- bekk_spec(model = list(type = "dbekk", asymmetric = FALSE))
x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)

summary(x1)

plot(x1)


# Fitting a symmetric scalar BEKK model
obj_spec <- bekk_spec(model = list(type = "sbekk", asymmetric = FALSE))
x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)

summary(x1)

plot(x1)

}
}
\references{
Hafner and Herwartz (2008). Analytical quasi maximum likelihood inference in multivariate volatility models. Metrika, 67, 219-239.

Fülle, M. J., A. Lange, C. M. Hafner, and H. Herwartz (2024). BEKKs: An R package for estimation of
conditional volatility of multivariate time series. Journal of Statistical Software 111 (4), 1–34. <doi:10.18637/jss.v111.i04>.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backtest.R
\name{backtest}
\alias{backtest}
\title{Backtesting via Value-at-Risk (VaR)}
\usage{
backtest(
  x,
  window_length = 1000,
  p = 0.99,
  portfolio_weights = NULL,
  n.ahead = 1,
  distribution = "empirical",
  nc = 1
)
}
\arguments{
\item{x}{An object of class "bekkFit" from the function \link{bekk_fit}.}

\item{window_length}{An integer specifying the length of the rolling window.}

\item{p}{A numerical value that determines the confidence level. The default value is set at 0.99 in accordance with the Basel Regulation.}

\item{portfolio_weights}{A vector determining the portfolio weights to calculate the portfolio VaR. If set to "NULL", the univariate VaR for each series are calculated.}

\item{n.ahead}{Number of periods to predict conditional volatility. Default is a one-period ahead forecast.}

\item{distribution}{A character string determining the assumed distribution of the residuals. Implemented are "normal", "empirical" and "t". The default is assuming the empirical distribution of the residuals.}

\item{nc}{Number of cores to be used for parallel computation.}
}
\value{
Returns a S3 class "backtest" object containing the VaR forecast, out-of-sample returns and backtest statistics according to the R-package "GAS". conf
}
\description{
Method for backtesting a model obtained from \link{bekk_fit} in terms of VaR-forcasting accuracy using a rolling window approach.
}
\examples{
\donttest{
\dontshow{Sys.setenv("OMP_THREAD_LIMIT"="1")}
data(StocksBonds)
obj_spec <- bekk_spec()
x1 <- bekk_fit(obj_spec, StocksBonds, QML_t_ratios = FALSE, max_iter = 50, crit = 1e-9)

# backtesting
x2 <- backtest(x1, window_length = 6000, n.ahead = 1, nc = 1)
plot(x2)
# backtesting using 5 day-ahead forecasts
x3 <- backtest(x1, window_length = 6000, n.ahead = 5, nc = 1)
plot(x3)
# backtesting using 20 day-ahead forecasts and portfolio
x4 <- backtest(x1, window_length = 6000, portfolio_weights = c(0.5,0.5), n.ahead = 20, nc = 1)
plot(x4)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BEKKs.R
\name{BEKKs}
\alias{BEKKs}
\title{BEKKs: Volatility modelling}
\description{
\loadmathjax
This package implements estimation, simulation and forecasting techniques for conditional volatility modelling using the BEKK model. We refer the reader to Fülle et al. (2024) for a package overview.
The full BEKK(1,1,1) model of Engle and Kroner (1995)
\mjdeqn{H_t = CC' + A' r_{t-1} r_{t-1}'A + G' H_{t-1}G ,}{H_t = CC' + A' r_{t-1} r_{t-1}'A + G' H_{t-1}G ,}  the asymmetric extensions of Kroner and Ng (1998) and Grier et. al. (2004)
\mjdeqn{H_t = CC' + A' r_{t-1} r_{t-1}'A +B'\gamma_{t-1} \gamma_{t-1}' B+G'H_{t-1}G}{H_t = CC' +A'r_{t-1} r_{t-1}'A +B'\gamma_{t-1} \gamma_{t-1}' B+G'H_{t-1}G,,}
with \mjdeqn{\gamma_t = r_t I\left(r_t lower than 0 \right)}{\gamma_t = r_t I(r_t &lt; 0 )} are implemented.
Moreover, the diagonal BEKK, where the parameter matrices A, B and G are reduced to diagonal matrices and
the scalar BEKK model of Ding and Engle (2001)
\mjdeqn{H_t = CC' + a r_{t-1} r_{t-1}' + g H_{t-1},}{H_t = CC' + a r_{t-1} r_{t-1}' + g H_{t-1},}
where a and g are scalar parameters and are implemented to allow faster but less flexible estimation in higher dimensions.
}
\details{
The main functions are:
\itemize{
\item \tabular{ll}{ \code{\link{bekk_spec}} \tab Specifies the model type to be estimated.}
\item \tabular{ll}{\code{\link{bekk_fit}} \tab Estimates a BEKK(1,1,1) model of a given series and
specification object \link{bekk_spec}.}
\item \tabular{ll}{ \code{\link{simulate}} \tab Simulates a BEKK(1,1,1) process using either a \link{bekk_fit} or \link{bekk_spec} object.}
\item \tabular{ll}{ \code{\link{predict}} \tab Forecasts  conditional volatility using a \link{bekk_fit} object.}
\item \tabular{ll}{ \code{\link{VaR}} \tab Estimates (portfolio) Value-at-Risk using a fitted BEKK(1,1,1) model.}
\item \tabular{ll}{ \code{\link{backtest}} \tab Backtesting estimated (portfolio) value-at-risks of a fitted BEKK(1,1,1) model.}
\item \tabular{ll}{ \code{\link{virf}} \tab Calculates volatility impulse response functions for fitted symmetric BEKK(1,1,1) models.}
}
}
\references{
Engle, R. F. and K. F. Kroner (1995). Multivariate simultaneous generalized arch. Econometric Theory 11(1),122-150.

Fülle, M. J., A. Lange, C. M. Hafner, and H. Herwartz (2024). BEKKs: An R package for estimation of
conditional volatility of multivariate time series. Journal of Statistical Software 111 (4), 1–34. \url{doi:10.18637/jss.v111.i04}.

Kroner, K. F. and V. K. Ng (1998). Modeling asymmetric comovements of asset returns. Review of Financial Studies 11(4), 817-44.

Ding, Zhuanxin and Engle, Robert F (2001). Large scale conditional covariance matrix modeling, estimation and testing. NYU working paper No. Fin-01-029.

Grier, K. B., Olan T. Henry, N. Olekalns, and K. Shields (2004). The asymmetric effects of uncertainty on inflation and output growth. Journal of Applied Econometrics  19(5), 551-565.

Hafner CM, Herwartz H (2006). Volatility impulse responses for multivariate GARCH models:  An exchange rate illustration. Journal of International Money and Finance,25,719-740.
}
\author{
\itemize{
\item Markus J. Fülle  \email{fuelle@uni-goettingen.de}
\item Helmut Herwartz \email{hherwartz@uni-goettingen.de}
\item Alexander Lange \email{alexander.lange@uni-goettingen.de}
\item Christian M. Hafner \email{christian.hafner@uclouvain.be}
}
}

\name{StocksBonds}
\alias{StocksBonds}
\docType{data}
\title{
Daily stock and Bond returns
}
\description{
Bivariate data set consisting of daily S&P 500 bond and MSCI World returns from December 1995 to December 2019.
}
\usage{data("StocksBonds")}
\format{
  A time series matrix of class mts with 6073 observations on the following 2 variables.
  \describe{
    \item{S&P 500 Bonds}{a numeric vector}
    \item{MSCI World}{a numeric vector}
  }
}
\source{
Yahoo Finance.
}
\examples{
data(StocksBonds)
## maybe str(StocksBonds) ; plot(StocksBonds) ...
}
\keyword{datasets}
